/***************************************************************************
 *
 * Copyright (C) 2007-2008  SMSC
 *
 * This driver supports 9118/7/6/5 , 9218/7/5, 9211/10, 9221/20 SMSC LAN Ethernet controllers 
 ***************************************************************************
 * File: Lan9118End.c
 *
 * modification history
 * -------------------------------------
 * 09/03/2008 (v3.00) - AH
 *     - Cleaned up and updated per latest changes in LAN9118 PAS
 *     - Removed registers definition from the header that is reserved in the PAS
 *     - Removed support for the following Chip ID to comply with the latest PAS
 *       (0x01180000, 0x01170000, 0x01180001, 0x01170001, 0x01160001, 0x01150001)
 *       Note: Driver won't load in case it sees one of the above IDs.
 *     - Revised and modified to comply with read-after-write time delay restriction
 *     - Fixed bug 2877 "AutoMdix hardware strap override option via insmod does not work"
 * 05/30/2008 (v2.08) - AH
 *      - Added support for LAN9221/20 (Bedford Family).
 * 01/24/2008 (v2.07) - AH
 *      - Updated Phy_GetAutoMdixSts() to reflect AutoMDIX user parameter setup.
 *      - Changed default AMDIX in interface init string from 2 (AutoMDIX enable) to 3 (controlled by strap)
 *      - Fixed driver flow control bug.
 * 01/18/2008 (v2.06) - AH
 *      - To avoid any confusion in ifShow counters, END_ADD_IF_INERR (input error counter add macro) is removed from the code.
 *      - Added MAC address change support. "macAddrString" is deleted from interface init string and link will come up with
 *        either valid MAC address stored in EEPROM or default MAC address (00:80:0F:11:70:00).
 * 01/16/2008 (v2.05) - AH
 *      - Changing the link mode from 10FD to 10HD causes the driver to hang. To resolve the issue, MAC Rx is turned off when the link
 *        goes down and turned back on after link comes back up and becomes stable.
 *      - Update interface counters in if structure which updates ifShow counters.
 *      - Added prompt to show link up/down and speed/duplex mode
 * 01/11/2008 (v2.04) - AH
 *      - Added mux call (muxTxRestart) to restart Tx after waking the stopped Tx queue. Without this call, after stopping the Tx queue
 *        (i.e. changing link speed/duplex while traffic is running), the stack will never retransmitte and the driver will hang.
 *      - Added polling counter for RxFifo status check and limited to RX_POLLING_COUNT before re-enabling the interrupt
 * 01/04/2008 (v2.03) - AH
 *      - Modified pingLib.c to resolve max ping length issue which is limited to 8192 bytes in VxWorks 5.5.
 *         _pingSockBufSize is added in pingLib.c to control transmit and receive socket buffer length and 
 *         PING_MAXPACKET changed to 65536 bytes. _pingSockBufSize default value is 8192 byte that can be changed in run time.
 *      - Modified ftp_client.c test case to add ftp_put loop. ftploop is also changed to ftp_get. 
 * 12/18/2007 (v2.02) - AH
 *      - DMA cycle mode changed from cycle-steal to burst mode. In cycle-steal mode in this platform (w/ SH7727), DMA counter returns zero
 *        as soon as the DMA starts and cpu gets confused about the end of the cycle.In multi-segment packet transmit,
 *        there is a chance that one buffer sent to TX FIFO in the middle of the previous segment which will eventually cause
 *        transmit error. (TXE in INT_STS). DMA_CYCLE_STEAL is added to select DMA cycle.
 *      - Added Boylston Auto family support
 * 12/10/2007 (v2.01) - AH
 *      - Added compiler directive UNCASHED_TX_BUF to control Tx Buffer Memory allocation
 *        By enabling this option, flushing the DMA buffer (takes 100 usec) in Tx routine is not necessay.
 *      - Added compiler directive SCATTER_GATHER_EN to enable fragmented packet transmitte w/o copying
 *        Mblks into a contiguous buffer (for 1500 byte packet takes 60 usec)
 *        Wind OS devides 1500 byte ping into fragmented ICMPs with the following packet size:
 *            - 1st Packet: 1514 byte
 *            - 2nd Packet: Frag1: 36 bytes, Frag2: 20 bytes
 *        The following is transfer time to Tx fifo in Ver2.00 and V2.01 w/ SCATTER_GATHER_EN
 *             Ver2.00         Ver2.01
 *        DMA: 400 usec        343 usec
 *        PIO: 362 usec        300 usec
 *
 * 12/04/2007 (v2.00) - AH
 *		- Complete Driver Overhaul
 *		- Still needs code clean up
 */


/*

NOTE: External Interface

The only external interface is the LAN9118EndLoad() routine, which expects
the <initString> parameter as input.  This parameter passes in a 
colon-delimited string of the format:

"<unit>:<baseAdrs>:<intVec>:<intLvl>:<bufSize>:<link_mode>:<macAddrString>"

The LAN9118EndLoad() function uses strtok() to parse the string.

TARGET-SPECIFIC PARAMETERS

.IP "unit"

A convenient holdover from the former model.  This parameter is used only
in the string name for the driver.

.IP "baseAdrs"

Base address at which the NIC hardware device registers are located.

.IP "vecNum"

This is the interrupt vector number of the hardware interrupt generated by
this Ethernet device.

.IP "intLvl"

This parameter defines the level of the hardware interrupt.

.IP "bufSize"

Size of the NIC buffer memory in bytes.

.IP "link_mode"

Link mode options:
	The driver parameter link_mode is a bit wise field that specifies
	any combination of 4 or fewer link speeds
	0x01=10HD
	0x02=10FD
	0x04=100HD
	0x08=100FD
	0x10=Symmetrical Pause
	0x20=Asymmetrical Pause
	0x40=Auto Negitiate

.IP "AutoMdix"
	AutoMdix string enable s are disable the MDXI facility in 9218/6/5
	for  other Ehternet controller give numerical value >3


.IP "macAddrString"
	ASCII string of MAC address value in hexadecimal.

EXTERNAL SUPPORT REQUIREMENTS

This driver requires four external support functions, and provides a hook
function:

.iP "void sysLanIntEnable (int level)" "" 9 -1

This routine provides a target-specific interface for enabling Ethernet device
interrupts at a specified interrupt level.

.iP "void sysLanIntDisable (void)" "" 9 -1

This routine provides a target-specific interface for disabling Ethernet device
interrupts.

SYSTEM RESOURCE USAGE

This driver requires the following system resources:

    - one mutual exclusion semaphore
    - one interrupt vector
 */



/* configurations */

#undef	VXWORKS_DEBUG	   /**/ /*	 #define 	VXWORKS_DEBUG        1  */	/* Enable debug messages */
#undef	LAN9118_INSTRUMENT /**/ /*	 #define	LAN9118_INSTRUMENT   1	*/	/* instrument the driver */
#undef	LAN9118_DEBUG 	   /**/ /*	 #define	LAN9118_DEBUG        1  */	/* log 118 debug messages */

#undef	UNCASHED_TX_BUF	   /**/ /*	 #define	UNCASHED_TX_BUF      1  	/* Uncashed Memory for Tx Buffer */
/*#undef	SCATTER_GATHER_EN */   	 #define	SCATTER_GATHER_EN    1    /* Scatter-Gather method for Tx */
#undef	DMA_CYCLE_STEAL    /**/ /*	 #define	DMA_CYCLE_STEAL      1  	/* DMA Cycle Mode */

/* for debugging purpose only */
/* To trace the receive/transmitte path in the driver, enable one of the following option at a time */
#undef	RX_LED_TRACE_EN  /* #define 	RX_LED_TRACE_EN        1  */
#undef	TX_LED_TRACE_EN  /* #define 	TX_LED_TRACE_EN        1  */

/* includes */
#include "vxWorks.h"
#include "iv.h"
#include "vme.h"
#include "net/mbuf.h"
#include "lstLib.h"
#include "semLib.h"
#include "sys/times.h"
#include "net/unixLib.h"
#include "net/protosw.h"
#include "sys/socket.h"
#include "sys/ioctl.h"
#include "errno.h"
#include "memLib.h"
#include "intLib.h"
#include "net/route.h"
#include "iosLib.h"
#include "errnoLib.h"
#include "logLib.h"

#include "cacheLib.h"
#include "netLib.h"
#include "stdio.h"
#include "stdlib.h"
#include "sysLib.h"

#include "etherLib.h"
#include "net/systm.h"
#include "net/if_subr.h"

#include "ifLib.h"
#include "ipProto.h"
#include "usrConfig.h"

#ifdef WR_IPV6
#include "adv_net.h"
#endif /*WR_IPV6*/

#include  "LAN9118End.h"

/*#undef	ETHER_MAP_IP_MULTICAST*/

#include "etherMultiLib.h"
#include "end.h"
#include "endLib.h"

#define TX_DMA	0
#define RX_DMA	1

#define RX_POLLING_COUNT    10

/**************************************************************************/
#ifdef LAN9118_DEBUG
int	LAN9118Debug_mode = LAN9118_DEBUG_GPIO;
#else
int	LAN9118Debug_mode = 0;
#endif

BOOL E2Prm_LoadFlag=0;


typedef struct _DMA_XFER 
{
	UINT32 dwLanReg;
	UINT32 *pdwBuf;
	UINT32 dwDmaCh;
	UINT32 dwDwCnt;
	UINT32 fMemWr;
} DMA_XFER;

typedef struct _FLOW_CONTROL_PARAMETERS
{
	UINT32 MaxThroughput;
	UINT32 MaxPacketCount;
	UINT32 PacketCost;
	UINT32 BurstPeriod;
	UINT32 IntDeas;
} FLOW_CONTROL_PARAMETERS, *PFLOW_CONTROL_PARAMETERS;

typedef struct rx_frame
    {
    UINT16		pad1;
    ENET_HDR	enetHdr;		/* ethernet header */
    UINT8		data[ETHERMTU + 3 * PLATFORM_CACHE_LINE_BYTES];	/* frame data + CRC (rounded to a cache line) + 2 * cache lines  */
    } LAN9118_RX_FRAME;


typedef char* LAN9118_CLUSTER;

/*********************************************************************************/
/* Mutual exclusion related functions
/**********************************************************************************/
typedef unsigned long VL_KEY;

typedef struct _VERIFIABLE_LOCK {
	SEM_ID Lock;
	VL_KEY KeyCode;
} VERIFIABLE_LOCK, * PVERIFIABLE_LOCK;

void Vl_InitLock(PVERIFIABLE_LOCK pVl);
BOOL Vl_CheckLock(PVERIFIABLE_LOCK pVl,VL_KEY keyCode);
VL_KEY Vl_WaitForLock(PVERIFIABLE_LOCK pVl,UINT32 *pdwIntFlags);
void Vl_ReleaseLock(PVERIFIABLE_LOCK pVl,VL_KEY keyCode,UINT32 *pdwIntFlags);


/*********************************************************************************/
/* END Device Data Structure
/**********************************************************************************/
typedef struct LAN9118_end_device	/* driver control structure */
    {
    END_OBJ	          endObj;                         /* The class we inherit from. */
    END_ERR	          lastError;                      /* Last error passed to muxError */
    int               unit;                           /* unit number */
    int			      rringSize;                      /* RMD ring size */
    UCHAR             enetAddr[LAN9118_EADR_LEN];	  /* ethernet address */
    UINT32            dwLanBase;                      /* address of NIC chip */
    LAN9118_CLUSTER	  pCluster;                       /* Rx frame memory */
    CL_POOL_ID        pClPoolId;                      /* cluster pool */
    UINT32         	  ivec;                           /* interrupt vector */
    UINT32            ilevel;                         /* interrupt level */
	UINT8             AutoMdix;
    UINT8             regPage;                        /* current register page */
    UINT8             txStartPage;                    /* ptr to Tx start buffer */
    UINT8             rxStartPage;                    /* ptr to Rx start buffer */
    UINT8             rxStopPage;                     /* ptr to Rx stop buffer */
    UINT8             nextPage;                       /* ptr to next page */
    LAN9118_DRV_FLAG  flags;                          /* device specific flags */
    BOOL              rxHandling;                     /* rcv task is scheduled */
    BOOL              txBlocked;                      /* to implement flow control */
    BOOL              rxOvw;                          /* overwrite warning */
    BOOL              txResend;

    char              macAddrString[13];
    UINT32            dwBitWidth;
    UINT32            irq_pol;
    UINT32            irq_type;
    UINT32            link_mode;
    UINT32            phy_addr;

    UINT32            dwIdRev;
    UINT32            dwGeneration;
    UINT32            dwFpgaRev;

    BOOL              UseScatterGather;
    BOOL              UseTxCsum;
    BOOL              UseRxCsum;

    UINT32            LanInitialized;
	BOOL              ExtPhy;

    SEM_ID            IntEnableLock;
    VERIFIABLE_LOCK   MacPhyLock;

	UINT8             *txData;
    UINT32            dwTxDmaCh;
    UINT32            TxDmaChReserved;
    DMA_XFER          TxDmaXfer;
    UINT32            dwTxDmaThreshold;
    UINT32            dwTxQueueDisableMask;
    UINT32            TxInitialized;

    UINT32            dwRxDmaCh;
	DMA_XFER          RxDmaXfer;
    UINT32            RxDmaChReserved;
    UINT32            dwRxDmaThreshold;
    UINT32            RxCongested;
    UINT32            dwRxOffCount;
    UINT32            RxOverrun;
    UINT32            RxOverrunCount;
    UINT32            RxStatusDWReadCount;
    UINT32            RxDataDWReadCount;
    UINT32            RxPacketReadCount;
    UINT32            RxFastForwardCount;
    UINT32            RxPioReadCount;
    UINT32            RxDmaReadCount;
    UINT32            RxCongestedCount;
    UINT32            RxDumpCount;
    UINT32            LastReasonForReleasingCPU;
    UINT32            LastRxStatus1;
    UINT32            LastRxStatus2;
    UINT32            LastRxStatus3;
    UINT32            LastIntStatus1;
    UINT32            LastIntStatus2;
    UINT32            LastIntStatus3;
    UINT32            imask;
    UINT32            RxUnloadProgress;
    UINT32            RxUnloadPacketProgress;
    UINT32            RxMaxDataFifoSize;

	struct ifnet      *pIf;	
    SEM_ID            ifCounterLock;
    char              *pIfAddr;
	
	int               RxWorkLimit;
	int               RxPacketsReceived;

    UINT32            RxFlowCurrentThroughput;
    UINT32            RxFlowCurrentPacketCount;
    UINT32            RxFlowCurrentWorkLoad;
    UINT32            MeasuringRxThroughput;
    UINT32            RxFlowMeasuredMaxThroughput;
    UINT32            RxFlowMeasuredMaxPacketCount;

/*RX_FLOW_ACTIVATION specifies the percentage that RxFlowCurrentWorkLoad must exceed
     RxFlowMaxWorkLoad in order to activate flow control */
     
#define RX_FLOW_ACTIVATION	(4UL)

/*RX_FLOW_DEACTIVATION specifies the percentage that RxFlowCurrentWorkLoad must reduce
     from RxFlowMaxWorkLoad in order to deactivate flow control */
     
#define RX_FLOW_DEACTIVATION (25UL)


    UINT32            RxFlowMaxWorkLoad;

    FLOW_CONTROL_PARAMETERS RxFlowParameters;

    UINT32            RxFlowBurstWorkLoad;
    UINT32            RxFlowBurstMaxWorkLoad;
    UINT32            RxFlowControlActive;
    UINT32            RxFlowBurstActive;
    UINT32            RxInterrupts;
    BOOL              FCRxInterruptsEn;

#define GPT_SCHEDULE_DEPTH	(4)
    void              *GptFunction[GPT_SCHEDULE_DEPTH];
    UINT32            GptCallTime[GPT_SCHEDULE_DEPTH];
    UINT32            Gpt_scheduled_slot_index;
    SEM_ID            GpTimerLock;

    UINT32            Running;

	UINT32            dwPhyAddress;
    UINT32            dwPhyId;
#ifdef USE_LED1_WORK_AROUND
    UINT32            NotUsingExtPhy;
#endif
    UINT8             bPhyModel;
    UINT8             bPhyRev;
    UINT32            dwLinkSpeed;
    UINT32            dwLinkSettings;
    UINT32            dwRemoteFaultCount;
	/* struct timer_list LinkPollingTimer; */
    UINT32            StopLinkPolling;
    UINT16            wLastADV;
    UINT16            wLastADVatRestart;
#ifdef USE_PHY_WORK_AROUND
#define MIN_PACKET_SIZE (64)
    UINT32            dwTxStartMargen;
    UINT8             LoopBackTxPacket[MIN_PACKET_SIZE];
    UINT32            dwTxEndMargen;
    UINT32            dwRxStartMargen;
    UINT8             LoopBackRxPacket[MIN_PACKET_SIZE];
    UINT32            dwRxEndMargen;
    UINT32            dwResetCount;
#endif /*USE_PHY_WORK_AROUND*/

    UINT8             SoftwareInterruptSignal;

#define SMSC_IF_NAME_SIZE	(10)
	char              ifName[SMSC_IF_NAME_SIZE];

/* for Rx Multicast work around */
    volatile UINT32   HashLo;
    volatile UINT32   HashHi;
    volatile UINT32   MulticastUpdatePending;
    volatile UINT32   set_bits_mask;
    volatile UINT32   clear_bits_mask;

    UINT32            rx_dma;
    UINT32            tx_dma;
    UINT32            afc_cfg;
    UINT32            dma_threshold;
    UINT32            max_throughput;
    UINT32            max_packet_count;
    UINT32            packet_cost;
    UINT32            burst_period;
    UINT32            max_work_load;
    UINT32            tx_fif_sz;
    UINT32            int_deas;

	SMSC9118_IOCTL_DATA ioctlData;

    } LAN9118_END_DEV, *PLAN9118_END_DEV;

/* network buffers configuration */

M_CL_CONFIG LAN9118MclBlkConfig = 	/* network mbuf configuration table */
    {
    /* 
    no. mBlks	no. clBlks	memArea		memSize
    ---------	----------	-------		-------
    */
    0, 		0, 		NULL, 		0
    };

CL_DESC LAN9118ClDescTbl [] =		/* cluster pool configuration table */
    {
    /* 
    clSize	num	memArea		memSize
    ------	----	-------		-------
    */
    {0,		0,	NULL,		0}
    }; 


NET_POOL   LAN9118NetPool;	/* global for easy access of end.pNetPool */
void *     LAN9118EndDevice;	/* global for easy access of the device */
CL_POOL_ID LAN9118ClPoolId;	/* global for easy access to cluster pool */
SEM_ID     LAN9118DmaSemId;	/* global for easy access to sem ID */



int LAN9118ClDescTblNumEnt = (NELEMENTS(LAN9118ClDescTbl));


/* globals */

#ifdef	LAN9118_INSTRUMENT
UINT32  LAN9118TxTimeout = 0;   /* number of transmit time-out */
UINT32  LAN9118TxError = 0;     /* number of transmit errors */
UINT32	LAN9118RestartNb = 0;	/* number of restart due to ring overflow */
UINT32	LAN9118InitNb = 0;		/* number of time device is re-initialized */
UINT32	LAN9118TxNb = 0;		/* number of transmitted packets */
UINT32	LAN9118IntNb = 0;		/* number of receive interrupt */
UINT32	LAN9118Len = 0;			/* lenght of the current received packet */
UINT32	LAN9118HdrStat = 0;		/* status byte of the current received packet */
UINT32	LAN9118NextPage = 0;	/* page pointer to the next received packet */
UINT32	LAN9118CurrentPage = 0;	/* start page of the current packet */
#endif




/* platform retated decelaration */
LOCAL void   Platform_SetRegDW (UINT32 dwLanBase,UINT32 dwOffset,UINT32 dwVal);
LOCAL UINT32 Platform_GetRegDW (UINT32 dwLanBase,UINT32 dwOffset);
LOCAL void   Platform_Initialize (PLAN9118_END_DEV pDrvCtrl);
LOCAL UINT32 Platform_CurrentIRQ (PLAN9118_END_DEV pDrvCtrl);
LOCAL void   Platform_CleanUp (PLAN9118_END_DEV pDrvCtrl);
LOCAL BOOL   Platform_Is16BitMode (PLAN9118_END_DEV pDrvCtrl);
LOCAL void   Platform_FreeIRQ (PLAN9118_END_DEV pDrvCtrl);
LOCAL BOOL   Platform_IsValidDmaChannel (UINT32 dwDmaCh);
LOCAL BOOL   Platform_DmaInitialize (PLAN9118_END_DEV pDrvCtrl,	UINT32 dwDmaCh);
LOCAL BOOL   Platform_DmaDisable (PLAN9118_END_DEV pDrvCtrl, const UINT32 dwDmaCh);
LOCAL void   Platform_CacheInvalidate (PLAN9118_END_DEV pDrvCtrl, const void * const pStartAddress,	const UINT32 dwLengthInBytes);
LOCAL void   Platform_CachePurge (PLAN9118_END_DEV pDrvCtrl, const void * const pStartAddress, const UINT32 dwLengthInBytes);
LOCAL UINT32 Platform_RequestDmaChannel (PLAN9118_END_DEV pDrvCtrl,BOOL Dir);
LOCAL void   Platform_ReleaseDmaChannel (PLAN9118_END_DEV pDrvCtrl, BOOL Dir);
LOCAL BOOL   Platform_DmaStartXfer (PLAN9118_END_DEV pDrvCtrl, const DMA_XFER * const pDmaXfer);
LOCAL UINT32 Platform_DmaGetDwCnt (PLAN9118_END_DEV pDrvCtrl, const UINT32 dwDmaCh);
LOCAL void   Platform_DmaComplete (PLAN9118_END_DEV pDrvCtrl, const UINT32 dwDmaCh);
LOCAL void   Platform_GetFlowControlParameters (PLAN9118_END_DEV pDrvCtrl, PFLOW_CONTROL_PARAMETERS flowControlParameters, BOOL useDma);
LOCAL void   Platform_WriteFifo (UINT32 dwLanBase, UINT32 *pdwBuf, UINT32 dwDwordCount);
LOCAL void   Platform_ReadFifo (UINT32 dwLanBase, UINT32 *pdwBuf, UINT32 dwDwordCount);

/* LAN MAC Functions */
LOCAL STATUS Mac_Initialize (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS MacNotBusy (PLAN9118_END_DEV pDrvCtrl);
LOCAL UINT32 Mac_GetRegDW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegOffset,VL_KEY keyCode);
LOCAL void Mac_SetRegDW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegOffset, UINT32 dwVal,VL_KEY keyCode);

/* LAN PHY Functions */
LOCAL STATUS Phy_Initialize (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Phy_SetLink (PLAN9118_END_DEV pDrvCtrl);
LOCAL UINT16 Phy_GetRegW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegIndex,VL_KEY keyCode);
LOCAL void Phy_SetRegW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegIndex, UINT16 wVal,VL_KEY keyCode);
LOCAL void Phy_UpdateLinkMode (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Phy_GetLinkMode (PLAN9118_END_DEV pDrvCtrl,VL_KEY keyCode);

void Phy_CheckLink(unsigned long ptr);
LOCAL UINT16 Phy_GetAutoMdixSts(PLAN9118_END_DEV pDrvCtrl);
LOCAL void Phy_SetAutoMdixSts(PLAN9118_END_DEV pDrvCtrl, UINT16 wAutoMdixSts);

/* GP Timer Functions */
LOCAL TIME_SPAN Gpt_FreeRunCompare (UINT32 time1, UINT32 time2);
LOCAL void Gpt_ScheduleInterrupt (PLAN9118_END_DEV pDrvCtrl, TIME_SPAN timeSpan);
LOCAL void Gpt_CancelInterrupt (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Gpt_CancelCallBack (PLAN9118_END_DEV pDrvCtrl, void (*callBackFunction) (PLAN9118_END_DEV pDrvCtrl));
/* 100uS units relative to now */
LOCAL void Gpt_ScheduleCallBack (PLAN9118_END_DEV pDrvCtrl, void (*callBackFunction) (PLAN9118_END_DEV pDrvCtrl), UINT32 callBackTime);
LOCAL void Gpt_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);

void GptCB_RxCompleteMulticast (PLAN9118_END_DEV pDrvCtrl);
void GptCB_RestartBurst (PLAN9118_END_DEV pDrvCtrl);
void GptCB_MeasureRxThroughput (PLAN9118_END_DEV pDrvCtrl);

/* Tx/Rx Functions */
LOCAL void Tx_Initialize (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Tx_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl,UINT32 dwIntSts);

#ifdef SCATTER_GATHER_EN
LOCAL void Tx_MblkWalk (M_BLK * pMblk, int *pFragNum, int *pPktLen);
LOCAL STATUS Tx_SendFragBuf (PLAN9118_END_DEV pDrvCtrl, M_BLK* pMblk);
#else
LOCAL STATUS Tx_SendBuf (PLAN9118_END_DEV pDrvCtrl, M_BLK* pMblk);
#endif

LOCAL void Tx_StopQueue (PLAN9118_END_DEV pDrvCtrl,UINT32 dwSource);
LOCAL void Tx_WakeQueue (PLAN9118_END_DEV pDrvCtrl,UINT32 dwSource);
LOCAL UINT32 Tx_GetTxStatusCount (PLAN9118_END_DEV pDrvCtrl);
LOCAL UINT32 Tx_CompleteTx (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Tx_UpdateTxCounters (PLAN9118_END_DEV pDrvCtrl);
LOCAL void CalculateTxChecksumOffset (PLAN9118_END_DEV pDrvCtrl, UINT16 *csum_start_offset);

LOCAL void Rx_Initialize (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Rx_CompleteMulticastUpdate (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Rx_HandleOverrun (PLAN9118_END_DEV pDrvCtrl);
LOCAL UINT32 Rx_PopRxStatus (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL void Rx_CountErrors (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRxStatus);
LOCAL void Rx_FastForward (PLAN9118_END_DEV pDrvCtrl, UINT32 dwDwordCount);
LOCAL void Rx_ProcessPackets(PLAN9118_END_DEV pDrvCtrl);
LOCAL void Rx_BeginMulticastUpdate (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Rx_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL void RxStop_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL void Rx_ReceiverOff (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Rx_ReceiverOn (PLAN9118_END_DEV pDrvCtrl, VL_KEY callerKeyCode);

/* EEPROM Functions */
void Eeprom_EnableAccess(PLAN9118_END_DEV pDrvCtrl);
void Eeprom_DisableAccess(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_IsMacAddressLoaded(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_IsBusy(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_Timeout(PLAN9118_END_DEV pDrvCtrl);

BOOL Eeprom_ReadLocation(PLAN9118_END_DEV pDrvCtrl,BYTE address, BYTE * data);
BOOL Eeprom_EnableEraseAndWrite(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_DisableEraseAndWrite(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_WriteLocation(PLAN9118_END_DEV pDrvCtrl,BYTE address,BYTE data);
BOOL Eeprom_EraseAll(PLAN9118_END_DEV pDrvCtrl);
BOOL Eeprom_Reload(PLAN9118_END_DEV pDrvCtrl);

BOOL Eeprom_SaveMacAddress(PLAN9118_END_DEV pDrvCtrl,UINT32 dwHi16,UINT32 dwLo32);


/* LAN Helper Functions */
LOCAL STATUS Lan_Initialize (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntCfg);
LOCAL void Lan_EnableInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntEnMask);
LOCAL void Lan_DisableInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntEnMask);
LOCAL void Lan_EnableIRQ (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Lan_DisableIRQ (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Lan_SetIntDeas (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntDeas);
LOCAL void Lan_SetTDFL (PLAN9118_END_DEV pDrvCtrl, UINT8 level);
LOCAL void Lan_SetTSFL (PLAN9118_END_DEV pDrvCtrl, UINT8 level);
LOCAL void Lan_SetRSFL (PLAN9118_END_DEV pDrvCtrl, UINT8 level);

LOCAL void Lan_SignalSoftwareInterrupt (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Lan_HandleSoftwareInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL void Lan_udelay (PLAN9118_END_DEV pDrvCtrl, UINT32 uSec);

LOCAL void Lan_ShowMacRegs (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Lan_ShowSysRegs (PLAN9118_END_DEV pDrvCtrl);
LOCAL void Lan_ShowPhyRegs (PLAN9118_END_DEV pDrvCtrl);

LOCAL void Led_TestPoint(PLAN9118_END_DEV pDrvCtrl, UINT8 TestPointNum);
LOCAL void Lan_SetGPIO (PLAN9118_END_DEV pDrvCtrl, UINT8 GpiosToSet);
LOCAL void Lan_ClrGPIO (PLAN9118_END_DEV pDrvCtrl, UINT8 GpiosToClr);

LOCAL int Lan_GetDebug (void);
#ifdef	LAN9118_DEBUG
LOCAL void DumpBuffer(UINT8 *data, UINT32 len);
#endif

LOCAL STATUS    Lan_AddrFilterSet (PLAN9118_END_DEV pDrvCtrl, char* pAddr, BOOL bSet);
LOCAL int       Lan_HashIndex (char* eAddr);
LOCAL void      Lan_HandleRxInt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL STATUS    Lan_ReadFrameBufferSupplied (PLAN9118_END_DEV pDrvCtrl, M_BLK* pMblk);
LOCAL M_BLK*    Lan_ReadFrameBufferNotSupplied (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts);
LOCAL void      Lan_EnetAddrGet (PLAN9118_END_DEV pDrvCtrl, char* addr);

LOCAL void      LAN9118_DrvDetails (void);

/* END Specific interfaces. */
END_OBJ *		LAN9118EndLoad (char *initString);
LOCAL STATUS	LAN9118Unload (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS	LAN9118Start (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS	LAN9118Stop (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS	LAN9118Send (PLAN9118_END_DEV pDrvCtrl, M_BLK *pMblk);
LOCAL int		LAN9118Ioctl (PLAN9118_END_DEV pDrvCtrl, int cmd, caddr_t data);

LOCAL STATUS	LAN9118MCastAddrAdd (PLAN9118_END_DEV pDrvCtrl, char* pAddress);
LOCAL STATUS	LAN9118MCastAddrDel (PLAN9118_END_DEV pDrvCtrl, char* pAddress);
LOCAL STATUS	LAN9118MCastAddrGet (PLAN9118_END_DEV pDrvCtrl, MULTI_TABLE *pTable);

LOCAL STATUS	LAN9118PollStart (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS	LAN9118PollStop (PLAN9118_END_DEV pDrvCtrl);
LOCAL STATUS	LAN9118PollSend (PLAN9118_END_DEV pDrvCtrl, M_BLK *pMblk);
LOCAL STATUS	LAN9118PollReceive (PLAN9118_END_DEV pDrvCtrl, M_BLK *pMblk);

LOCAL STATUS    LAN9118InitParse (PLAN9118_END_DEV pDrvCtrl, char * initString);
LOCAL STATUS    LAN9118MemInit (PLAN9118_END_DEV pDrvCtrl);
LOCAL void      LAN9118Isr (PLAN9118_END_DEV pDrvCtrl);

void cmd (char *str, UINT32 arg1, UINT32 arg2, UINT32 arg3);

/***************************************************************************************
 * Declare our function table.  This is static across all driver
 * instances.
 */
LOCAL NET_FUNCS LAN9118FuncTable =
    {
    (FUNCPTR)LAN9118Start,			/* Function to start the device. */
    (FUNCPTR)LAN9118Stop,			/* Function to stop the device. */
    (FUNCPTR)LAN9118Unload,		/* Unloading function */
    (FUNCPTR)LAN9118Ioctl,			/* Ioctl function */
    (FUNCPTR)LAN9118Send,			/* Send function */
    (FUNCPTR)LAN9118MCastAddrAdd,	/* Multicast address add */
    (FUNCPTR)LAN9118MCastAddrDel,	/* Multicast address delete */
    (FUNCPTR)LAN9118MCastAddrGet,	/* Multicast table retrieve */
    (FUNCPTR)LAN9118PollSend,		/* Polling send function  */
    (FUNCPTR)LAN9118PollReceive,	/* Polling receive function */
    endEtherAddressForm,			/* Put address info into a packet.  */
    endEtherPacketDataGet,			/* Get a pointer to packet data. */
    endEtherPacketAddrGet,			/* Get packet addresses. */
    NULL							/* Bind function */
    };



volatile UINT32 g_GpioSetting = 0x0; /* fixme - better be gaurded by semaphore */
#ifdef USE_LED1_WORK_AROUND
volatile UINT32 g_GpioSettingOriginal = 0x0; /* fixme - better be gaurded by semaphore */
#endif

PLAN9118_END_DEV saved_pDrvCtrl;
/************************************************************************/
/*	Set Bus wait states for specific memory area			    		*/
/************************************************************************/
LOCAL const UINT16 wWaitArea[7] = {AREA0_WAIT_CTRL_,
									0,
									AREA2_WAIT_CTRL_,
								 	AREA3_WAIT_CTRL_, 
									AREA4_WAIT_CTRL_,
									AREA5_WAIT_CTRL_,
									AREA6_WAIT_CTRL_
									};

LOCAL const UINT16 wWaitState[11] = {
									WAIT_STATES_0,
									WAIT_STATES_1,
									WAIT_STATES_2,
								 	WAIT_STATES_3,
									WAIT_STATES_4,
									0,
									WAIT_STATES_6,
									0,
									WAIT_STATES_8,
									0,
									WAIT_STATES_10};



/*********************************************************************************/
/*platform related functions
/**********************************************************************************/



/************************************************************

FUNCTION: Platform_SetRegDW

PARAMETERS: 

	UINT32 dwLanBase, the virtual base address returned by 

			Platform_Initialize

	UINT32 dwOffset, the byte offset into the Lan address space

			where the register of interest is located.

			dwOffset should be UINT32 aligned.

	UINT32 dwVal, the value to write to the register of interest

DESCRIPTION:

	This function is used to provide a platform independent

		method of register access. It allows platform code 

		for big endian systems to do byte swapping, if necessary.

	Because this function is very simple, it is recommended that

		it should be implemented as inline or as a macro.

RETURN VALUE: void

************************************************************/
LOCAL void Platform_SetRegDW(
		UINT32 dwLanBase,
		UINT32 dwOffset,
		UINT32 dwVal)
{
	(*(volatile UINT32 *)(dwLanBase+dwOffset))=dwVal;
}

/************************************************************

FUNCTION: Platform_GetRegDW

PARAMETERS:

	UINT32 dwLanBase, the virtual base address returned by

			Platform_Initialize

	UINT32 dwOffset, the byte offset into the Lan address space

			where the register of interest is located.

			dwOffset should be UINT32 aligned.

DESCRIPTION:

	This function is used to provide a platform independent

		method of register access. It allows platform code 

		for big endian systems to do byte swapping if necessary.

	Because this function is very simple, it is recommended that

		it should be implemented as inline or as a macro.

RETURN VALUE:

	The UINT32 value read from the register of interest;

************************************************************/
LOCAL UINT32 Platform_GetRegDW(
	UINT32 dwLanBase,
	UINT32 dwOffset)
{
	return (*(volatile UINT32 *)(dwLanBase+dwOffset));
}

/************************************************************

FUNCTION: Platform_Initialize

PARAMETERS:

	platformData, pointer to platform specified data structure

	dwLanBase,	user specified physical base address for the LAN9118

		= 0, let this function decide

	dwBusWidth,     

		= 16, user specifies 16 bit mode operation, bypass autodetection

		= 32, user specifies 32 bit mode operation, bypass autodetection

		= any other value, driver should auto detect bus width

DESCRIPTION:

    This is the first Platform_xxx function that will be called

	This function will initialize the PLATFORM_DATA structure.

	This function will prepare the system to access the LAN9118.

	It will properly initialize the bus and return the virtual 

	memory location where the LAN9118 can be access.

RETURN VALUE:

	0 = failed,

	any other value is the virtual base address where the LAN9118 

	   can be accessed 

************************************************************/
LOCAL void Platform_Initialize (PLAN9118_END_DEV pDrvCtrl)
{
	int 	iAreaNum=0;
	/* initially set to slowest setting incase the chip is 115/116 */
	int 	iWaitStNum=10;
	UINT32	dwLanBase;

	UINT16 wWaitCtrlReg, wTemp;
	UINT16 wBcrMask=1;

#ifdef VXWORKS_DEBUG 
   	DRV_LOG (LAN9118_DEBUG_LOAD, "Platform_Initialize()\n", 1, 2, 3, 4, 5, 6);
#endif
	SMSC_ASSERT(pDrvCtrl!=NULL);
	pDrvCtrl->dwBitWidth=0;
	pDrvCtrl->dwIdRev=0;
	dwLanBase = pDrvCtrl->dwLanBase;
	if(dwLanBase==0x0) {
		dwLanBase=LAN9118_CSBASE;
	}

	if(0xA8000000<=dwLanBase) 
	{
		iAreaNum=2;
	}
	if(0xAC000000<=dwLanBase)
	{
		iAreaNum=3;
	}
	if(0xB0000000<=dwLanBase)
	{
		iAreaNum=4;
	}
	if(0xB4000000<=dwLanBase)
	{
		iAreaNum=5;
	}
	if(0xB8000000<=dwLanBase)
	{
		iAreaNum=6;
	}
	if(iAreaNum==0) {
#ifdef VXWORKS_DEBUG
       	DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Invalid Lan Base Address = 0x%08X\n", dwLanBase, 2, 3, 4, 5, 6);
		DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "  must be in Area 2,3,4,5, or 6 and uncached\n", 1, 2, 3, 4, 5, 6);
#endif 
		pDrvCtrl->dwLanBase = 0;
		return;
	}

#ifdef VXWORKS_DEBUG	
	DRV_LOG (LAN9118_DEBUG_LOAD, " Lan Base at 0x%08lX in Area %d\n", dwLanBase, iAreaNum, 3, 4, 5, 6);
#endif
	/* Init BSC registers for 11x */
	wWaitCtrlReg = WCR2;
	
	/* Clear specified Area */
	wWaitCtrlReg &= (UINT16)(~(wWaitArea[iAreaNum]));	
	wTemp = wWaitState[iWaitStNum];

	switch (iAreaNum)
	{
		case AREA_2:
			wTemp <<= 3;
			wBcrMask <<=4;
			break;
	
		case AREA_3:
			wTemp <<= 5;
			wBcrMask <<=6;
			break;
	
		case AREA_4:
			wTemp <<= 7;
			wBcrMask <<=8;
			break;
	
		case AREA_5:
			wTemp <<= 10;
			wBcrMask <<=10;
			break;
	
		case AREA_6:
			wTemp <<= 13;
			wBcrMask <<= 12;
			break;
	
		default:
			break;
	}
	
	/* Inserted 2 wait states for Area 5 and 2. */
	wWaitCtrlReg |= wTemp;	
	WCR2 = wWaitCtrlReg;
	wTemp=BCR2;
	{
		UINT32 test=0;
		UINT16 dummy=0;
		BCR2|=wBcrMask;
#define BYTE_TEST_OFFSET	(0x64)
		test=(*(volatile unsigned long *)(dwLanBase+BYTE_TEST_OFFSET));
		dummy=(*(volatile unsigned short *)(dwLanBase+BYTE_TEST_OFFSET+2));
		if(test==0x87654321) {
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_LOAD, " 32 bit mode detected\n", 1, 2, 3, 4, 5, 6);
#endif
			pDrvCtrl->dwBitWidth=32;
		} else {
			BCR2&=(~(wBcrMask));
			test=(*(volatile unsigned long *)(dwLanBase+BYTE_TEST_OFFSET));
			if(test==0x87654321) {
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_LOAD, " 16 bit mode detected\n", 1, 2, 3, 4, 5, 6);
#endif
				pDrvCtrl->dwBitWidth=16;
			} else {
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, " neither 16 nor 32 bit mode detected, test=0x%08x\n", test, 2, 3, 4, 5, 6);
#endif
				BCR2=wTemp;
				dwLanBase=0;
				dummy=dummy;
			}
		}
	}

	if(dwLanBase!=0)
	{
		UINT32 idRev=0;
	#define ID_REV_OFFSET	(0x50)
		idRev=(*(volatile unsigned long *)(dwLanBase+ID_REV_OFFSET));
		pDrvCtrl->dwIdRev=idRev;
		
		switch(idRev&0xFFFF0000) {
		case	0x01150000UL:
		case	0x01160000UL:
		case	0x115A0000UL:
		case	0x215A0000UL:
		case	0x216A0000UL:
		case	0x92100000UL:
		case	0x92200000UL:
			break;
			
		default:		
			iWaitStNum=2; 
			wWaitCtrlReg = WCR2;
			wWaitCtrlReg &= (UINT16)(~(wWaitArea[iAreaNum]));	/* Clear specified Area */
			wTemp = wWaitState[iWaitStNum];

			switch (iAreaNum)
			{
				case AREA_2:
					wTemp <<= 3;
					wBcrMask <<=4;
					break;
	
				case AREA_3:
					wTemp <<= 5;
					wBcrMask <<=6;
					break;
		
				case AREA_4:
					wTemp <<= 7;
					wBcrMask <<=8;
					break;
	
				case AREA_5:
					wTemp <<= 10;
					wBcrMask <<=10;
					break;
	
				case AREA_6:
					wTemp <<= 13;
					wBcrMask <<= 12;
					break;
	
				default:break;	
			}
			wWaitCtrlReg |= wTemp;	/* Inserted 2 wait states for Area 5 and 2. */
			WCR2 = wWaitCtrlReg;
			break;	
		}
	}
	pDrvCtrl->dwLanBase = dwLanBase;
	
	/* Clear All Leds */
	sysLedSet(0);
}

/*********************************************************************************
Function Name:  Platform_CurrentIRQ

Description : this function returns the current IRQ used by the device.

input: 		PLAN9118_END_DEV

output:  UINT32 

**********************************************************************************/
LOCAL UINT32 Platform_CurrentIRQ (PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl != NULL);
	return pDrvCtrl->ivec;
}

/***********************************************************

FUNCTION: Platform_CleanUp

PARAMETERS:

	platformData, pointer to platform specified data structure

DESCRIPTION:

	This function is called so the platform layer can clean up

	any resources that may have been acquired in Platform_Initialize.

	If Platform_Initialize returned 0, then this function will

	not be called.

***********************************************************/
LOCAL void Platform_CleanUp (PLAN9118_END_DEV pDrvCtrl)
{
}

/***********************************************************

FUNCTION: Platform_Is16BitMode

PARAMETER:

	platformData, pointer to platform specific data structure

RETURN VALUE:

	TRUE if the platform is configured for 16 bit mode

	FALSE if the platform is configured for 32 bit mode

NOTE: this function is not currently used. 

***********************************************************/
LOCAL BOOL Platform_Is16BitMode (PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(pDrvCtrl->dwBitWidth==16) {
		return TRUE;
	}
	return FALSE;
}

/*****************************************************

FUNCTION: Platform_GetFlowControlParameters

PARAMETERS:

	platformData, pointer to platform specific data structure

	flowControlParameters, pointer to structure which receives 

	     the flow control parameters.

    useDma, flag which specifies whether to get flow control

         parameters for PIO(FALSE) or DMA(TRUE). 

DESCRIPTION:

	fills the structure pointed to by flowControlParameters,

	with the appropriate flow control parameters for the 

	current combination of ((118/7/6/5),(218/7/5) or (116/115)), 

	(16 or 32 bit mode), and (PIO or DMA).

******************************************************/
LOCAL void Platform_GetFlowControlParameters(
	PLAN9118_END_DEV pDrvCtrl,
	PFLOW_CONTROL_PARAMETERS flowControlParameters,
	BOOL useDma)
{
	memset(flowControlParameters,0,sizeof(FLOW_CONTROL_PARAMETERS));
	flowControlParameters->BurstPeriod=100;
	flowControlParameters->IntDeas=0;
	if(useDma) {
		if(Platform_Is16BitMode(pDrvCtrl)) {
			switch(pDrvCtrl->dwIdRev&0xFFFF0000) {

			case 0x01160000UL:
			case 0x01150000UL:
			case 0x115A0000UL:
			case 0x215A0000UL:
			case 0x216A0000UL:
			case 0x92100000UL:
			case 0x92200000UL:
				/*115/116,16 bit,DMA*/
				flowControlParameters->MaxThroughput=0x89DF2UL;
				flowControlParameters->MaxPacketCount=0x175UL;
				flowControlParameters->PacketCost=0x167UL;
				flowControlParameters->BurstPeriod=0x3EUL;
				flowControlParameters->IntDeas=0x05UL;
				break;
			default:
				flowControlParameters->MaxThroughput=(0xABE0AUL);
				flowControlParameters->MaxPacketCount=(0x1D1UL);
				flowControlParameters->PacketCost=(0x1CEUL);
				flowControlParameters->BurstPeriod=(0x51UL);
				flowControlParameters->IntDeas=(0x09UL);
				break;
			}

		} else {

			switch(pDrvCtrl->dwIdRev&0xFFFF0000) {
			case 0x01160000UL:
			case 0x01150000UL:
			case 0x115A0000UL:
			case 0x215A0000UL:
			case 0x216A0000UL:
			case 0x92100000UL:
			case 0x92200000UL:
				/*115/116,32 bit,DMA*/
				flowControlParameters->MaxThroughput=0xABE0AUL;
				flowControlParameters->MaxPacketCount=0x1D1UL;
				flowControlParameters->PacketCost=0x00UL;
				flowControlParameters->BurstPeriod=0x30UL;
				flowControlParameters->IntDeas=0x0A;
				break;

			default:
				flowControlParameters->MaxThroughput=(0xC7F82UL);
				flowControlParameters->MaxPacketCount=(0x21DUL);
				flowControlParameters->PacketCost=(0x17UL);
				flowControlParameters->BurstPeriod=(0x1EUL);
				flowControlParameters->IntDeas=(0x17UL);
				break;

			}

		}

	} else {

		if(Platform_Is16BitMode(pDrvCtrl)) {

			switch(pDrvCtrl->dwIdRev&0xFFFF0000) {

			case 0x01160000UL:
			case 0x01150000UL:
			case 0x115A0000UL:
			case 0x215A0000UL:
			case 0x216A0000UL:
			case 0x92100000UL:
			case 0x92200000UL:
				/*115/116,16 bit,PIO*/

				flowControlParameters->MaxThroughput=(0x787FCUL);
				flowControlParameters->MaxPacketCount=(0x146UL);
				flowControlParameters->PacketCost=(0x112UL);
				flowControlParameters->BurstPeriod=(0x40UL);
				flowControlParameters->IntDeas=(0x36UL);
				break;
			default:
				flowControlParameters->MaxThroughput=(0x9F4F6UL);
				flowControlParameters->MaxPacketCount=(0x1AFUL);
				flowControlParameters->PacketCost=(0x181UL);
				flowControlParameters->BurstPeriod=(0x32UL);
				flowControlParameters->IntDeas=(0x45UL);
				break;
			}

		} else {

			switch(pDrvCtrl->dwIdRev&0xFFFF0000) {

			case 0x01160000UL:
			case 0x01150000UL:
			case 0x115A0000UL:
			case 0x215A0000UL:
			case 0x216A0000UL:
			case 0x92100000UL:
			case 0x92200000UL:
				/*115/116,32 bit,PIO*/

				flowControlParameters->MaxThroughput=(0x9E338UL);
				flowControlParameters->MaxPacketCount=(0x1ACUL);
				flowControlParameters->PacketCost=(0xD2UL);
				flowControlParameters->BurstPeriod=(0x60UL);
				flowControlParameters->IntDeas=(0x0EUL);
				break;

			default:
				flowControlParameters->MaxThroughput=(0xAE5C8UL);
				flowControlParameters->MaxPacketCount=(0x1D8UL);
				flowControlParameters->PacketCost=(0UL);
				flowControlParameters->BurstPeriod=(0x57UL);
				flowControlParameters->IntDeas=(0x14UL);
				break;


			}

		}

	}

}

/**********************************************************

FUNCTION: Platform_FreeIRQ

PARAMETERS:

    platformData, pointer to platform specific data structure

DESCRIPTION:

    uninstalls the ISR installed from Platform_RequestIRQ

**********************************************************/
LOCAL void Platform_FreeIRQ (PLAN9118_END_DEV pDrvCtrl)
{
	STATUS result;

	if(pDrvCtrl==NULL) {
		SMSC_ASSERT(FALSE);
		return;
	}
	
    SYS_INT_DISCONNECT (pDrvCtrl, LAN9118Isr, (int)pDrvCtrl, &result);

    if (result == ERROR)
	{
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_TMP, "Could not diconnect interrupt!\n",
                 1, 2, 3, 4, 5, 6);
#endif
	}

    pDrvCtrl->ivec = 0;
}

/*********************************************************

FUNCTION: Platform_IsValidDmaChannel

PARAMETERS:

	dwDmaCh, the dma channel number to test for validity

DESCRIPTION:

    This function is used to test the validity of the 

    channels request with parameters rx_dma, and tx_dma

RETURN VALUE:

	TRUE if the dma channel may be used

	FALSE if the dma channel may not be used

NOTE: this function does not use PLATFORM_DATA because

    it is called before the driver allocates memory for

    PLATFORM_DATA

*********************************************************/
LOCAL BOOL Platform_IsValidDmaChannel(UINT32 dwDmaCh)
{
	if(dwDmaCh<=3) {
		return TRUE;
	}
	return FALSE;
}

/*********************************************************************************
Function Name: GetRegDW & SetRegDW macros

Description : Memory access.  

input: 	Base Address and Offset	

output: void in SetRegDW, memory value in GetRegDW

**********************************************************************************/
#define GetRegDW(dwBase,dwOffset) \
	((UINT32)(*(volatile UINT32 *)(dwBase + dwOffset)))

#define SetRegDW(dwBase,dwOffset,dwVal) \
	((*(volatile UINT32 *)(dwBase + dwOffset)) = dwVal)

/********************************************************

FUNCTION: Platform_DmaDisable

PARAMETERS:

	platformData, pointer to the platform specific data structure

	dwDmaCh, the Dma channel to disable

RETURN VALUE:

	TRUE on success

	FALSE on failure

********************************************************/
LOCAL BOOL Platform_DmaDisable(
	PLAN9118_END_DEV pDrvCtrl,
	const UINT32 dwDmaCh)
{
	const UINT32	dwDmaCh0CtrlReg = 0xA400002CUL;
	UINT32		dwTemp;

	if (dwDmaCh > 3UL)
	{
		SMSC_WARNING("Platform_DmaDisable(), Invalid DMA Channel!\n");
		return FALSE;
	}
	dwTemp = GetRegDW(dwDmaCh0CtrlReg,(dwDmaCh * 0x10UL)); 
	dwTemp &= ~DMAC_ENBL_;	/*  Disable DMA controller */
	SetRegDW(dwDmaCh0CtrlReg, (dwDmaCh * 0x10UL), dwTemp); 
	return TRUE;
}

/*******************************************************

FUNCTION: Platform_CacheInvalidate

PARAMETERS:

	platformData, pointer to the platform specific data structure

	pStartAddress, the starting virtual address of the region

	dwLengthInBytes, the length in bytes of the region to invalidate

DESCRIPTION:

	Invalidates a specified memory region if it exists in cache.

	  Does not necessarily write the data back to memory.

*******************************************************/
LOCAL void Platform_CacheInvalidate(
	PLAN9118_END_DEV pDrvCtrl,
	const void * const pStartAddress,
	const UINT32 dwLengthInBytes)
{
	Platform_CachePurge(
		pDrvCtrl,
		pStartAddress,
		dwLengthInBytes);
}

/*******************************************************

FUNCTION: Platform_CachePurge

PARAMETERS:

    platformData, pointer to the platform specific data structure

	pStartAddress, the starting virtual address of the region

	dwLengthInBytes, the length in bytes of the region to purge

DESCRIPTION:

    Writes back data to a specified memory region if it 

      exists in cache.

*******************************************************/
LOCAL void Platform_CachePurge(
	PLAN9118_END_DEV pDrvCtrl,
	const void * const pStartAddress,
	const UINT32 dwLengthInBytes)
{
	UINT32	dwAddrIndex;
	UINT32	begin, end;

	const UINT32	dwStartAddr = CpuToPhysicalAddr(pStartAddress);

	begin = dwStartAddr & ~(PLATFORM_CACHE_LINE_BYTES - 1UL);
	end = (dwStartAddr + dwLengthInBytes + PLATFORM_CACHE_LINE_BYTES - 1UL) & 
				~(PLATFORM_CACHE_LINE_BYTES - 1UL);

	for (dwAddrIndex = begin; 
		dwAddrIndex < end; 
		dwAddrIndex += PLATFORM_CACHE_LINE_BYTES)
	{
		UINT32	dwData, dwAddr;

		dwData = (dwAddrIndex & CACHE_PHYSADDR_MASK);	 /* _Virtual_ address, ~U, ~V */
		dwAddr = CACHE_OC_ADDRESS_ARRAY | (dwAddrIndex & CACHE_OC_ENTRY_MASK) | 
				CACHE_OC_ASSOC_BIT_;
		(*(volatile unsigned long *)dwAddr) = dwData;
	}
}

/******************************************************

FUNCTION: Platform_RequestDmaChannel

PARAMETERS:

	platformData, pointer to the platform specific data structure

DESCRIPTION:

	If the OS supports Dma Channel allocation then this function

	  will use that support to reserve a dma channel.

	If the OS does not support Dma Channel allocation, or a

	  channel can not be reserved then this function 

	  will return TRANSFER_PIO

RETURN VALUE:

	returns the DMA channel number that has been reserved

	if a channel can not be reserved then return TRANSFER_PIO

******************************************************/
LOCAL UINT32 Platform_RequestDmaChannel(
	PLAN9118_END_DEV pDrvCtrl,
	BOOL Dir)
{
	if((Dir==TX_DMA)&&(pDrvCtrl->tx_dma<TRANSFER_PIO))
	{
	    return PLATFORM_TX_DMA;/* DMA Channel for TX_DMA*/
	}
	else if((Dir==RX_DMA)&&(pDrvCtrl->rx_dma<TRANSFER_PIO))
	{
	    return PLATFORM_RX_DMA;/* DMA Channel for RX_DMA*/
	}
	return TRANSFER_PIO;		
}

/*****************************************************

FUNCTION: Platform_ReleaseDmaChannel

PARAMETERS:

	platformData, pointer to the platform specific data structure

	dwDmaChannel, the Dma channel number to be released

DESCRIPTION:

	Releases the DMA channel specified by dwDmaChannel,

	  which was previously returned by Platform_RequestDmaChannel

    If the OS supports it this function will notify the OS

	  that the dma channel is free to be used by other devices

******************************************************/
LOCAL void Platform_ReleaseDmaChannel(
	PLAN9118_END_DEV pDrvCtrl,
	BOOL Dir)
{
	if((Dir==TX_DMA)&&(pDrvCtrl->tx_dma<TRANSFER_PIO))
	{
		pDrvCtrl->dwTxDmaCh=0;
		pDrvCtrl->TxDmaChReserved=FALSE;
	}
	else if((Dir==RX_DMA)&&(pDrvCtrl->rx_dma<TRANSFER_PIO))
	{
		pDrvCtrl->dwRxDmaCh=0;
	 	pDrvCtrl->RxDmaChReserved=FALSE;
	}
}

/******************************************************

FUNCTION: Platform_DmaComplete

PARAMETERS:

    platformData, pointer to platform specific data structure

	dwDmaCh, dma channel number

DESCRIPTION:

	Waits for the specified dma channel to finish 

	transfering data. Upon return the dma channel should

	be ready to handle another request from 

	Platform_DmaStartXfer

	This function should be the same as waiting for 

	Platform_DmaGetDwCnt to return 0

******************************************************/
LOCAL void Platform_DmaComplete(
	PLAN9118_END_DEV pDrvCtrl,
	const UINT32 dwDmaCh)
{
	UINT32 dwTimeOut=1000000;
	SMSC_ASSERT(dwDmaCh<4);

	while ((Platform_DmaGetDwCnt(pDrvCtrl,dwDmaCh)) && dwTimeOut) {
		sysFlashDelay(1);
		dwTimeOut--;
	}

	if(!Platform_DmaDisable(pDrvCtrl,dwDmaCh)) {
		SMSC_WARNING("Failed Platform_DmaDisable");
	}

	if(dwTimeOut==0) {
		SMSC_WARNING("Platform_DmaComplete: Timed out");
	}
}

/*******************************************************

FUNCTION: Platform_DmaGetDwCnt

PARAMETERS:

    platformData, pointer to the platform specific data structure

	dwDmaCh, Dma channel number

RETURN VALUE:

    0, if the DMA channel is ready to handle another 

	   request from Platform_DmaStartXfer

   non zero, is the number of UINT32S left for the 

       dma channel to transfer

*******************************************************/
LOCAL UINT32 Platform_DmaGetDwCnt(
	PLAN9118_END_DEV pDrvCtrl,
	const UINT32 dwDmaCh)
{
	const UINT32	dwDmaCh0XferCntReg = 0xA4000028UL;

	if (dwDmaCh > 3UL)
	{
		SMSC_WARNING("DmaGetDwCnt(), Invalid DMA Channel!\n");
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		return 0UL;
	}
	return (GetRegDW(dwDmaCh0XferCntReg,(dwDmaCh * 0x10UL)));
}

/********************************************************

FUNCTION: Platform_DmaInitialize

PARAMETERS:

    platformData, pointer to the platform specific data structure

	dwDmaCh, the Dma channel to initialize

RETURN VALUE:

	TRUE, on Success

	FALSE, on Failure

********************************************************/
LOCAL BOOL Platform_DmaInitialize(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwDmaCh)
{
	DMATCR0=0;
	DMATCR1=0;
	DMATCR2=0;
	DMATCR3=0;
	DMAOR|=DMA_MSTR_EN_;
	return TRUE;
}

/*****************************************************

FUNCTION: Platform_DmaStartXfer

PARAMETERS:

    platformData, pointer to the platform specific data structure

    pDmaXfer, pointer to DMA_XFER structure

	    which describes the requested transfer

DESCRIPTION:

    Begins a new dma transfer,

	Should not be called if another transfer is in progress

RETURN VALUE:

    TRUE if dma has begun the transfer

	FALSE for any error

******************************************************/
LOCAL BOOL Platform_DmaStartXfer(
	PLAN9118_END_DEV pDrvCtrl,
	const DMA_XFER * const pDmaXfer)
{
	const UINT32	dwDmaSrcBaseReg = (0xA4000020UL + (pDmaXfer->dwDmaCh * 0x10UL));
	const UINT32	dwDmaDstBaseReg = dwDmaSrcBaseReg + 0x04UL;
	const UINT32	dwDmaXferCntReg = dwDmaSrcBaseReg + 0x08UL;
	const UINT32	dwDmaCtrlReg    = dwDmaSrcBaseReg + 0x0CUL;

	UINT32 dwTemp, dwSrcAddr, dwDestAddr;
	UINT32 dwAlignMask;
	UINT32 dwLanPhysAddr, dwMemPhysAddr;

	/*  1. validate the requested channel # */
	if (pDmaXfer->dwDmaCh > 3UL)
	{
		SMSC_WARNING("Platform_DmaStartXfer -- bad dwDmaCh");
		return FALSE;
	}

	/*  2. make sure the channel's not already running */
	if (GetRegDW(dwDmaXferCntReg,0) != 0UL)
	{
		SMSC_WARNING("Platform_DmaStartXfer -- requested channel is still running");
		return FALSE;
	}

	/*  3. calculate the physical transfer addresses */
	dwLanPhysAddr = CpuToPhysicalAddr((void *)pDmaXfer->dwLanReg);
	dwMemPhysAddr = CpuToPhysicalAddr((void *)pDmaXfer->pdwBuf);

	/*  4. validate the address alignments */
	/*  need CL alignment for CL bursts */
	dwAlignMask = (PLATFORM_CACHE_LINE_BYTES - 1UL);
	
	if ((dwLanPhysAddr & dwAlignMask) != 0UL)
	{
		SMSC_WARNING("Platform_DmaStartXfer -- bad dwLanPhysAddr () alignment");
		return FALSE;
	}


	if ((dwMemPhysAddr & dwAlignMask) != 0UL)
	{
		SMSC_WARNING("Platform_DmaStartXfer -- bad dwMemPhysAddr () alignment");
		return FALSE;
	}

	/*  5. validate the transfer size & alignment */
	if (pDmaXfer->dwDwCnt >= (1UL<<24))
	{
		SMSC_WARNING("Platform_DmaStartXfer -- dwDwCnt is too big (16777216 counts max on SH3)");
		return FALSE;
	}

	if ((pDmaXfer->dwDwCnt & 3UL) != 0UL)
	{
		SMSC_WARNING("Platform_DmaStartXfer -- bad dwDwCnt alignment");
		return FALSE;
	}

	/*  6. Config Control reg */
	dwTemp = GetRegDW(dwDmaCtrlReg,0); 

	/*  Disable the selected channel first */
	dwTemp &= ~DMAC_ENBL_;
	SetRegDW((dwDmaCtrlReg),0, dwTemp); 

	/*  Select correct ch and set SRC, DST and counter */
	if (pDmaXfer->fMemWr == TRUE)
	{
		dwSrcAddr = dwLanPhysAddr;
		dwDestAddr = dwMemPhysAddr;

		dwTemp &= ~DST_ADDR_MODE_;
		dwTemp |= INC_DST_ADDR;		/*  Always to increment mem buffer */

		dwTemp &= ~SRC_ADDR_MODE_;
	}
	else
	{
		dwSrcAddr = dwMemPhysAddr;
		dwDestAddr = dwLanPhysAddr;

		dwTemp &= ~SRC_ADDR_MODE_;
		dwTemp |= INC_SRC_ADDR;		/*  Always increment mem buffer */

		dwTemp &= ~DST_ADDR_MODE_;
	}


	/*  Set Source and destination addresses */
	SetRegDW(dwDmaSrcBaseReg,0, dwSrcAddr);
	SetRegDW(dwDmaDstBaseReg,0, dwDestAddr); 

	/*  Set the transmit size in terms of the xfer mode */
	dwTemp &= ~XMIT_SIZE_;

#ifdef DMA_CYCLE_STEAL
	dwTemp &= ~XMIT_MODE_;	/* Clear to 0:  Cycle-steal mode */
#else	
	dwTemp |= XMIT_MODE_;	/* Set to 1:  Burst mode */
#endif	

	 
	dwTemp |= XMIT_QUINT32;	/*  Set 16 bytes (QUINT32) xfer */
	SetRegDW(dwDmaXferCntReg,0, (pDmaXfer->dwDwCnt/4UL));

	dwTemp &= ~XMIT_RESRC_;
	dwTemp |= AUTO_REQ;		/*  Auto request */
	dwTemp &= ~XFER_END_;
	SetRegDW(dwDmaCtrlReg,0, dwTemp); 


	/*  Clrea NMIF and AE bits and enable DMA master */
	DMAOR &= ~(NMI_FLAG_ | ADDR_ERR_FLAG_);

	/*  Enable DMA controller ch x */
	dwTemp = GetRegDW(dwDmaCtrlReg,0); 
	dwTemp |= DMAC_ENBL_;
	SetRegDW(dwDmaCtrlReg,0, dwTemp);

	return TRUE;
}

/******************************************************

FUNCTION: Platform_WriteFifo

PARAMETERS:

	UINT32 dwLanBase, the lan base address

	UINT32 * pdwBuf, a pointer to and array of UINT32s to 

		be written to the TX_DATA_FIFO

	UINT32 dwDwordCount, the number of UINT32s in the UINT32

		array pointed to by pdwBuf

DESCRIPTION:

	This function is during PIO transfers to write packet

	data to the TX_DATA_FIFO. This function was made

	platform dependent to allow platforms to perform

	byte swapping on big endian systems, if necessary.

******************************************************/
LOCAL void Platform_WriteFifo(
	UINT32 dwLanBase,
	UINT32 *pdwBuf,
	UINT32 dwDwordCount)
{
	volatile UINT32 *pdwReg;
	pdwReg = (volatile UINT32 *)(dwLanBase+TX_DATA_FIFO);
	while(dwDwordCount)
	{
		*pdwReg = *pdwBuf++;
		dwDwordCount--;
	}
}

/******************************************************

FUNCTION: Platform_ReadFifo

PARAMETERS:

	UINT32 dwLanBase, the lan base address

	UINT32 * pdwBuf, a pointer to and array of UINT32s which 

		will be read from the RX_DATA_FIFO

	UINT32 dwDwordCount, the number of UINT32s in the UINT32

		array pointed to by pdwBuf

DESCRIPTION:

	This function is during PIO transfers to read packet

	data from the RX_DATA_FIFO. This function was made

	platform dependent to allow platforms to perform

	byte swapping on big endian systems, if necessary.

******************************************************/
LOCAL void Platform_ReadFifo(
	UINT32 dwLanBase,
	UINT32 *pdwBuf,
	UINT32 dwDwordCount)
{
	const volatile UINT32 * const pdwReg = 
		(const volatile UINT32 * const)(dwLanBase+RX_DATA_FIFO);
	
	while (dwDwordCount)
	{
		*pdwBuf++ = *pdwReg;
		dwDwordCount--;
	}
}

/******************************************************************************
*
* LAN9118EndLoad - initialize the driver and device
*
* This routine initializes the driver and the device to the operational state.
* All of the device-specific parameters are passed in <initString>.
* This routine can be called in two modes. If it is called with an empty but
* allocated string, it places the name of this device (that is, "ln") into 
* the <initString> and returns 0.
*
* If the string is allocated and not empty, the routine attempts to load
* the driver using the values specified in the string.
*
* RETURNS: An END object pointer, or NULL on error, or 0 and the name of the
* device if the <initString> was NULL.
*
* "<unit>:<baseAdrs>:<intVec>:<intLvl>:<bufSize>:<link_mode>:<macAddrString>"
*
*/
END_OBJ* LAN9118EndLoad
    (
    char* initString			/* string to be parsed */
    )
    {

    PLAN9118_END_DEV	pDrvCtrl;
    UINT32		local_MACCR;

#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "Loading LAN9118EndLoad...\n", 1, 2, 3, 4, 5, 6);
#endif

    if (initString == NULL) {
        DRV_LOG (LAN9118_DEBUG_LOAD, "LAN9118EndLoad: NULL initStr\r\n",0,0,0,0,0,0);
        return (NULL);
    }
    
    if (initString[0] == '\0') {
        bcopy((char *)LAN9118_DEV_NAME, initString, LAN9118_DEV_NAME_LEN);
        return (0);
    }
    
    /* allocate the device structure */
    pDrvCtrl = (LAN9118_END_DEV *)calloc (sizeof (LAN9118_END_DEV), 1);
    if (pDrvCtrl == NULL) {
	    goto errorExit;
	}
	
    memset (pDrvCtrl, 0, sizeof (LAN9118_END_DEV));

    /* parse the init string, filling in the device structure */
    if (LAN9118InitParse (pDrvCtrl, initString) == ERROR) {
	    goto errorExit;
	}
    
    /* initialize the END and MIB2 parts of the structure */
    if (END_OBJ_INIT (&pDrvCtrl->endObj, (DEV_OBJ *)pDrvCtrl, LAN9118_DEV_NAME,
                       pDrvCtrl->unit, &LAN9118FuncTable,
                      "LAN9118 Enhanced Network Driver") == ERROR) {
	    goto errorExit;
	}


   /* MIB2 initialization is moved to Mac_Initialize routine */
	
   /* set buffer address; reserve 8 pages for TX */
   /* fixme - Rx/Rx page variables have not been used in any routine */
    pDrvCtrl->txStartPage = 0;
    pDrvCtrl->rxStartPage = pDrvCtrl->txStartPage + 8;
    pDrvCtrl->nextPage    = pDrvCtrl->rxStartPage;

    if (pDrvCtrl->rxStopPage >= (pDrvCtrl->rxStartPage + 1)) {
        pDrvCtrl->rringSize = pDrvCtrl->rxStopPage - pDrvCtrl->rxStartPage + 1;
    } else {
        DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "rxStopPage is less than rxStartPage...\n", 1, 2, 3, 4, 5, 6);
        pDrvCtrl->rxStopPage = pDrvCtrl->rxStartPage + 1;
        pDrvCtrl->rringSize = 2;
    }

    /* Perform memory allocation */
    if (LAN9118MemInit (pDrvCtrl) == ERROR) {
#ifdef VXWORKS_DEBUG
	    DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Error in MemInit...\n", 1, 2, 3, 4, 5, 6);
#endif
	    goto errorExit;
	}

#ifndef SCATTER_GATHER_EN
	#ifdef UNCASHED_TX_BUF
    if ((pDrvCtrl->txData = cacheDmaMalloc((ENET_HDR_REAL_SIZ + ETHERMTU + 3 * PLATFORM_CACHE_LINE_BYTES))) == NULL)
    {
		SMSC_WARNING("Not able to allocate uncached Mem\n");
	    goto errorExit;
    }
	#else
    if ((pDrvCtrl->txData = calloc ((ENET_HDR_REAL_SIZ + ETHERMTU + 3 * PLATFORM_CACHE_LINE_BYTES), 1)) == NULL)
    {
		SMSC_WARNING("Not able to allocate Mem\n");
	    goto errorExit;
    }
	#endif
#endif

    /* set the flags to indicate readiness */
    END_OBJ_READY (&pDrvCtrl->endObj,
                   IFF_NOTRAILERS | IFF_MULTICAST | IFF_BROADCAST);
    

    /* save the device address */

#ifdef LAN9118_DEBUG
    LAN9118EndDevice = pDrvCtrl;
    LAN9118DmaSemId  = pDrvCtrl->endObj.txSem;
#endif

    Platform_Initialize (pDrvCtrl);

    if (pDrvCtrl->dwLanBase == 0) {
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_LOAD, "pDrvCtrl->dwLanBase == 0\n", 1, 2, 3, 4, 5, 6);
#endif
	    goto errorExit;
    }
    
    pDrvCtrl->rx_dma = TRANSFER_REQUEST_DMA;/* Use TRANSFER_PIO; to set PIO Mode*/
    pDrvCtrl->tx_dma = TRANSFER_REQUEST_DMA;/* Use TRANSFER_PIO; to set PIO Mode*/
    pDrvCtrl->int_deas = 0xffffffff;
    pDrvCtrl->afc_cfg = 0xffffffff;
    pDrvCtrl->max_throughput = 0xffffffff;
    pDrvCtrl->max_packet_count = 0xffffffff;
    pDrvCtrl->packet_cost = 0xffffffff;
    pDrvCtrl->burst_period = 0xffffffff;
    pDrvCtrl->max_work_load = 0xffffffff;
    pDrvCtrl->tx_fif_sz = 0x00030000;
	pDrvCtrl->dwTxDmaThreshold = PLATFORM_DMA_THRESHOLD;	
    pDrvCtrl->dwRxDmaThreshold = PLATFORM_DMA_THRESHOLD;

    return (&pDrvCtrl->endObj);
	errorExit:
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Error LAN9118End...\n", 1, 2, 3, 4, 5, 6);
#endif
    LAN9118Unload (pDrvCtrl);
    return NULL;
    }

/******************************************************************************
*
* LAN9118Unload - unload a driver from the system
*
* This function first brings down the device, and then frees any
* stuff that was allocated by the driver in the load function.
*/
LOCAL STATUS LAN9118Unload
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
    if (pDrvCtrl != NULL) {
        END_OBJECT_UNLOAD (&pDrvCtrl->endObj);

        /* free resources */
        if (pDrvCtrl->pCluster != NULL)
            free (pDrvCtrl->pCluster);
            
#ifndef SCATTER_GATHER_EN
	#ifdef UNCASHED_TX_BUF
        if (pDrvCtrl->txData) {
            cacheDmaFree(pDrvCtrl->txData);
            pDrvCtrl->txData = NULL;
        }
    #endif
#endif
    }

    return (OK);
}

/*******************************************************************************
*
* LAN9118Start - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* RETURNS: OK or ERROR
*
*/
LOCAL STATUS LAN9118Start
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
    UINT32	acquired_isr=FALSE;
    int		i;
    STATUS 	result = ERROR, LanInitResult = OK;

    saved_pDrvCtrl = pDrvCtrl;

    /* Clear statuses */
    pDrvCtrl->rxHandling = FALSE;
    pDrvCtrl->txBlocked = FALSE;
    pDrvCtrl->rxOvw = FALSE;
    pDrvCtrl->txResend = FALSE;
    for (i = 0; i < GPT_SCHEDULE_DEPTH; i++) {
	    pDrvCtrl->GptFunction [i] = NULL;
    }
    pDrvCtrl->Gpt_scheduled_slot_index = GPT_SCHEDULE_DEPTH;
    
    /* initialize the LAN9118 */
    {
        UINT32 dwIntCfg=0;

		pDrvCtrl->irq_pol=PLATFORM_IRQ_POL;
		pDrvCtrl->irq_type=PLATFORM_IRQ_TYPE; 
        if(pDrvCtrl->irq_pol) {
            dwIntCfg|=INT_CFG_IRQ_POL_;
        }
        if(pDrvCtrl->irq_type) {
            dwIntCfg|=INT_CFG_IRQ_TYPE_;
        }
		pDrvCtrl->tx_fif_sz = 0x00050000;
		pDrvCtrl->afc_cfg = 0xFFFFFFFF;
        if ((LanInitResult = Lan_Initialize(pDrvCtrl,dwIntCfg)) != OK)
        {
#ifdef VXWORKS_DEBUG
            DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Failed Lan_Initialize.\n", 1, 2, 3, 4, 5, 6);
#endif
            goto DONE;
        }
    }

    /* connect interrupt handler */
    SYS_INT_CONNECT (pDrvCtrl, LAN9118Isr, (int)pDrvCtrl, &result);
    if (result == ERROR) {
		goto DONE;
    }
    acquired_isr=TRUE;
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "Interrupt connected.\n", 1, 2, 3, 4, 5, 6);
#endif

    /* reset to default value */
    result = ERROR;	

    /* mark the interface -- up */
    END_FLAGS_SET (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));

    /* Enable interrupts */
    SYS_INT_ENABLE (pDrvCtrl);
    
    /* must now test the IRQ connection to the ISR */
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "Testing ISR using IRQ %d\n", Platform_CurrentIRQ (pDrvCtrl), 2, 3, 4, 5, 6);
#endif
    {
        UINT32 dwTimeOut=100000;
        Lan_SignalSoftwareInterrupt (pDrvCtrl);
        do {
            Lan_udelay(pDrvCtrl, 10);
            dwTimeOut--;
        } while((dwTimeOut)&&(pDrvCtrl->SoftwareInterruptSignal == FALSE));
        
        if(pDrvCtrl->SoftwareInterruptSignal == FALSE) {
#ifdef VXWORKS_DEBUG
            DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "ISR failed signaling test\n", 1, 2, 3, 4, 5, 6);
#endif
            goto DONE;
        }
    }
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "ISR passed test using IRQ %d\n", Platform_CurrentIRQ (pDrvCtrl), 2, 3, 4, 5, 6);
#endif

	if (Mac_Initialize (pDrvCtrl) == ERROR) {
        DRV_LOG (LAN9118_DEBUG_LOAD, "Failed Mac_Initialize\n", 1, 2, 3, 4, 5, 6);
        goto DONE;
    }

    pDrvCtrl->MulticastUpdatePending = FALSE;

#ifdef USE_PHY_WORK_AROUND
    pDrvCtrl->phy_addr = 32; /* search for the address */
    if (Phy_Initialize (pDrvCtrl) == ERROR)
    {
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_PHY, "Failed to initialize Phy\n", 1, 2, 3, 4, 5, 6);
#endif
        goto DONE;
    }
#endif /*USE_PHY_WORK_AROUND*/

    {
        if(pDrvCtrl->rx_dma==TRANSFER_REQUEST_DMA) {
			pDrvCtrl->dwRxDmaCh=Platform_RequestDmaChannel(pDrvCtrl,RX_DMA);
            SMSC_ASSERT(pDrvCtrl->dwRxDmaCh!=TRANSFER_PIO);
            if(pDrvCtrl->dwRxDmaCh!=TRANSFER_PIO) {
                pDrvCtrl->RxDmaChReserved=TRUE;
            }
        } else {
       		pDrvCtrl->dwRxDmaCh=TRANSFER_PIO;
	        pDrvCtrl->RxDmaChReserved=FALSE;        	
       	}
			
       if(pDrvCtrl->tx_dma==TRANSFER_REQUEST_DMA) {
#ifdef VXWORKS_DEBUG
	       DRV_LOG (LAN9118_DEBUG_PHY, "TX DMA request\n", 1, 2, 3, 4, 5, 6);
#endif
            pDrvCtrl->dwTxDmaCh=Platform_RequestDmaChannel(pDrvCtrl,TX_DMA);
            SMSC_ASSERT(pDrvCtrl->dwTxDmaCh!=TRANSFER_PIO);
            if(pDrvCtrl->dwTxDmaCh!=TRANSFER_PIO) {
                pDrvCtrl->TxDmaChReserved=TRUE;
            }
        } else {
	    	pDrvCtrl->dwTxDmaCh=TRANSFER_PIO;      
            pDrvCtrl->TxDmaChReserved=FALSE;
        }
	   
#ifdef VXWORKS_DEBUG
	 DRV_LOG (LAN9118_DEBUG_PHY, "\nndwTxDmaCh%d \ndwRxDmaCh %d\n",pDrvCtrl->dwTxDmaCh, pDrvCtrl->dwRxDmaCh, 3, 4, 5, 6);
#endif
        Tx_Initialize(pDrvCtrl);
        Rx_Initialize(pDrvCtrl);
    }

    pDrvCtrl->Running=TRUE;
    Tx_StopQueue(pDrvCtrl,0x01); 

	spin_lock_init(&(pDrvCtrl->GpTimerLock));
    Lan_EnableInterrupt(pDrvCtrl,INT_EN_GPT_INT_EN_);

#ifndef USE_PHY_WORK_AROUND
    pDrvCtrl->phy_addr =32; /* search for the address */
    if (Phy_Initialize (pDrvCtrl) == ERROR)
    {
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_PHY, "Failed to initialize Phy\n", 1, 2, 3, 4, 5, 6);
#endif
        goto DONE;
    }
#endif /*USE_PHY_WORK_AROUND*/

DONE:
    if(result!=0) {
        if(pDrvCtrl!=NULL) {
            if(pDrvCtrl->TxDmaChReserved) {
                Platform_ReleaseDmaChannel(pDrvCtrl, TX_DMA);
                }
            if(pDrvCtrl->RxDmaChReserved) {
                Platform_ReleaseDmaChannel(pDrvCtrl, RX_DMA);
                }
            if(acquired_isr) {
                Platform_FreeIRQ(pDrvCtrl);
            }
        }
    }
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "LAN9118Start, result=%d\n", result, 2, 3, 4, 5, 6);
#endif
	if (LanInitResult == -ENODEV) {
		LAN9118Unload(pDrvCtrl);
	    return (-ENODEV);
	} else {
	    return (OK);
	}
}

/*******************************************************************************
*
* LAN9118Stop - stop the device
*
* This function calls BSP functions to disconnect interrupts and stop
* the device from operating in interrupt mode.
*
* RETURNS: OK or ERROR
*/
LOCAL STATUS LAN9118Stop
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
    STATUS		result = OK;
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_TMP, "-->LAN911x_stop(dev=0x%08X)\n", (UINT32) pDrvCtrl, 2, 3, 4, 5, 6);
#endif
    if(pDrvCtrl==NULL) {
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_TMP, "LAN911x_stop(pDrvCtrl==NULL)\n", 1, 2, 3, 4, 5, 6);
#endif
	result = ERROR;
	goto DONE;
    }

    pDrvCtrl->StopLinkPolling=TRUE;

    /* stop LAN9118 timer interrupt */
    Lan_DisableInterrupt(pDrvCtrl,INT_EN_GPT_INT_EN_);
    Gpt_CancelInterrupt (pDrvCtrl);

    /* stop all timer callbacks */
    Gpt_CancelCallBack (pDrvCtrl, GptCB_RestartBurst);
    Gpt_CancelCallBack (pDrvCtrl, GptCB_MeasureRxThroughput);
    Gpt_CancelCallBack (pDrvCtrl, GptCB_RxCompleteMulticast);
    Gpt_CancelCallBack (pDrvCtrl, (void (*) (LAN9118_END_DEV *)) Phy_CheckLink);

    Tx_UpdateTxCounters(pDrvCtrl);
    pDrvCtrl->Running=FALSE;

    Lan_DisableIRQ(pDrvCtrl);

    Tx_StopQueue(pDrvCtrl,0x01UL);
    
    if(pDrvCtrl->TxDmaChReserved) {
        Platform_ReleaseDmaChannel (pDrvCtrl,TX_DMA );
        pDrvCtrl->TxDmaChReserved=FALSE;
    }
    
    if(pDrvCtrl->RxDmaChReserved) {
        Platform_ReleaseDmaChannel (pDrvCtrl, RX_DMA);
        pDrvCtrl->RxDmaChReserved=FALSE;
    }

    END_FLAGS_CLR (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));

    /* Disable LAN interrupts */
    SYS_INT_DISABLE (pDrvCtrl);

    /* disconnect interrupt */
    SYS_INT_DISCONNECT (pDrvCtrl, LAN9118Isr, (int)pDrvCtrl, &result);

    if (result == ERROR)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_TMP, "Could not diconnect interrupt!\n",
                 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

    pDrvCtrl->ivec = 0;

	/* freeing semaphores */
	semTake(pDrvCtrl->ifCounterLock, WAIT_FOREVER);
	semDelete(pDrvCtrl->ifCounterLock);
	semTake(pDrvCtrl->IntEnableLock, WAIT_FOREVER);
	semDelete(pDrvCtrl->IntEnableLock);
	semTake(pDrvCtrl->MacPhyLock.Lock, WAIT_FOREVER);
	semDelete(pDrvCtrl->MacPhyLock.Lock);

DONE:
#ifdef VXWORKS_DEBUG
    if (result == OK) {
    	DRV_LOG (LAN9118_DEBUG_TMP, "<--LAN9118Stop, result=OK\n", 1, 2, 3, 4, 5, 6);
    }
    else {
    	DRV_LOG (LAN9118_DEBUG_TMP, "<--LAN9118Stop, result=ERROR\n", 1, 2, 3, 4, 5, 6);
    }
#endif

    return (result);
}

/*******************************************************************************
*
* LAN9118Send - the driver's actual output routine
*
* This routine takes a M_BLK_ID sends off the data in the M_BLK_ID.
* The buffer must already have the addressing information properly installed
* in it.  This is done by a higher layer.  The last arguments are a free
* routine to be called when the device is done with the buffer and a pointer
* to the argument to pass to the free routine.
*
* This routine accepts outgoing packets from the send queue, and then 
* gains exclusive access to the transmit hw (through a mutex semaphore),
* then calls Tx_SendBuf() to send the packet out onto the interface.
*
* RETURNS: OK, or ERROR if the packet could not be transmitted.
*/
LOCAL STATUS LAN9118Send
    (
    PLAN9118_END_DEV pDrvCtrl,
    M_BLK* pMblk
    )
{
    
	#ifdef TX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 0);
	#endif

    /* Check interface to see if it is up and running */
    if ((END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_UP | IFF_RUNNING)) != (IFF_UP | IFF_RUNNING)) {
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_TX, "Device is NOT UP and RUNNING\n", 1, 2, 3, 4, 5, 6);
#endif
		if (!LAN9118_IS_IN_POLL_MODE()) {
			/* free the Mblk chain */
			netMblkClChainFree (pMblk);
		}
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		errno = EINVAL;
        return (ERROR);
	}

	#ifdef TX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 1);
	#endif
	/* Check Device Mode */
	if (LAN9118_IS_IN_POLL_MODE()) {
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		/* free the Mblk chain */
		netMblkClChainFree (pMblk);
		errno = EINVAL;
        return (ERROR);
	}

    /*
     * Obtain exclusive access to transmitter.  This is necessary because
     * we might have more than one stack transmitting at once.
     */
	LAN9118_SEM_TAKE (pDrvCtrl, WAIT_FOREVER);

	#ifdef TX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 2);
	#endif

	if (pDrvCtrl->dwTxQueueDisableMask != 0) {	
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		/* Transmit queue is stopped - just signal we are out of resources without freeing packet */
	    LAN9118_SEM_GIVE (pDrvCtrl);
        return (END_ERR_BLOCK);
	}

    /* send packet out over interface */
#ifdef SCATTER_GATHER_EN
	if (Tx_SendFragBuf (pDrvCtrl, pMblk) == OK)
#else	
	if (Tx_SendBuf (pDrvCtrl, pMblk) == OK)
#endif
	{
		/* Success, free the Mblk chain */
		netMblkClChainFree (pMblk);
		END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
	} else {
		/* Failure: free rhe Mblk chain, update stats, release the semaphore and return */
		#ifdef VXWORKS_DEBUG
			#ifdef SCATTER_GATHER_EN
   			DRV_LOG (LAN9118_DEBUG_TX, "FAILED Tx_SendFragBuf\n", 0, 0, 0, 0, 0, 0);
   			#else
   			DRV_LOG (LAN9118_DEBUG_TX, "FAILED Tx_SendBuf\n", 0, 0, 0, 0, 0, 0);
   			#endif
		#endif
	    /* update statistics */
    	END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_ERRS, +1);
		#ifdef	LAN9118_INSTRUMENT
		LAN9118TxError++;
		#endif

		netMblkClChainFree (pMblk);
		errno = EINVAL;
	    LAN9118_SEM_GIVE (pDrvCtrl);
       	return (ERROR);
    }

    /* Release semaphore */
    LAN9118_SEM_GIVE (pDrvCtrl);
		
#ifdef	LAN9118_INSTRUMENT
    LAN9118TxNb++;
#endif
#ifdef VXWORKS_DEBUG
  	DRV_LOG (LAN9118_DEBUG_TX, "LAN9118Send\n", 0, 0, 0, 0, 0, 0);
#endif

    return OK;    
}

/*******************************************************************************
*
* LAN9118Ioctl - the driver's I/O control routine
*
* Perform device-specific commands.
*
* RETURNS: 0, or EINVAL if the command 'cmd' is not supported.
*********************************************************************************/

LOCAL int LAN9118Ioctl
    (
    PLAN9118_END_DEV pDrvCtrl,/* device receiving command */
    int			cmd,/* ioctl command code */
    caddr_t		data /* command argument */   
    )
{
    int 		error = 0;
    long		value;
    int			saveFlags;
    UINT32 		dwHigh16;
    UINT32 		dwLow32;
    UINT32		local_MACCR;
	UINT32      tempdata=0;
	BYTE        index=0;



#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "LAN9118Ioctl Command %d\n",cmd, 2, 3, 4, 5, 6);
#endif
    
    switch ((UINT)cmd)
	{
 		/******************* Device commands **************************/     

		/******************* set MAC address *****************************/
	    case EIOCSADDR:/* set device MAC address */
#ifdef VXWORKS_DEBUG
            DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCSADDR:\n", 1, 2, 3, 4, 5, 6);
#endif
		    if (data == NULL) {
			    return (EINVAL);
			}
	        
	        bcopy ((char *)data, (char *)END_HADDR(&pDrvCtrl->endObj),
			END_HADDR_LEN(&pDrvCtrl->endObj));

		    dwHigh16 = (pDrvCtrl->enetAddr [4]) |
			           (pDrvCtrl->enetAddr [5] << 8);

		    dwLow32 = (pDrvCtrl->enetAddr [3] << 24) | 
			          (pDrvCtrl->enetAddr [2] << 16) |
			          (pDrvCtrl->enetAddr [1] << 8) |
			          (pDrvCtrl->enetAddr [0]);

		 	{
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
	        	Mac_SetRegDW (pDrvCtrl, ADDRH, dwHigh16,keyCode);
	        	Mac_SetRegDW (pDrvCtrl, ADDRL, dwLow32,keyCode);
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			}
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_MAC, "Mac Address is set by default to 0x%04X%08X\n", dwHigh16, dwLow32, 3, 4, 5, 6);
#endif
        	break;
        	
		/******************* get MAC address *****************************/
        case EIOCGADDR:/* get MAC address */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCGADDR:\n", 1, 2, 3, 4, 5, 6);
#endif
	    	if (data == NULL) {
				return (EINVAL);
			}
			
            bcopy ((char *)END_HADDR(&pDrvCtrl->endObj), (char *)data,
		    END_HADDR_LEN(&pDrvCtrl->endObj));
            break;

		/***************  set (or clear) flags  **************************/
        case EIOCSFLAGS:/* set (or clear) flags */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCSFLAGS:\n", 1, 2, 3, 4, 5, 6);
#endif

		    saveFlags = DRV_FLAGS_GET();

		    value = (long) data;
		    if (value < 0) {
				value = -value;
				value--;
				END_FLAGS_CLR (&pDrvCtrl->endObj, value);
			} else {
				END_FLAGS_SET (&pDrvCtrl->endObj, value);
			}
           break;

        case EIOCGFLAGS: /* get flags */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCGFLAGS:\n", 1, 2, 3, 4, 5, 6);
#endif
	    	if (data == NULL) {
				error = EINVAL;
			} else {
				*(int *)data = END_FLAGS_GET(&pDrvCtrl->endObj);
			}
	    	break;
		
		/****************** Multicast commands ****************************/
        case EIOCMULTIADD:                  /* add m'cast addr */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCMULTIADD:\n", 1, 2, 3, 4, 5, 6);
#endif
            error = LAN9118MCastAddrAdd (pDrvCtrl, (char *)data);
            break;

        case EIOCMULTIDEL:                  /* del m'cast addr */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCMULTIDEL:\n", 1, 2, 3, 4, 5, 6);
#endif
            error = LAN9118MCastAddrDel (pDrvCtrl, (char *)data);
            break;

        case EIOCMULTIGET:                  /* get m'cast addr */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCMULTIGET:\n", 1, 2, 3, 4, 5, 6);
#endif
            error = LAN9118MCastAddrGet (pDrvCtrl, (MULTI_TABLE *)data);
            break;

		/********************* Polling functions ***************************/
        case EIOCPOLLSTART:                  /* Start Polling */
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCPOLLSTART:\n", 1, 2, 3, 4, 5, 6);
#endif
            error = LAN9118PollStart (pDrvCtrl);
            break;

        case EIOCPOLLSTOP:                   /* Stop Polling*/
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCPOLLSTOP:\n", 1, 2, 3, 4, 5, 6);
#endif
            error = LAN9118PollStop (pDrvCtrl);
            break;            
		
		/******************* MibII functions **************************** */
        case EIOCGMIB2233:
        case EIOCGMIB2:
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCGMIB2:\n", 1, 2, 3, 4, 5, 6);
#endif
            if (data == NULL) {
                return (EINVAL);
            }
            
            bcopy((char *)&pDrvCtrl->endObj.mib2Tbl, (char *)data,
                  sizeof(pDrvCtrl->endObj.mib2Tbl));
            break;
		
		/*****************Network Protocol Toolkit Command*************/
        case EIOCGNPT:	/* return true in case of toolkit device*/
		/* this driver is END so this call is not supported*/
#ifdef VXWORKS_DEBUG
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCGNPT:\n", 1, 2, 3, 4, 5, 6);
#endif
            return (EINVAL);
            break;
            
		/************** Access driver's internal 2233 stats counters ******/
        case EIOCGPOLLCONF:
        	DRV_LOG (LAN9118_DEBUG_IOCTL, "EIOCGPOLLCONF:\n", 1, 2, 3, 4, 5, 6);
            return (EINVAL);
#if 0
	    (*(END_IFDRVCONF **) data)->ifPollInterval = 2;
	    (*(END_IFDRVCONF **) data)->ifValidCounters = 0;
	    (*(END_IFDRVCONF **) data)->ifWatchdog = NULL;
	    (*(END_IFDRVCONF **) data)->ifEndObj = &pDrvCtrl->endObj;
	    (*(END_IFDRVCONF **) data)->ifMuxCookie = NULL;
	    (*(END_IFDRVCONF **) data)->ifPollRtn = NULL;
#endif
            break;

		case COMMAND_GET_SIGNATURE:
			break;

		case COMMAND_GET_FLOW_PARAMS:
			pDrvCtrl->ioctlData.Data[0]=pDrvCtrl->RxFlowMeasuredMaxThroughput;
			pDrvCtrl->ioctlData.Data[1]=pDrvCtrl->RxFlowMeasuredMaxPacketCount;
			pDrvCtrl->ioctlData.Data[2]=pDrvCtrl->RxFlowParameters.MaxThroughput;
			pDrvCtrl->ioctlData.Data[3]=pDrvCtrl->RxFlowParameters.MaxPacketCount;
			pDrvCtrl->ioctlData.Data[4]=pDrvCtrl->RxFlowParameters.PacketCost;
			pDrvCtrl->ioctlData.Data[5]=pDrvCtrl->RxFlowParameters.BurstPeriod;
			pDrvCtrl->ioctlData.Data[6]=pDrvCtrl->RxFlowMaxWorkLoad;
			pDrvCtrl->ioctlData.Data[7]=Lan_GetRegDW(INT_CFG)>>24;
			bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			break;
		
		case COMMAND_SET_FLOW_PARAMS:
			if(!(pDrvCtrl->RxFlowControlActive)) {
				bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
				pDrvCtrl->RxFlowParameters.MaxThroughput=pDrvCtrl->ioctlData.Data[2];
				pDrvCtrl->RxFlowParameters.MaxPacketCount=pDrvCtrl->ioctlData.Data[3];
				pDrvCtrl->RxFlowParameters.PacketCost=pDrvCtrl->ioctlData.Data[4];
				pDrvCtrl->RxFlowParameters.BurstPeriod=pDrvCtrl->ioctlData.Data[5];
				if(pDrvCtrl->ioctlData.Data[6]==0xFFFFFFFFUL) {
					pDrvCtrl->RxFlowMaxWorkLoad=
						pDrvCtrl->RxFlowParameters.MaxThroughput+
						(pDrvCtrl->RxFlowParameters.MaxPacketCount*
						pDrvCtrl->RxFlowParameters.PacketCost);
				} else {
					pDrvCtrl->RxFlowMaxWorkLoad=pDrvCtrl->ioctlData.Data[6];
				}
				Lan_SetIntDeas(pDrvCtrl,pDrvCtrl->ioctlData.Data[7]);
				pDrvCtrl->RxFlowBurstMaxWorkLoad=
					(pDrvCtrl->RxFlowMaxWorkLoad*
					pDrvCtrl->RxFlowParameters.BurstPeriod)/1000;

			}
			break;
			
		case COMMAND_GET_CONFIGURATION:
			bcopy((char*)&pDrvCtrl,(char*)data,sizeof(pDrvCtrl));
			break;
			
		case COMMAND_MAC_GET_REG:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if((pDrvCtrl->ioctlData.Data[0]<=0xC)&&(pDrvCtrl->LanInitialized)) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				
				pDrvCtrl->ioctlData.Data[1]=
					Mac_GetRegDW(pDrvCtrl,pDrvCtrl->ioctlData.Data[0],keyCode);
					
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
				bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			} else {
				SMSC_WARNING("Reading Mac Register Failed");
				goto MAC_ACCESS_FAILURE;
			}
			break;
			
		case COMMAND_MAC_SET_REG:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if((pDrvCtrl->ioctlData.Data[0]<=0xC)&&(pDrvCtrl->LanInitialized)) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				Mac_SetRegDW(
					pDrvCtrl,
					pDrvCtrl->ioctlData.Data[0],
					pDrvCtrl->ioctlData.Data[1],keyCode);
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			} else {
				SMSC_WARNING("Writing Mac Register Failed");
MAC_ACCESS_FAILURE:
				if(!(pDrvCtrl->LanInitialized)) {

					SMSC_WARNING("  LAN Not Initialized,");
					SMSC_WARNING("    Use ifconfig to bring interface UP");
				}
				if(!(pDrvCtrl->ioctlData.Data[0]<=0xC)) {
					SMSC_WARNING("  Invalid index ");
				}
			}
			break;

		case COMMAND_PHY_GET_REG:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if((pDrvCtrl->ioctlData.Data[0]<32)&&(pDrvCtrl->LanInitialized)) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				pDrvCtrl->ioctlData.Data[1]=((UINT32)
				Phy_GetRegW(pDrvCtrl,pDrvCtrl->ioctlData.Data[0],keyCode));
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
				bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			} else {
				SMSC_WARNING("Reading Phy Register Failed");
				goto PHY_ACCESS_FAILURE;
			}
			break;
			
		case COMMAND_PHY_SET_REG:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if((pDrvCtrl->ioctlData.Data[0]<32)&&(pDrvCtrl->LanInitialized)) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				Phy_SetRegW(
					pDrvCtrl,
					pDrvCtrl->ioctlData.Data[0],
					((UINT16)(pDrvCtrl->ioctlData.Data[1])),keyCode);
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			} else {
				SMSC_WARNING("Writing Phy Register Failed");
PHY_ACCESS_FAILURE:
				if(!(pDrvCtrl->LanInitialized)) {
					SMSC_WARNING("  Lan Not Initialized,");
					SMSC_WARNING("    Use ifconfig to bring interface UP");
				}
				if(!(pDrvCtrl->ioctlData.Data[0]<32)) {
					SMSC_WARNING("  Invalid index");
				}
			}
			break;

		case COMMAND_DUMP_LAN_REGS:
			Lan_ShowSysRegs(pDrvCtrl);
			break;
			
		case COMMAND_DUMP_MAC_REGS:
			if(pDrvCtrl->LanInitialized) {
				Lan_ShowMacRegs(pDrvCtrl);
			} else {
				SMSC_WARNING("Mac Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;

		case COMMAND_DUMP_PHY_REGS:
			if(pDrvCtrl->LanInitialized) {
			Lan_ShowPhyRegs(pDrvCtrl);
			} else {
				SMSC_WARNING("Phy Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;

		case COMMAND_DUMP_EEPROM:
			{
				BYTE tempdata=0;
				BYTE index=0;
				Eeprom_EnableAccess(pDrvCtrl);
				for(index=0;index<8;index++) {
					if(Eeprom_ReadLocation(pDrvCtrl,index,&tempdata)) {
						pDrvCtrl->ioctlData.Data[index]=(UINT32)tempdata;
					} else {
						break;
					}
				}
				Eeprom_DisableAccess(pDrvCtrl);
				bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			}
			break;
			
		case COMMAND_GET_MAC_ADDRESS:
			if(pDrvCtrl->LanInitialized) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				pDrvCtrl->ioctlData.Data[0]=Mac_GetRegDW(pDrvCtrl,ADDRH,keyCode);
				pDrvCtrl->ioctlData.Data[1]=Mac_GetRegDW(pDrvCtrl,ADDRL,keyCode);
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
				bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			} else {
				SMSC_WARNING("Lan Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;

		case COMMAND_SET_MAC_ADDRESS:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if(pDrvCtrl->LanInitialized) 
			{
				UINT32 dwLow32=pDrvCtrl->ioctlData.Data[1];
				UINT32 dwHigh16=pDrvCtrl->ioctlData.Data[0];
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
				Mac_SetRegDW(pDrvCtrl,ADDRH,dwHigh16,keyCode);
				Mac_SetRegDW(pDrvCtrl,ADDRL,dwLow32,keyCode);
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			} else {
				SMSC_WARNING("Lan Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;

		case COMMAND_LOAD_MAC_ADDRESS:
			if(pDrvCtrl->LanInitialized) {
				Eeprom_EnableAccess(pDrvCtrl);
				if(Eeprom_Reload(pDrvCtrl)) {
					if(Eeprom_IsMacAddressLoaded(pDrvCtrl)) {
						UINT32 dwIntFlags=0;
						VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
						pDrvCtrl->ioctlData.Data[0]=Mac_GetRegDW(pDrvCtrl,ADDRH,keyCode);
						pDrvCtrl->ioctlData.Data[1]=Mac_GetRegDW(pDrvCtrl,ADDRL,keyCode);
						Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
					} else {
						SMSC_WARNING("Failed to Load Mac Address(1)");
					}
				} else {
					SMSC_WARNING("Failed to Load Mac Address(2)");
				}
				Eeprom_DisableAccess(pDrvCtrl);

			} else {
				SMSC_WARNING("Lan Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;
			
		case COMMAND_SAVE_MAC_ADDRESS:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			if(pDrvCtrl->LanInitialized) {
				if(Eeprom_SaveMacAddress(pDrvCtrl,
					pDrvCtrl->ioctlData.Data[0],pDrvCtrl->ioctlData.Data[1])) {
				}
			} else {
				SMSC_WARNING("Lan Not Initialized,");
				SMSC_WARNING("  Use ifconfig to bring interface UP");
			}
			break;

#ifdef LAN9118_DEBUG
		case COMMAND_SET_DEBUG_MODE:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			LAN9118Debug_mode=	pDrvCtrl->ioctlData.Data[0];
			if(LAN9118Debug_mode&LAN9118_DEBUG_GPIO) {
				/* Set GPIO / LED pins as push/pull outputs */
				if(OLD_REGISTERS(pDrvCtrl)) 	
				{
					g_GpioSetting=0x00270700UL;	
				} else {
					/* switch EEPROM pins to mirror MII data signals */
					/* g_GpioSetting=0x00670700; */
					/* switch EEPROM to GPO3 and GPO4 */
					g_GpioSetting=0x00170700UL;
				}
				Lan_SetRegDW(GPIO_CFG,g_GpioSetting);
			} else {
				/* Set GPIO / LED pins as LEDs, and EEPROM pins as such */
				Lan_SetRegDW(GPIO_CFG,0x70070000);
			}
			break;
#endif

		case COMMAND_SET_LINK_MODE:
			bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
			pDrvCtrl->link_mode=(pDrvCtrl->ioctlData.Data[0]&0x7FUL);
			if(pDrvCtrl->LanInitialized) {
				Phy_SetLink(pDrvCtrl);
			}
			break;		
			
		case COMMAND_GET_LINK_MODE:
			pDrvCtrl->ioctlData.Data[0]=pDrvCtrl->link_mode;
			bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			break;
			
		case COMMAND_CHECK_LINK:
			 Phy_UpdateLinkMode(pDrvCtrl);
			break;

		case COMMAND_SET_AUTOMDIX:
			switch((pDrvCtrl->dwIdRev) &0xFFFF0000UL) {
				case 0x118A0000UL: 
				case 0x117A0000UL:
				case 0x115A0000UL:
				case 0x92100000UL:
				case 0x92110000UL:
				case 0x92200000UL:
				case 0x92210000UL:
					bcopy((char*)data,(char*)&pDrvCtrl->ioctlData,sizeof(pDrvCtrl->ioctlData));
					pDrvCtrl->AutoMdix=(UINT8)(pDrvCtrl->ioctlData.Data[0]&0xf);
					if(pDrvCtrl->LanInitialized) {
						Phy_SetAutoMdixSts(pDrvCtrl,pDrvCtrl->AutoMdix);
					}
					break;

					default:
					printf("Auto MDIX Not supported");
					break;
			}
			
		case COMMAND_GET_AUTOMDIX:
			pDrvCtrl->ioctlData.Data[0]=Phy_GetAutoMdixSts(pDrvCtrl);
			bcopy((char*)&pDrvCtrl->ioctlData,(char*)data,sizeof(pDrvCtrl->ioctlData));
			break;
	    
	    default:
	        return (EINVAL);
	        break;
	}
    
    return (error);
}  

/*******************************************************************************
*
* LAN9118MCastAddrAdd - add a multicast address
*
* This routine adds a multicast address to whatever the driver
* is already listening for.
*
*
* RETURNS: OK on success, ERROR otherwise.
*/
LOCAL STATUS LAN9118MCastAddrAdd
    (
    PLAN9118_END_DEV pDrvCtrl,
    char *		pAddr
    )
{
    int retVal;

    retVal = etherMultiAdd (&pDrvCtrl->endObj.multiList, pAddr);
    if (retVal == ENETRESET)
        return Lan_AddrFilterSet (pDrvCtrl, pAddr, TRUE);

    return ((retVal == OK) ? OK : ERROR);
}

/*******************************************************************************
*
* LAN9118MCastAddrDel - remove a multicast address
*
*
* This routine removes a multicast address from whatever the driver
* is listening for.
*
* RETURNS: OK on success, ERROR otherwise.
*/
LOCAL STATUS LAN9118MCastAddrDel
    (
    PLAN9118_END_DEV pDrvCtrl,
    char *		pAddr
    )
{
    int retVal;
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_IOCTL, "MCastAddrDel\n", 0, 0, 0, 0, 0, 0);
#endif
    retVal = etherMultiDel (&pDrvCtrl->endObj.multiList, pAddr);

    if (retVal == ENETRESET)
        return Lan_AddrFilterSet (pDrvCtrl, pAddr, FALSE);

    return ((retVal == OK) ? OK : ERROR);
}

/*******************************************************************************
*
* LAN9118MCastAddrGet - retreive current multicast address list
*
* This routine gets the multicast list of whatever the driver
* is already listening for.
*
* RETURNS: OK on success; otherwise ERROR.
*/
LOCAL STATUS LAN9118MCastAddrGet
    (
    PLAN9118_END_DEV pDrvCtrl,
    MULTI_TABLE *	pTable
    )
{
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_IOCTL, "MCastAddrGet\n", 0, 0, 0, 0, 0, 0);
#endif

    return (etherMultiGet (&pDrvCtrl->endObj.multiList, pTable));
}

/*******************************************************************************
*
* LAN9118PollStart - starting polling mode
*
* RETURNS: OK, always.
*/

LOCAL STATUS LAN9118PollStart
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
	UINT32 intLevel;
	
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_POLL, "S ", 0, 0, 0, 0, 0, 0);
#endif

    /* disable all interrupts */
    Lan_DisableIRQ(pDrvCtrl);
    /* atomic access */
    intLevel = intLock(); 
    DRV_FLAGS_SET (LAN9118_FLAG_POLL);
    intUnlock(intLevel);

    return (OK);    
}

/*******************************************************************************
*
* LAN9118PollStop - stop polling mode
*
* This function terminates polled mode operation.  The device returns to
* interrupt mode.
*
* The device interrupts are enabled, the current mode flag is switched
* to indicate interrupt mode and the device is then reconfigured for
* interrupt operation.
*
* RETURNS: OK, always.
*/

LOCAL STATUS LAN9118PollStop
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
	UINT32 intLevel;

    /* enable interrupts */
    Lan_EnableIRQ(pDrvCtrl);
    /* atomic access */
    intLevel = intLock(); 
    DRV_FLAGS_CLR(LAN9118_FLAG_POLL);
    intUnlock(intLevel);

#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_POLL, "s", 0, 0, 0, 0, 0, 0);
#endif    

    return (OK);
}

/*******************************************************************************
*
* LAN9118PollSend - send a packet in polled mode
*
* Polled mode operation takes place without any kernel or other OS
* services available.  Use extreme care to insure that this code does not
* call any kernel services.  Polled mode is only for WDB system mode use.
* Kernel services, semaphores, tasks, etc, are not available during WDB
* system mode.
*
* A typical implementation is to set aside a fixed buffer for polled send
* operation.  Copy the mblk data to the buffer and pass the fixed buffer
* to the device.  Performance is not a consideration for polled operations.
*
* An alternate implementation is a synchronous one.  The routine accepts
* user data but does not return until the data has actually been sent.  If an
* error occurs, the routine returns EAGAIN and the user will retry the request.
*
* If the device returns OK, then the data has been sent and the user may free
* the associated mblk.  The driver never frees the mblk in polled mode.
* The calling routine will free the mblk upon success.
*
* RETURNS: OK upon success.  EAGAIN if device is busy or no resources.
* A return of ERROR indicates a hardware fault or no support for polled mode
* at all.
*/
LOCAL STATUS LAN9118PollSend
    (
    PLAN9118_END_DEV pDrvCtrl,
    M_BLK*		pMblk
    )
{
    /* Check interface to see if it is up and running */
    if ((END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_UP | IFF_RUNNING)) != (IFF_UP | IFF_RUNNING)) {
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_TX, "Device is NOT UP and RUNNING\n", 1, 2, 3, 4, 5, 6);
#endif
		errno = EINVAL;
        return (ERROR);
	}
	
	/* Check Device Mode */
	if (!LAN9118_IS_IN_POLL_MODE()) {
		errno = EINVAL;
        return (ERROR);
	}

	if (pDrvCtrl->dwTxQueueDisableMask != 0)
	{	
		/* Transmit queue is stopped - just signal we are out of resources */
        return (END_ERR_BLOCK);
	}

    /* send packet out over interface */
#ifdef SCATTER_GATHER_EN
	if (Tx_SendFragBuf (pDrvCtrl, pMblk) == OK)
#else
	if (Tx_SendBuf (pDrvCtrl, pMblk) == OK)
#endif	
    {
		/* Success, don't free the Mblk chain, for polled mode caller will */
		END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
	}
	else
	{
		/* Failure: update stats, don't free rhe Mblk chain (in poll mode caller will), return */
		#ifdef VXWORKS_DEBUG
   		DRV_LOG (LAN9118_DEBUG_TX, "FAILED Tx_SendBuf\n", 0, 0, 0, 0, 0, 0);
		#endif
	    /* update statistics */
    	END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_ERRS, +1);
		#ifdef	LAN9118_INSTRUMENT
		LAN9118TxError++;
		#endif

		errno = EINVAL;
    	return (ERROR);
    }

#ifdef	LAN9118_INSTRUMENT
    LAN9118TxNb++;
#endif
#ifdef VXWORKS_DEBUG
  	DRV_LOG (LAN9118_DEBUG_TX, "LAN9118Send\n", 0, 0, 0, 0, 0, 0);
#endif

    return OK;    
}

/*******************************************************************************
*
* LAN9118PollReceive - get a packet in polled mode
*
* Polled mode operation takes place without any kernel or other OS
* services available.  Use extreme care to insure that this code does not
* call any kernel services.  Polled mode is only for WDB system mode use.
* Kernel services, semaphores, tasks, etc, are not available during WDB
* system mode.
*
* The WDB agent polls the device constantly looking for new data.  Typically
* the device has a ring of RFDs to receive incoming packets.  This routine
* examines the ring for any new data and copies it to the provided mblk.
* The concern here is to keep the device supplied with empty buffers at all
* time.
*
* RETURNS: OK upon success.  EAGAIN is returned when no packet is available.
* A return of ERROR indicates a hardware fault or no support for polled mode
* at all.
*/
LOCAL STATUS LAN9118PollReceive
    (
    PLAN9118_END_DEV pDrvCtrl,
    M_BLK*		pMblk
    )
{
    STATUS		nRetValue = OK;
    
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_POLL, "Start Poll Read!\n", 1, 2, 3, 4, 5, 6);
#endif

    if (Lan_ReadFrameBufferSupplied (pDrvCtrl, pMblk) == ERROR) {
        nRetValue = EAGAIN;
    }

#ifdef VXWORKS_DEBUG   
    DRV_LOG (LAN9118_DEBUG_POLL, "End Poll Read!\n", 1, 2, 3, 4, 5, 6);        
#endif

    return nRetValue;
}

/*******************************************************************************
*
* LAN9118InitParse - parse the init string
*
* Parse the input string.  Fill in values in the driver control structure.
*
* The muxLib.o module automatically prepends the unit number to the user's
* initialization string from the BSP (configNet.h).
*
* "<unit>:<baseAdrs>:<intVec>:<intLvl>:<bufSize>:<link_mode>:<macAddrString>"
*
* .IP <unit>
* Device unit number, a small integer.
* .IP <vecNum>
* Interrupt vector number
* .IP <intLvl>
* Interrupt level
* .LP
*
* RETURNS: OK or ERROR for invalid arguments.
*/
/* Need to parse initString instead of hard coding driver parameters */ 
LOCAL STATUS LAN9118InitParse
    (
    PLAN9118_END_DEV pDrvCtrl,
    char * initString
    )
    {
    int		i;
    char	digit1, digit0;
    char	error;
    char *	tok;
    char *	pHolder = NULL;
	unsigned long temp_mac=0;
    long	address;
    UINT32 dwHigh16=0;
	UINT32 dwLow32=0;
	UINT32 dwTimeOut=0;
	UINT32 e2p_cmd=0;
	/* For some reason, initString always comes up empty! Need more work here. */	
		
  /* Parse the initString */
	/* fixme - do boundary check on the values that are passed to the driver */
	
    /* Unit number. */
    tok = strtok_r (initString, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
    pDrvCtrl->unit = atoi (tok); 

    /* NIC address. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
    address = strtoul (tok, NULL, 16);
    pDrvCtrl->dwLanBase = (UINT32) address; 

    /* Interrupt vector. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
     pDrvCtrl->ivec = (UINT32)atoi (tok); 

    /* Interrupt level. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
     pDrvCtrl->ilevel = atoi (tok); 

    /* Get ring buffer size in pages */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
     pDrvCtrl->rxStopPage = strtoul (tok, NULL, 16) / LAN9118_PAGE_SIZE; 

    /* Get link mode */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	return ERROR;
     pDrvCtrl->link_mode = strtoul (tok, NULL, 16); 

	/* Get MDIX data*/
	tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
	{
        pDrvCtrl->AutoMdix= 3; /* Auto Mdxi controlled by STRAP*/
	} else {
        pDrvCtrl->AutoMdix =(UINT32)atoi (tok); 
	}

    /* get optional MAC address */
    tok = strtok_r (NULL, ":", &pHolder);
    if ((tok == NULL)){
		E2Prm_LoadFlag=1;
	} 
    else {
		if(strlen (tok) != 12){
			E2Prm_LoadFlag=1;
		}
		strcpy (pDrvCtrl->macAddrString,tok);
    }
	

	
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "Processed all arugments\n", 1, 2, 3, 4, 5, 6);

    DRV_LOG (LAN9118_DEBUG_LOAD, "\nAddress %x \nLvl %u \nVec %d \nRing %p \nlink_mode 0x%02x\n",
               (int)pDrvCtrl->dwLanBase, pDrvCtrl->ilevel, pDrvCtrl->ivec, 
	       0x1111, pDrvCtrl->link_mode, 6);    
/*    DRV_LOG (LAN9118_DEBUG_LOAD, "MAC = %s\n", pDrvCtrl->macAddrString, 2, 3, 4, 5, 6);*/
#endif

    return OK;
    }
    
/*******************************************************************************
*
* LAN9118MemInit - initialize memory for the LAN9118 chip
*
* This routine is highly specific to the device.
*
* Using data in the control structure, setup and initialize the memory
* areas needed.  If the memory address is not already specified, then allocate
* cache safe memory.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS LAN9118MemInit
    (
    PLAN9118_END_DEV pDrvCtrl		/* device to be initialized */
    )
    {
    /* allocate netpool */

    if ((pDrvCtrl->endObj.pNetPool = malloc (sizeof(NET_POOL))) == NULL) {
        return (ERROR);
    }

    /* Set number of M-Blks and CL-Blks*/

    LAN9118MclBlkConfig.mBlkNum	= pDrvCtrl->rringSize;
    LAN9118MclBlkConfig.clBlkNum= pDrvCtrl->rringSize / 2;
    LAN9118ClDescTbl[0].clNum	= LAN9118MclBlkConfig.clBlkNum;

    /* Calculate the total memory for all the M-Blks and CL-Blks. */

    LAN9118MclBlkConfig.memSize = LAN9118MclBlkConfig.mBlkNum * 
				   (MSIZE + sizeof (long)) +
				  LAN9118MclBlkConfig.clBlkNum * 
				   (CL_BLK_SZ + sizeof(long));

    /* allocate memory for M-Blks and CL-Blks */

    LAN9118MclBlkConfig.memArea	= (char *) memalign (sizeof(long),
				   LAN9118MclBlkConfig.memSize);
	

    if (LAN9118MclBlkConfig.memArea == NULL) {
	    return (ERROR);
	}
    
    /* Calculate the memory size of all the clusters. */

    LAN9118ClDescTbl[0].clSize  = LAN9118_RX_BUF_SZ;
    LAN9118ClDescTbl[0].memSize = sizeof(long) + LAN9118ClDescTbl[0].clNum * 
				  (LAN9118ClDescTbl[0].clSize + 8);

    /* Allocate memory for clusters */

    LAN9118ClDescTbl[0].memArea = (char *) memalign (sizeof(long),
				   LAN9118ClDescTbl[0].memSize);
   
    if (LAN9118ClDescTbl[0].memArea == NULL) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "system memory unavailable\n", 
		 	1, 2, 3, 4, 5, 6);
#endif
		return (ERROR);
	}

    /* Save clusters start address for later use */

    pDrvCtrl->pCluster = (LAN9118_CLUSTER) LAN9118ClDescTbl[0].memArea;

    /* Initialize the net buffer pool with buffers */

    if (netPoolInit (pDrvCtrl->endObj.pNetPool, &LAN9118MclBlkConfig, 
		     &LAN9118ClDescTbl[0], LAN9118ClDescTblNumEnt, NULL) == ERROR) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Could not init buffering\n",
		 	1, 2, 3, 4, 5, 6);
#endif
		return (ERROR);
	}
    
    /* Store the cluster pool id as others need it later. */

    pDrvCtrl->pClPoolId = netClPoolIdGet (pDrvCtrl->endObj.pNetPool, 
					  LAN9118_RX_BUF_SZ, FALSE);
   
    if (pDrvCtrl->pClPoolId  == NULL) {
		return (ERROR);
	}

#ifdef LAN9118_DEBUG
    LAN9118NetPool = *pDrvCtrl->endObj.pNetPool;
    LAN9118ClPoolId = pDrvCtrl->pClPoolId;
#endif

    return (OK);
    }

/*******************************************************************************
*
* LAN9118Isr - The driver's interrupt handler
*
* This function clears the cause of the device interrupt(s) and then acts
* on the individual possible causes.  The primary goal of this routine is to
* minimize the time spent in it.  This is accomplished by deferring processing
* to the netTask via the netJobAdd() function.
*
* Note that in case the receiver overruns, we promptly mark the interface as
* "down" and leave error handling to task-level.   This is in case netTask
* is in the midst of DMA activity, we must allow it to complete.  The receive
* handler will give up when it discovers the interface is down, which will
* then allow netTask to run our OVW handler.  This provides orderly error 
* recovery.
*
* RETURNS: N/A.
*/
LOCAL void LAN9118Isr
    (
    PLAN9118_END_DEV pDrvCtrl
    )
{
	UINT32 dwIntCfg=0;
	UINT32 dwIntSts=0;

	if(pDrvCtrl==NULL) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_INT, "LAN911x_ISR(pDrvCtrl==NULL)\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	if(pDrvCtrl->dwLanBase==0) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_INT, "LAN911x_ISR(dwLanBase==0)\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

    SET_GPIO(GPIO_ISR);

	DRV_LOG (LAN9118_DEBUG_INT, "dwIntSts %x , %x, %x ", pDrvCtrl->LastIntStatus1, pDrvCtrl->LastIntStatus2, pDrvCtrl->LastIntStatus3, 4, 5, 6);

    /* Check if interface is up and running */
    if ((END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_UP | IFF_RUNNING)) != (IFF_UP | IFF_RUNNING)) {
	    goto DONE;
	}    

#ifdef	LAN9118_INSTRUMENT
        LAN9118IntNb++;
#endif

	dwIntCfg=Lan_GetRegDW(INT_CFG);
	if((dwIntCfg&0x00001100)!=0x00001100) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_INT, "In ISR, not my interrupt, dwIntCfg=0x%08X\n", dwIntCfg, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	{
		UINT32 reservedBits;
		if(OLD_REGISTERS(pDrvCtrl)) {
			reservedBits=0x00FFEEEEUL;
		} else {
			reservedBits=0x00FFCEEEUL;
		}
		if(dwIntCfg&reservedBits) {
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_INT, "In ISR, reserved bits are high.\n", 1, 2, 3, 4, 5, 6);
#endif
			goto DONE;
		}
	}

	dwIntSts=Lan_GetRegDW(INT_STS);
	pDrvCtrl->LastIntStatus3=pDrvCtrl->LastIntStatus2; 
	pDrvCtrl->LastIntStatus2=pDrvCtrl->LastIntStatus1;
	pDrvCtrl->LastIntStatus1=dwIntSts;
	
	/* Acknowledge the interrupts that are not masked */
	dwIntSts &= pDrvCtrl->imask;
	Lan_SetRegDW(INT_STS, dwIntSts);

    if (dwIntSts&0x00904058) { /* RXDFH + RXD + RXE + RXDF + RSFF + RSFL */
    	/* Disable Rx Interrupts */
		pDrvCtrl->imask &= (~pDrvCtrl->RxInterrupts);
		Lan_ClrBitsDW (INT_EN, pDrvCtrl->RxInterrupts);
    	
        /* Defer the handler to the task level */
        netJobAdd ((FUNCPTR) Rx_HandleInterrupt, (int) pDrvCtrl, dwIntSts, 0, 0, 0);
    }

    if (dwIntSts&INT_STS_TDFA_) {
    	/* Disable TDFA Interrupt */
		pDrvCtrl->imask &= (~INT_EN_TDFA_EN_);
		Lan_ClrBitsDW (INT_EN, INT_EN_TDFA_EN_);

        /* Defer the handler to the task level */
        netJobAdd ((FUNCPTR) Tx_HandleInterrupt, (int) pDrvCtrl, dwIntSts, 0, 0, 0);
    }
    
    if (dwIntSts&INT_STS_RXSTOP_INT_) {
    	/* Disable Rx Stop Interrupt */
		pDrvCtrl->imask &= (~INT_EN_RXSTOP_INT_EN_);
		Lan_ClrBitsDW (INT_EN, INT_EN_RXSTOP_INT_EN_);

        /* Defer the handler to the task level */
        netJobAdd ((FUNCPTR) RxStop_HandleInterrupt, (int) pDrvCtrl, dwIntSts, 0, 0, 0);
    }

    if (dwIntSts&INT_STS_GPT_INT_) {
    	/* Disable GPT Interrupt */
		pDrvCtrl->imask &= (~INT_EN_GPT_INT_EN_);
		Lan_ClrBitsDW (INT_EN, INT_EN_GPT_INT_EN_);

        /* Defer the handler to the task level */
        netJobAdd ((FUNCPTR) Gpt_HandleInterrupt, (int) pDrvCtrl, dwIntSts, 0, 0, 0);
    }
    
    if (dwIntSts & INT_STS_SW_INT_) {
    	/* Disable S/W Interrupt */
		pDrvCtrl->imask &= (~INT_EN_SW_INT_EN_);
		Lan_ClrBitsDW (INT_EN, INT_EN_SW_INT_EN_);
        Lan_HandleSoftwareInterrupt(pDrvCtrl, dwIntSts);
    }
    
DONE:
    CLR_GPIO(GPIO_ISR);
}

/*******************************************************************************
*
* LAN9118_DrvDetails - Driver detailed info.
*
*
* RETURNS: N/A.
*/
typedef struct _SHOW_REG
{
	char  szName[20];
	UINT32 dwOffset;
} SHOW_REG;

LOCAL void LAN9118_DrvDetails(void)
{
	UINT32 test=0;
#ifndef	BYTE_TEST_OFFSET
	#define BYTE_TEST_OFFSET	(0x64)
#endif
	PLAN9118_END_DEV pDrvCtrl;
	pDrvCtrl = saved_pDrvCtrl;
	printf("\n	Driver Version = %1X.%02X",
		(DRIVER_VERSION >> 8),(DRIVER_VERSION & 0xFFUL));
	printf("\n	PLATFORM_NAME %s\n",PLATFORM_NAME);
	switch((pDrvCtrl->dwIdRev)&0xFFFF0000UL) {
	case 0x01180000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
		case 1UL:
			printf("  	No support for this device, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			goto DONE;
			break;

		case 2UL:
			printf("  	LAN9118 Concord A1 identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			break;
		default:
			printf("  	LAN9118 Concord A1 identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			break;
		};break;
	case 0x01170000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
		case 1UL:
			printf("  	No support for this device, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			goto DONE;
			break;

		case 2UL:
			printf("  	LAN9117 Concord A1 identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			break;
		default:
			printf("  	LAN9117 Concord A1 identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			break;
		};break;
	case 0x01160000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			goto UNKNOWN_CHIP;
		case 1UL:
			printf("  	No support for this device, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			goto DONE;
			break;

		case 2UL:
			printf("  	LAN9116 Concord A1 identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			break;
		default:
			printf("  	LAN9116 Concord A1 identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;
	case 0x01150000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			goto UNKNOWN_CHIP;
		case 1UL:
			printf("  	No support for this device, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);
			goto DONE;
			break;

		case 2UL:
			printf("  	LAN9115 Concord A1 identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN9115 Concord A1 identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;
	case 0x118A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9218 Boylston identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN9218 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;
	case 0x117A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9217 Boylston identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN9217 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;
	case 0x115A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9215 Boylston identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN9215 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x215A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN89215 Boylston auto identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN89215 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x216A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN89216 Boylston auto identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN89216 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x217A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN89217 Boylston auto identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN89217 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x218A0000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN89218 Boylston auto identified, dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		default:
			printf("  	LAN89218 Boylston identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX ",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x92100000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9210 Boylston Lite identified, dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
	
		default:
			printf("  	LAN9210 Boylston Lite identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x92110000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9211 Boylston Lite identified, dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
	
		default:
			printf("  	LAN9211 Boylston Lite identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x92200000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9220 Bedford identified, dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
	
		default:
			printf("  	LAN9220 Bedford identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	case 0x92210000UL:
		switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
		case 0UL:
			printf("  	LAN9221 Bedford identified, dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
	
		default:
			printf("  	LAN9221 Bedford identified (NEW), dwIdRev=0x%08lX, Generation=0x%08lX",pDrvCtrl->dwIdRev,pDrvCtrl->dwGeneration);

			break;
		};break;

	default:
	UNKNOWN_CHIP:
		{
		printf("  	LAN9118 Family NOT Identified, dwIdRev=0x%08lX",pDrvCtrl->dwIdRev);
		};break;
	}
	printf("\n	Date Code: %s  %s\n", __DATE__, __TIME__);
	printf("\n	***************Driver Parameters*****************\n");

	printf("\n	Lan Base at 0x%08lX\n",  pDrvCtrl->dwLanBase);
	test=(*(volatile unsigned long *)((pDrvCtrl->dwLanBase)+BYTE_TEST_OFFSET));
	if(pDrvCtrl->dwBitWidth==32) {
		printf("\n	MODE:	32 bit \n");
		}
	else
		{
		printf("\n	MODE:	16 bit \n");
		}
	if( pDrvCtrl->rx_dma ==TRANSFER_PIO)
	{
		printf( "	Receive Mode :	PIO");
	}
	else
	{
		printf( "	Receive Mode :	DMA");
	}
	if( pDrvCtrl->tx_dma ==TRANSFER_PIO)
	{
		printf( "	Transmit Mode :	PIO");
	}
	else
	{
		printf( "	Transmit Mode :	DMA");
	}
	 printf( "\n	ISR %d", Platform_CurrentIRQ (pDrvCtrl));
	 printf( "		Interupt Level: %u	Vector %d ",pDrvCtrl->ilevel, pDrvCtrl->ivec);
     printf("\n	ENET Addr: %02lX:%02lX:%02lX:%02lX:%02lX:%02lX \n",
            pDrvCtrl->enetAddr[0], pDrvCtrl->enetAddr[1], 
	     	pDrvCtrl->enetAddr[2], pDrvCtrl->enetAddr[3], 
	     	pDrvCtrl->enetAddr[4], pDrvCtrl->enetAddr[5]);
		printf("\n	Link_mode        = 0x%lX,\n	supported Modes %s,%s,%s,%s,%s,%s,%s\n",
			pDrvCtrl->link_mode,
			(pDrvCtrl->link_mode&LINK_SPEED_10HD)?"10HD":"",
			(pDrvCtrl->link_mode&LINK_SPEED_10FD)?"10FD":"",
			(pDrvCtrl->link_mode&LINK_SPEED_100HD)?"100HD":"",
			(pDrvCtrl->link_mode&LINK_SPEED_100FD)?"100FD":"",
			(pDrvCtrl->link_mode&LINK_ASYMMETRIC_PAUSE)?"ASYMP":"",
			(pDrvCtrl->link_mode&LINK_SYMMETRIC_PAUSE)?"SYMP":"",
			(pDrvCtrl->link_mode&LINK_AUTO_NEGOTIATE)?"ANEG":"");

		switch((pDrvCtrl->dwIdRev) &0xFFFF0000UL)
		{
		case 0x118A0000UL: 
		case 0x117A0000UL:
		case 0x115A0000UL:
		case 0x92100000UL:
		case 0x92110000UL:
		case 0x92200000UL:
		case 0x92210000UL:
			switch((pDrvCtrl->AutoMdix)&0xFFFFUL) 
			{
			case 0UL:
			printf("	Auto MDIX Disabled - Set to Straight Cable ");
			break;
			case 1UL:
			printf("	Auto MDIX Disabled - Set to Cross Over  Cable ");
			break;
			case 2UL:
			printf("	Auto MDIX Enabled  ");
			break;
			default:
			printf("	Auto MDIX Controlled By Strap");
			break;
			};
		break;
		default:		
 			printf("	Auto MDIX :	Not Supported");
			break;
		};	
  	printf("\n	**************************************************\n");

DONE:
}

/*****************************************************************************
**Function Name: Lan_Initialize
                 Called by LAN9118Start() to initialize the LAN9118 device.
** Desc:

**Input:LAN9118_END_DEVICE pointer 

**OutPut:STATUS

****************************************************************************/
LOCAL STATUS Lan_Initialize (
        PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwIntCfg)
{
	STATUS result = ERROR;
	UINT32 dwTimeOut = 0;
	UINT32 dwTemp = 0;
	UINT32 e2p_cmd;
	
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_LOAD, "-->Lan_Initialize(dwIntCfg=0x%08X)\n", dwIntCfg, 2, 3, 4, 5, 6);
#endif
	SMSC_ASSERT(pDrvCtrl != NULL);

	pDrvCtrl->dwIdRev=Lan_GetRegDW(ID_REV);
	if(HIWORD(pDrvCtrl->dwIdRev)==LOWORD(pDrvCtrl->dwIdRev)) {
	    /*this may mean the chip is set for 32 bit 
	      while the bus is reading as 16 bit*/
	    printf("  LAN9118 Family NOT Identified, dwIdRev=0x%08lX",pDrvCtrl->dwIdRev);
	    result=-ENODEV;
	    goto DONE;
	}
	
	switch((pDrvCtrl->dwIdRev)&0xFFFF0000UL) {

		case 0x01180000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
				case 1UL:
					result = -ENODEV;
					DRV_LOG(LAN9118_DEBUG_WARNING, "SMSC_WARNING: No support for this device, dwIdRev==0x%08lX \n", pDrvCtrl->dwIdRev,2,3,4,5,6);
					goto DONE;

				case 2UL:
					pDrvCtrl->dwGeneration=2;
					break;
				default:
					pDrvCtrl->dwGeneration=2;
					break;
			};
			break;

		case 0x01170000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
				case 1UL:
					result = -ENODEV;
					DRV_LOG(LAN9118_DEBUG_WARNING, "SMSC_WARNING: No support for this device, dwIdRev==0x%08lX \n", pDrvCtrl->dwIdRev,2,3,4,5,6);
					goto DONE;

				case 2UL:
					pDrvCtrl->dwGeneration=2;
					break;
				default:
					pDrvCtrl->dwGeneration=2;
					break;
			};
			break;
			
		case 0x01160000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					goto UNKNOWN_CHIP;
				case 1UL:
					result = -ENODEV;
					DRV_LOG(LAN9118_DEBUG_WARNING, "SMSC_WARNING: No support for this device, dwIdRev==0x%08lX \n", pDrvCtrl->dwIdRev,2,3,4,5,6);
					goto DONE;
				case 2UL:
					pDrvCtrl->dwGeneration=2;
					break;
				default:
					pDrvCtrl->dwGeneration=2;
					break;
				};
				break;

		case 0x01150000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					goto UNKNOWN_CHIP;
				case 1UL:
					result = -ENODEV;
					DRV_LOG(LAN9118_DEBUG_WARNING, "SMSC_WARNING: No support for this device, dwIdRev==0x%08lX \n", pDrvCtrl->dwIdRev,2,3,4,5,6);
					goto DONE;
				case 2UL:
					pDrvCtrl->dwGeneration=2;
					break;
				default:
					pDrvCtrl->dwGeneration=2;
					break;
				};
				break;

		case 0x118A0000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=3;
					break;
				default:
					pDrvCtrl->dwGeneration=3;
					break;
			};
			break;

		case 0x117A0000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=3;
					break;
				default:
					pDrvCtrl->dwGeneration=3;
					break;
			};
			break;

		case 0x116A0000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=3;
					break;
				default:
					pDrvCtrl->dwGeneration=3;
					break;
			};
			break;

		case 0x115A0000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=3;
					break;
				default:
					pDrvCtrl->dwGeneration=3;
					break;
			};
			break;

		case 0x92100000UL:
		case 0x92200000UL:
			pDrvCtrl->UseScatterGather=TRUE;
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
			
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};
			break;

		case 0x92110000UL:
		case 0x92210000UL:
			pDrvCtrl->UseScatterGather=TRUE;
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch((pDrvCtrl->dwIdRev)&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
			
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};
			break;

		case 0x93120000UL:
			pDrvCtrl->UseScatterGather=FALSE;
			pDrvCtrl->UseTxCsum=FALSE;
			pDrvCtrl->UseRxCsum=FALSE;		
			switch(pDrvCtrl->dwIdRev&0x0000FFFFUL) {
				case 1UL:
					pDrvCtrl->dwGeneration=3;
					break;
				default:
					pDrvCtrl->dwGeneration=3;
					break;
			};
			break;

		case 0x215A0000UL:
			pDrvCtrl->UseScatterGather=TRUE; 
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch(pDrvCtrl->dwIdRev&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};
			break;

		case 0x216A0000UL:
			pDrvCtrl->UseScatterGather=TRUE; 
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch(pDrvCtrl->dwIdRev&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};break;

		case 0x217A0000UL:
			pDrvCtrl->UseScatterGather=TRUE; 
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch(pDrvCtrl->dwIdRev&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};break;

		case 0x218A0000UL:
			pDrvCtrl->UseScatterGather=TRUE; 
			pDrvCtrl->UseTxCsum=FALSE; 
			pDrvCtrl->UseRxCsum=FALSE; 
			switch(pDrvCtrl->dwIdRev&0x0000FFFFUL) {
				case 0UL:
					pDrvCtrl->dwGeneration=4;
					break;
				default:
					pDrvCtrl->dwGeneration=4;
					break;
			};break;

		default:
		UNKNOWN_CHIP:
			result = ERROR;
			DRV_LOG(LAN9118_DEBUG_WARNING, "SMSC_WARNING: LAN9118 Family NOT Identified, dwIdRev==0x%08lX \n", pDrvCtrl->dwIdRev,2,3,4,5,6);
			goto DONE;
			break;
	}

	/* Reset the LAN9118 */
	/* LAN software reset */
	Lan_SetRegDW (HW_CFG, HW_CFG_SRST_);
	dwTimeOut = 1000000; /* 10 msec timeout to complete reset */

	do {
		Lan_udelay (pDrvCtrl, 10);
		dwTemp = Lan_GetRegDW (HW_CFG);
		dwTimeOut--;
	} while ((dwTimeOut > 0) && (dwTemp&HW_CFG_SRST_));

	if(dwTemp & HW_CFG_SRST_) {
#ifdef VXWORKS_DEBUG
    		DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "  Failed to complete reset.\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	/* Set Tx Data FIFO size */
	SMSC_ASSERT (pDrvCtrl->tx_fif_sz >= 0x00020000);
	SMSC_ASSERT (pDrvCtrl->tx_fif_sz <= 0x000E0000);
	SMSC_ASSERT ((pDrvCtrl->tx_fif_sz & (~HW_CFG_TX_FIF_SZ_)) == 0);
	Lan_SetRegDW (HW_CFG, pDrvCtrl->tx_fif_sz);

	/* Rx Data FIFO size */
	pDrvCtrl->RxMaxDataFifoSize = 0;
	switch (pDrvCtrl->tx_fif_sz >> 16) {
		case 2: pDrvCtrl->RxMaxDataFifoSize = 13440; break;
		case 3: pDrvCtrl->RxMaxDataFifoSize = 12480; break;
		case 4: pDrvCtrl->RxMaxDataFifoSize = 11520; break;
		case 5: pDrvCtrl->RxMaxDataFifoSize = 10560; break;
		case 6: pDrvCtrl->RxMaxDataFifoSize = 9600; break;
		case 7: pDrvCtrl->RxMaxDataFifoSize = 8640; break;
		case 8: pDrvCtrl->RxMaxDataFifoSize = 7680; break;
		case 9: pDrvCtrl->RxMaxDataFifoSize = 6720; break;
		case 10: pDrvCtrl->RxMaxDataFifoSize = 5760; break;
		case 11: pDrvCtrl->RxMaxDataFifoSize = 4800; break;
		case 12: pDrvCtrl->RxMaxDataFifoSize = 3840; break;
		case 13: pDrvCtrl->RxMaxDataFifoSize = 2880; break;
		case 14: pDrvCtrl->RxMaxDataFifoSize = 1920; break;
		default: SMSC_ASSERT (FALSE); break;
	}

	if(pDrvCtrl->afc_cfg == 0xFFFFFFFF) {
		switch (pDrvCtrl->tx_fif_sz) {

			/* AFC_HI is about ((Rx Data Fifo Size)*2/3)/64
			AFC_LO is AFC_HI/2
			BACK_DUR is about 5uS*(AFC_LO) rounded down
			*/
			case 0x00020000:/*13440 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x008C46AF;break;
			case 0x00030000:/*12480 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0082419F;break;
			case 0x00040000:/*11520 Rx Data Fifo Size */

				pDrvCtrl->afc_cfg=0x00783C9F;break;
			case 0x00050000:/*10560 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x006E374F;break;
			case 0x00060000:/* 9600 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0064328F;break;
			case 0x00070000:/* 8640 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x005A2D7F;break;
			case 0x00080000:/* 7680 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0050287F;break;
			case 0x00090000:/* 6720 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0046236F;break;
			case 0x000A0000:/* 5760 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x003C1E6F;break;
			case 0x000B0000:/* 4800 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0032195F;break;

			/*AFC_HI is ~1520 bytes less than RX Data Fifo Size */
			/*AFC_LO is AFC_HI/2 */
			/*BACK_DUR is about 5uS*(AFC_LO) rounded down */
			case 0x000C0000:/* 3840 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0024124F;break;
			case 0x000D0000:/* 2880 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0015073F;break;
			case 0x000E0000:/* 1920 Rx Data Fifo Size */
				pDrvCtrl->afc_cfg=0x0006032F;break;
			default:SMSC_ASSERT(FALSE);break;
		}
	}
	Lan_SetRegDW (AFC_CFG, (pDrvCtrl->afc_cfg&0xFFFFFFF0));

	/*make sure EEPROM has finished loading before setting GPIO_CFG */
	dwTimeOut=1000;
	e2p_cmd = Lan_GetRegDW (E2P_CMD);
	while((dwTimeOut>0)&&(e2p_cmd&E2P_CMD_EPC_BUSY_)) {
		Lan_udelay(pDrvCtrl, 5);
		dwTimeOut--;
		e2p_cmd = Lan_GetRegDW (E2P_CMD);
	}
	
	if(dwTimeOut==0) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_LOAD_ERROR, "Lan_Initialize: Timed out waiting for EEPROM busy bit to clear\n", 1, 2, 3, 4, 5, 6);
#endif
	}

	if(LAN9118Debug_mode&LAN9118_DEBUG_GPIO) {
		/* Set GPIO / LED pins as push/pull outputs */
		if(OLD_REGISTERS(pDrvCtrl)) 
		{
			g_GpioSetting=0x00270700;
		} else {
			/* switch EEPROM pins to mirror MII data signals */
			/* g_GpioSetting=0x00670700; */
			/* switch EEPROM to GPO3 and GPO4 */
			g_GpioSetting=0x00170700;
		}
	} else {
		/* Set GPIO / LED pins as LEDs, and EEPROM pinsfl as such */
		g_GpioSetting = 0x70070000;
	}
	Lan_SetRegDW (GPIO_CFG, g_GpioSetting);
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_LOAD, "g_GpioSetting set\n", 1, 2, 3, 4, 5, 6);
#endif
	
	/*initialize interrupts */
	Lan_SetRegDW (INT_EN, 0);
	Lan_SetRegDW (INT_STS, 0xFFFFFFFF);
	dwIntCfg|=INT_CFG_IRQ_EN_;
	Lan_SetRegDW (INT_CFG, dwIntCfg);

	Vl_InitLock(&(pDrvCtrl->MacPhyLock));
	spin_lock_init(&(pDrvCtrl->IntEnableLock));

	spin_lock_init(&(pDrvCtrl->ifCounterLock));
	pDrvCtrl->pIf = NULL;
	pDrvCtrl->pIfAddr = NULL;

	pDrvCtrl->LanInitialized=TRUE;
	result = OK;

DONE:
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_LOAD, "Lan_Initialize\n", 1, 2, 3, 4, 5, 6);
#endif
	return result;
}

/*******************************************************************************
*
* Lan_EnetAddrGet - get the Ethernet address.
*
* Get ethernet address from the BSP.
*
* RETURNS: N/A.
*/
LOCAL void Lan_EnetAddrGet
    (
    PLAN9118_END_DEV pDrvCtrl, 
    char*		addr
    )
{
	int	i, j;
	char 	digit [2];
	int 	error;
	 for (i = 0; i < 6; i++) {
	    error = 0;
	    for (j = 0; j < 2; j++) {
	        digit [j] = pDrvCtrl->macAddrString [2*i+j];
	        if (digit [j] <= '9' && digit [j] >= '0') {
		    digit [j] -= '0';
	        }
	        else if (digit [j] <= 'F' && digit [j] >= 'A') {
	    	    digit [j] -= 'A';
	    	    digit [j] += 10;
	        }
	        else if (digit [j] <= 'f' && digit [j] >= 'a') {
			    digit [j] -= 'a';
	    	    digit [j] += 10;
	        }
	        else {
		    error = 1;
		    break;
	        }
	    }
		

	    if (error) {
		/* default set your own MAC address as allocated */
        	addr [0] = 0x00;
        	addr [1] = 0x80;
        	addr [2] = 0x0F;
        	addr [3] = 0x11;
        	addr [4] = 0x70;
        	addr [5] = 0x00;
		break;
	    }
	    else {
	        addr [i] = digit [0] * 16 + digit [1];
		    }
	}
	
}

/*******************************************************************************
*
* Lan_ReadFrameBufferSupplied - read a packet off the interface ring buffer into a cluster
*
* Allocates a new cluster from the cluster pool, and reads the frame from the
* device into the cluster.
*
* RETURNS: a cluster or NULL if any error
*/
LOCAL STATUS Lan_ReadFrameBufferSupplied
    (
    PLAN9118_END_DEV pDrvCtrl,
    M_BLK* pMblk
    )
{
    UINT32		dwRxStatus;
    UINT32 		dwPacketLength;

    /* Check receive status */
    if ((Lan_GetRegDW (RX_FIFO_INF) & RX_FIFO_INF_RXSUSED_) == 0)
        return ERROR;

    if (!(END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_UP | IFF_RUNNING))) 
        return ERROR;

    /* read out LAN9118 RX status */
    /*Use PIO only*/	
    Lan_SetRegDW(RX_CFG,0x00000200UL);
	
    dwRxStatus=Rx_PopRxStatus(pDrvCtrl, Lan_GetRegDW(INT_STS));
    if (dwRxStatus == 0) 
    {
        return ERROR;
    }

    dwPacketLength=((dwRxStatus&0x3FFF0000UL)>>16);
#ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_RX, "Start Read (%d)!\n", dwPacketLength, 2, 3, 4, 5, 6);
#endif
    Rx_CountErrors(pDrvCtrl,dwRxStatus);
    if((dwRxStatus&RX_STS_ES_) == 0) 
    {
        pDrvCtrl->RxPacketReadCount++;
        pDrvCtrl->RxPioReadCount++;
        pDrvCtrl->RxDataDWReadCount+=
			(dwPacketLength+2+3)>>2;

        Platform_ReadFifo (pDrvCtrl->dwLanBase, ((UINT32 *)(pMblk->mBlkHdr.mData)), (dwPacketLength+2+3)>>2);

        pMblk->mBlkHdr.mFlags |= M_PKTHDR; /* set the packet header */
        pMblk->mBlkHdr.mLen = dwPacketLength - 4;	/* set the data len, remove CRC */
        pMblk->mBlkPktHdr.len = dwPacketLength - 4; /* set the total len, remove CRC */
        pMblk->mBlkHdr.mData += 2; /* Adjust mData for Rx Offset */

#ifdef	LAN9118_INSTRUMENT
	    LAN9118Len = dwPacketLength;
#endif
		END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX, "End Read (OK)!\n", 1, 2, 3, 4, 5, 6);
#endif
		return OK;
    }
    else 
    {
        /*if we get here then the packet is to be read */
        /*  out of the fifo and discarded */
		if(pDrvCtrl->dwRxDmaCh<TRANSFER_PIO)
		{
			dwPacketLength=((dwPacketLength+(PLATFORM_CACHE_LINE_BYTES-14)
					+PLATFORM_CACHE_LINE_BYTES-1)&(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;
	
		}
		else
		{
	        dwPacketLength+=(2+3);
    	    dwPacketLength>>=2;
		}
        Rx_FastForward(pDrvCtrl,dwPacketLength); 
        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX, "End Read (ERROR)!\n", 1, 2, 3, 4, 5, 6);
#endif
		return ERROR;
    }
   
    return OK;
}

/**********************************************************************************
*
* Lan_ReadFrameBufferNotSupplied - read a packet off the interface ring buffer into a cluster
* Allocates a new cluster from the cluster pool, and reads the frame from the
* device into the cluster. Pass the cluster to the Higer layer . read can be performed either through 
* PIO or DMA.
*
* RETURNS: a cluster or NULL if any error
*/
/**********************************************************************************/
LOCAL M_BLK*	Lan_ReadFrameBufferNotSupplied (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts)
{
    UINT32		dwRxStatus;
    UINT32 		dwPacketLength;
	UINT32		dwDwordCount, dwIntFlags;
    LAN9118_RX_FRAME *pRx, *pCpyRx;
    CL_BLK_ID	pClBlk;
    M_BLK_ID	pMblk;
    
    int iRxLoopCnt = RX_POLLING_COUNT;
    
    /* Check receive status */
	if ((Lan_GetRegDW (RX_FIFO_INF) & RX_FIFO_INF_RXSUSED_) == 0)
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "Lan_ReadFrameBufferNotSupplied: RX_FIFO_INF_RXSUSED_ = 0!\n", 1, 2, 3, 4, 5, 6);
#endif
		PULSE_GPIO(GPIO_RX_ERR);
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
        return NULL;
	}

	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 4);
	#endif
    if (!(END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_UP | IFF_RUNNING))) 
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "Lan_ReadFrameBufferNotSupplied: Interface down\n", 1, 2, 3, 4, 5, 6);
#endif
		PULSE_GPIO(GPIO_RX_ERR);
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
        return NULL;
	}

	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 5);
	#endif
	if((dwRxStatus=Rx_PopRxStatus(pDrvCtrl, dwIntSts))==0) 
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "Lan_ReadFrameBufferNotSupplied: Rx_PopRxStatus = 0\n", 1, 2, 3, 4, 5, 6);
#endif
		SMSC_WARNING("Rx_PopRxStatus problem");
		PULSE_GPIO(GPIO_RX_ERR);
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		return NULL;
	}

RUN_AGAIN:

	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 15);
	#endif
	Rx_CountErrors(pDrvCtrl,dwRxStatus);
	dwPacketLength=((dwRxStatus&0x3FFF0000UL)>>16) ;
	if((dwRxStatus&RX_STS_ES_)==0) 
	{
		UINT32 dwDwordCount;
		
		/* get packet buffer */
        if ((pRx = (LAN9118_RX_FRAME*) netClusterGet (pDrvCtrl->endObj.pNetPool, pDrvCtrl->pClPoolId)) == NULL) 
        {
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 31);
			#endif
            /* Tell the mux to give back some clusters */
			pDrvCtrl->lastError.errCode = END_ERR_NO_BUF;
	    	muxError (&pDrvCtrl->endObj, &pDrvCtrl->lastError);
            SMSC_WARNING ("Lan_ReadFrameBufferNotSupplied: Cannot get a Net Cluster!\n");
			PULSE_GPIO(GPIO_RX_ERR);
	    	goto ERROR_SKIP;
   		}
   		
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 16);
		#endif
        if ((pClBlk = netClBlkGet (pDrvCtrl->endObj.pNetPool, M_DONTWAIT)) == NULL) 
        {
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 31);
			#endif
 	    	pDrvCtrl->lastError.errCode = END_ERR_NO_BUF;
	    	muxError (&pDrvCtrl->endObj, &pDrvCtrl->lastError);
            SMSC_WARNING ("Lan_ReadFrameBufferNotSupplied: Cannot get a Cluster Block!\n");
            netClFree (pDrvCtrl->endObj.pNetPool, (UCHAR *)pRx);        
			PULSE_GPIO(GPIO_RX_ERR);
	    	goto ERROR_SKIP;
   	    }

        /*
   	     * OK we've got a spare, let's get an M_BLK_ID and marry it to the
       	 * one in the ring.
         */

		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 17);
		#endif
        if ((pMblk = mBlkGet (pDrvCtrl->endObj.pNetPool, M_DONTWAIT, MT_DATA)) == NULL) 
        {
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 31);
			#endif
            netClBlkFree (pDrvCtrl->endObj.pNetPool, pClBlk);
   	        netClFree (pDrvCtrl->endObj.pNetPool, (UCHAR *)pRx);
            SMSC_WARNING ("Lan_ReadFrameBufferNotSupplied: Cannot get a M Block!\n");
			PULSE_GPIO(GPIO_RX_ERR);
	    	goto ERROR_SKIP;
   	    }
        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);
		
		 /* Join the cluster to the MBlock */
	  	netClBlkJoin (pClBlk, (char*)pRx, dwPacketLength, NULL, 0, 0, 0);
     	netMblkClJoin (pMblk, pClBlk);
		
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 18);
		#endif
		if((pDrvCtrl->dwRxDmaCh<TRANSFER_PIO)&(dwPacketLength>=pDrvCtrl->dwRxDmaThreshold))
		{
			/* Cache line alignment */
			pCpyRx = (LAN9118_RX_FRAME*)(((UINT32)pRx + PLATFORM_CACHE_LINE_BYTES - 1) & (~(PLATFORM_CACHE_LINE_BYTES-1)));
		
			/*Use DMA and PIO*/
			pDrvCtrl->RxDmaXfer.dwLanReg=pDrvCtrl->dwLanBase+RX_DATA_FIFO;
			pDrvCtrl->RxDmaXfer.pdwBuf=NULL;/* this will be reset per dma request*/
			pDrvCtrl->RxDmaXfer.dwDmaCh=pDrvCtrl->dwRxDmaCh;
			pDrvCtrl->RxDmaXfer.dwDwCnt=0;/* this will be reset per dma request*/
			pDrvCtrl->RxDmaXfer.fMemWr=TRUE;

			dwDwordCount=((dwPacketLength+(PLATFORM_CACHE_LINE_BYTES-14)
						+PLATFORM_CACHE_LINE_BYTES-1)&(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;
			pDrvCtrl->RxDmaXfer.dwDwCnt=dwDwordCount;
			pDrvCtrl->RxDmaReadCount++;
			pDrvCtrl->RxDmaXfer.pdwBuf=(UINT32 *)pCpyRx;
		
			pDrvCtrl->RxPacketReadCount++;
		    pDrvCtrl->RxDataDWReadCount+=(dwPacketLength+2+3)>>2;
			cacheFlush (DATA_CACHE, pDrvCtrl->RxDmaXfer.pdwBuf, (pDrvCtrl->RxDmaXfer.dwDwCnt)<<2);
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 19);
			#endif
			switch(PLATFORM_CACHE_LINE_BYTES) 
			{
				case 16:Lan_SetRegDW(RX_CFG,(RX_CFG_RX_END_ALGN16_ |RX_CFG_RXDOFF_2_));break;
				case 32:Lan_SetRegDW(RX_CFG,(RX_CFG_RX_END_ALGN32_ |RX_CFG_RXDOFF_18_));break;
				default:SMSC_ASSERT(FALSE);
			}
			
			SET_GPIO(GPIO_RX_COPY);
			if(!Platform_DmaStartXfer(pDrvCtrl,&(pDrvCtrl->RxDmaXfer))) {
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 31);
				#endif
#ifdef VXWORKS_DEBUG
            	DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "Lan_ReadFrameBufferNotSupplied: Failed Platform_DmaStartXfer!\n", 1, 2, 3, 4, 5, 6);
#endif
				SMSC_WARNING("Failed Platform_DmaStartXfer");
			}

			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 20);
			#endif
			Platform_DmaComplete(pDrvCtrl,pDrvCtrl->dwRxDmaCh);
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 21);
			#endif
			CLR_GPIO(GPIO_RX_COPY);
		}
		else 
		{
			pCpyRx = pRx;

			Lan_SetRegDW(RX_CFG,0x00000200UL);
			SET_GPIO(GPIO_RX_COPY);
			Platform_ReadFifo (pDrvCtrl->dwLanBase, ((UINT32 *)(pCpyRx)), (dwPacketLength+2+3)>>2);
			CLR_GPIO(GPIO_RX_COPY);
	        pDrvCtrl->RxPioReadCount++;
			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 21);
			#endif
		}

		pMblk->mBlkHdr.mData = (char *)&pCpyRx->enetHdr;
		pMblk->mBlkHdr.mFlags |= M_PKTHDR;   /* set the packet header */
       	pMblk->mBlkHdr.mLen = dwPacketLength - 4;	/* set the data dwPacketLength, remove CRC  */
        pMblk->mBlkPktHdr.len = dwPacketLength - 4;	/* set the total len, remove CRC  */

#ifdef	LAN9118_INSTRUMENT
   		LAN9118Len = dwPacketLength;
#endif
		#ifdef LAN9118_DEBUG
		if (LAN9118Debug_mode & LAN9118_DEBUG_RX_DATA_DUMP) 
		{
			printf("Rx (length with crc appended = %d):", dwPacketLength, 3,4,5,6);
			DumpBuffer((UINT8 *) pMblk->mBlkHdr.mData, dwPacketLength);
		}
		#endif

       	/* Call the upper layer's receive routine. */
		END_RCV_RTN_CALL (&pDrvCtrl->endObj, pMblk);
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 22);
		#endif
	
		if (iRxLoopCnt)
		{
			UINT32 flag, dwTmpIntSts;
			
			iRxLoopCnt--;
			
			flag = intLock();
			dwTmpIntSts = Lan_GetRegDW(INT_STS);
			intUnlock(flag);
			
			if((dwRxStatus=Rx_PopRxStatus(pDrvCtrl, dwTmpIntSts))!=0)
			{
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 23);
				#endif
				goto RUN_AGAIN;
			}
		}

		return pMblk;
	}
	else
	{
ERROR_SKIP:
        /*if we get here then the packet is to be read */
        /*  out of the fifo and discarded */
		if(pDrvCtrl->dwRxDmaCh<TRANSFER_PIO)
		{
			dwPacketLength=((dwPacketLength+(PLATFORM_CACHE_LINE_BYTES-14)
				+PLATFORM_CACHE_LINE_BYTES-1)&(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;
		}
		else
		{
   	    	dwPacketLength+=(2+3);
        	dwPacketLength>>=2;
		}

        Rx_FastForward(pDrvCtrl,dwPacketLength);
   	    END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
		END_ADD_IF_DROP(pDrvCtrl, 1);
   	    
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "FFWD:: dwRxStatus=0x%08lX, dwPacketLength=0x%04lX\n", dwRxStatus, dwPacketLength, 3, 4, 5, 6);
        DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "ERROR_SKIP dwRxStatus (NULL)!0x%08lX\n", dwRxStatus, 2, 3, 4, 5, 6);
#endif
		return NULL;
	}
}

/*******************************************************************************
*
* Lan_HandleRxInt - deferred receive interrupt handler
*
* This function handles the received frames from the device.  It runs in the
* context of the netTask, which was triggered by a received packet interrupt.
* It resends any packet that was in the midst of transmission when the overflow
* hit.
*
* RETURNS: N/A.
*/
LOCAL void Lan_HandleRxInt
    (
    PLAN9118_END_DEV pDrvCtrl,
    UINT32 dwIntSts
	)
{
	SET_GPIO(GPIO_RX_JOB);
	
    /* empties the receive ring buffer of its packets */
    if(Lan_ReadFrameBufferNotSupplied (pDrvCtrl, dwIntSts) == NULL) 
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_RX_ERR, "Rx Error in Lan_HandleRxInt\n", 1, 2, 3, 4, 5, 6);
        DRV_LOG (LAN9118_DEBUG_RX_ERR, "dwIntSts = 0x%08lX\n", dwIntSts, 2, 3, 4, 5, 6);
#endif
	}

	CLR_GPIO(GPIO_RX_JOB);

	/* Ack the interrupt. */
	{
		UINT32 flag;

		flag = intLock();
		Lan_SetRegDW(INT_STS,INT_STS_RSFL_);	
		intUnlock(flag);
	}

	if (pDrvCtrl->FCRxInterruptsEn == TRUE)
	{
		Lan_EnableInterrupt(pDrvCtrl,pDrvCtrl->RxInterrupts);
	}
 }

/*******************************************************************************
*
* Lan_HashIndex - compute the hash index for an ethernet address
*
* RETURNS: hash index for an ethernet address.
*/
LOCAL int Lan_HashIndex
    (
    char *	eAddr
    )
{
    UINT8	eAddrByte;
    int		index;                       /* hash index - return value */
    int		byte;                        /* loop - counter */
    int		bit;                         /* loop - counter */
    UINT	crc = 0xffffffff;
    UINT8	msb;

    for (byte=0; byte<6; byte++)
        {
        eAddrByte = eAddr[byte];

        for (bit=0; bit<8; bit++)
            {
            msb = crc >> 31;
            crc <<= 1;
            if (msb ^ (eAddrByte & 0x1))
                {
                crc ^= LAN9118_CRC_POLY;
                crc |= 0x1;
                }
            eAddrByte >>= 1;
            }
        }

    /* Just want the 6 most significant bits. */
    
    index = crc >> 26 ;

    return index;
}

/******************************************************************************
*
* Lan_AddrFilterSet - set the address filter for multicast addresses
*
* This routine goes through all of the multicast addresses on the list
* of addresses (added with the MCastAddrAdd() routine) and sets the
* device's filter correctly.
*
* NOMANUAL
*/
LOCAL STATUS Lan_AddrFilterSet
    (
    PLAN9118_END_DEV pDrvCtrl,
    char*		pAddr,
    BOOL		bSet
    )
{

    UINT8		nHashIndex;
	VL_KEY keyCode=0;
	UINT32 dwIntFlags=0;

    SMSC_ASSERT(pDrvCtrl!=NULL);

    /* get hash index for the address */
    nHashIndex = Lan_HashIndex (pAddr);

    /* Turn on the corresponding bit in the filter. */
	keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

    if (END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_PROMISC)) {
		pDrvCtrl->set_bits_mask = MAC_CR_PRMS_;
		pDrvCtrl->clear_bits_mask = (MAC_CR_MCPAS_ | MAC_CR_HPFILT_);

		pDrvCtrl->HashHi = 0UL;
		pDrvCtrl->HashLo = 0UL;
		goto PREPARE;
	}

    if (END_FLAGS_GET(&pDrvCtrl->endObj) & (IFF_ALLMULTI)) {
		pDrvCtrl->set_bits_mask = MAC_CR_MCPAS_;
		pDrvCtrl->clear_bits_mask = (MAC_CR_PRMS_ | MAC_CR_HPFILT_);

		pDrvCtrl->HashHi = 0UL;
		pDrvCtrl->HashLo = 0UL;
		goto PREPARE;
	}

	pDrvCtrl->set_bits_mask = MAC_CR_HPFILT_;
	pDrvCtrl->clear_bits_mask = (MAC_CR_PRMS_ | MAC_CR_MCPAS_);

	pDrvCtrl->HashLo = Mac_GetRegDW(pDrvCtrl, HASHL,keyCode);
	pDrvCtrl->HashHi = Mac_GetRegDW(pDrvCtrl, HASHH,keyCode);
	if (bSet == TRUE) {
	    if (nHashIndex < 32) {
	        pDrvCtrl->HashLo |= (1 << nHashIndex);
	    }
	    else {
	        pDrvCtrl->HashHi |= (1 << (nHashIndex - 32));
	    }
	}
	else {
	    if (nHashIndex < 32) {
	        pDrvCtrl->HashLo &= ~(1 << nHashIndex);
	    }
	    else {
	        pDrvCtrl->HashHi &= ~(1 << (nHashIndex - 32));
	    }
	}

PREPARE:
	{
		UINT32 local_MACCR;

		Mac_SetRegDW(pDrvCtrl,HASHH,pDrvCtrl->HashHi,keyCode);
		Mac_SetRegDW(pDrvCtrl,HASHL,pDrvCtrl->HashLo,keyCode);
		local_MACCR = Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);
		local_MACCR |= pDrvCtrl->set_bits_mask;
		local_MACCR &= ~(pDrvCtrl->clear_bits_mask);
		Mac_SetRegDW(pDrvCtrl,MAC_CR,local_MACCR,keyCode);
	}
	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	
    return OK;
}

/*********************************************************************************
Function Name:  Lan_EnableInterrupt

Description :   enable the interrupts with the interrrupt mask 

input: 	LAN9118_END_DEVICE	device handler and mask

output: void 

**********************************************************************************/
LOCAL void Lan_EnableInterrupt(PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntEnMask)
{
	UINT32 dwIntFlags=0;

	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized==TRUE);

	spin_lock_irqsave(&(pDrvCtrl->IntEnableLock),dwIntFlags);
	pDrvCtrl->imask |= dwIntEnMask;
	Lan_SetBitsDW (INT_EN, dwIntEnMask);
	spin_unlock_irqrestore(&(pDrvCtrl->IntEnableLock),dwIntFlags);
}

/*********************************************************************************
Function Name:  Lan_DisableInterrupt

Description :   Disables the interrupt

input: 	LAN9118_END_DEVICE	device handler

output: void 

**********************************************************************************/
LOCAL void Lan_DisableInterrupt(PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntEnMask)
{
	UINT32 dwIntFlags=0;

	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized==TRUE);
	
	spin_lock_irqsave(&(pDrvCtrl->IntEnableLock),dwIntFlags);
	pDrvCtrl->imask &= (~dwIntEnMask);
	Lan_ClrBitsDW (INT_EN, dwIntEnMask);
	spin_unlock_irqrestore(&(pDrvCtrl->IntEnableLock),dwIntFlags);
}

/*********************************************************************************
Function Name:  Lan_SetTDFL

Description :   
Interrupt lack  for the following functions have been commented out
  because at this time they are not necessary.
These function are 
  Lan_SetTDFL
  Lan_SetTSFL
  Lan_SetRSFL
Both the Rx and Tx side of the driver use the FIFO_INT,
  but the Rx side only touches is during initialization,
  so it is sufficient that Tx side simple preserve the Rx setting
input: 	LAN9118_END_DEVICE	device handler 	

output: void 

**********************************************************************************/
LOCAL void Lan_SetTDFL(PLAN9118_END_DEV pDrvCtrl, UINT8 level)
{
	SMSC_ASSERT(pDrvCtrl != NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized == TRUE);
	{
		UINT32 temp = Lan_GetRegDW (FIFO_INT);
		temp &= 0x00FFFFFFUL;
		temp |= ((UINT32) level) << 24;
		Lan_SetRegDW (FIFO_INT, temp);
	}
}

/*********************************************************************************
Function Name:  Lan_SetTSFL

Description :   

input: 	LAN9118_END_DEVICE	device handler 		

output: void 

**********************************************************************************/
LOCAL void Lan_SetTSFL(PLAN9118_END_DEV pDrvCtrl, UINT8 level) 
{
	
	SMSC_ASSERT(pDrvCtrl != NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized == TRUE);
	{
		UINT32 temp = Lan_GetRegDW (FIFO_INT);
		temp &= 0xFF00FFFF;
		temp |= ((UINT32) level) << 16;
		Lan_SetRegDW (FIFO_INT, temp);
	}
}

/*********************************************************************************
Function Name:  

Description :   this function sets the Set the RX FIFO status Level . the interrupt will be 
				generated depends on this level

input: 		LAN9118_END_DEVICE	device handler 	and level

output: void 

**********************************************************************************/
LOCAL void Lan_SetRSFL(PLAN9118_END_DEV pDrvCtrl, UINT8 level) 
{
	SMSC_ASSERT(pDrvCtrl != NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized == TRUE);
	{
		UINT32 temp = Lan_GetRegDW (FIFO_INT);
		temp &= 0xFFFFFF00;
		temp |= ((UINT32) level);
		Lan_SetRegDW (FIFO_INT, temp);
	}
}

/*********************************************************************************
Function Name:  Lan_EnableIRQ

Description :   

input: LAN9118_END_DEVICE	device handler 		

output: void 

**********************************************************************************/
LOCAL void Lan_EnableIRQ(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 dwIntFlags=0;

	spin_lock_irqsave(&(pDrvCtrl->IntEnableLock),dwIntFlags);
	Lan_SetBitsDW (INT_CFG, INT_CFG_IRQ_EN_);
	spin_unlock_irqrestore(&(pDrvCtrl->IntEnableLock),dwIntFlags);
}

/*********************************************************************************
Function Name:  

Description :   

input: 		

output: void 

**********************************************************************************/
LOCAL void Lan_DisableIRQ(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 dwIntFlags=0;

	spin_lock_irqsave(&(pDrvCtrl->IntEnableLock),dwIntFlags);
	Lan_ClrBitsDW (INT_CFG, INT_CFG_IRQ_EN_);
	spin_unlock_irqrestore(&(pDrvCtrl->IntEnableLock),dwIntFlags);
}

/*********************************************************************************
Function Name:  Lan_SetIntDeas

Description :   

input: 		LAN9118_END_DEVICE	device handler 	

output: void 

**********************************************************************************/
LOCAL void Lan_SetIntDeas(PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntDeas)
{
	UINT32 dwIntFlags=0;
	UINT32 dwTemp=0;

	spin_lock_irqsave(&(pDrvCtrl->IntEnableLock),dwIntFlags);
	{
		Lan_ClrBitsDW (INT_CFG, INT_CFG_INT_DEAS_);

		/* Added to comply with read-after-write time delay restriction
		   for high performance devices. */
		dwTemp=Lan_GetRegDW(BYTE_TEST);
		dwTemp = 1UL;
		dwTemp=Lan_GetRegDW(BYTE_TEST);
		dwTemp = 2UL;
		dwTemp=Lan_GetRegDW(BYTE_TEST);
		dwTemp = dwTemp;

		Lan_SetBitsDW (INT_CFG, (dwIntDeas << 24));
	}
	spin_unlock_irqrestore(&(pDrvCtrl->IntEnableLock),dwIntFlags);
}

/*********************************************************************************
Function Name:  Lan_SignalSoftwareInterrupt

Description :   

input: 		LAN9118_END_DEVICE	device handler 	

output: void 

**********************************************************************************/
LOCAL void Lan_SignalSoftwareInterrupt (PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl != NULL);
	SMSC_ASSERT(pDrvCtrl->dwLanBase != 0);

	pDrvCtrl->SoftwareInterruptSignal = FALSE;
	Lan_EnableInterrupt (pDrvCtrl, INT_EN_SW_INT_EN_);
}

/*********************************************************************************
Function Name:  Lan_HandleSoftwareInterrupt

Description :   

input: 		LAN9118_END_DEVICE	device handler 	

output: void 

**********************************************************************************/
LOCAL void Lan_HandleSoftwareInterrupt(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwIntSts)
{
	SMSC_ASSERT(pDrvCtrl != NULL);
	pDrvCtrl->SoftwareInterruptSignal = TRUE;
	if (pDrvCtrl->MulticastUpdatePending) {
        
        /* Defer the handler to the task level */
        netJobAdd ((FUNCPTR) Rx_BeginMulticastUpdate, (int) pDrvCtrl, 0, 0, 0, 0);
	}
}

/*********************************************************************************
Function Name: Lan_udelay

Description :  * FREE_RUN register is at 25MHz  1 count is 1000 nanoseconds / 25 = 40 nsec

input: LAN9118_END_DEVICE, uSec

output: void 

**********************************************************************************/
LOCAL void Lan_udelay (PLAN9118_END_DEV pDrvCtrl, UINT32 uSec)
{
	volatile UINT32 startTime, currentTime, deltaTime;

	startTime=Lan_GetRegDW(FREE_RUN);
	deltaTime=uSec * 25;
	while (1) {
		currentTime=Lan_GetRegDW(FREE_RUN);
		if (currentTime-startTime >= deltaTime)
			break;
	}
}

/*********************************************************************************
Function Name: Lan_ShowMacRegs

Description :  print MAC registers

input: LAN9118_END_DEV pointer

output: N?A 

/*************************************************************/
LOCAL void Lan_ShowMacRegs (PLAN9118_END_DEV pDrvCtrl)
{
	static const SHOW_REG macCsr[] = {
		{ "MAC_CR",		MAC_CR		},
		{ "ADDRH",		ADDRH		},
		{ "ADDRL",		ADDRL		},
		{ "HASHH",		HASHH		},
		{ "HASHL",		HASHL		},
		{ "MII_ACC",	MII_ACC		},
		{ "MII_DATA",	MII_DATA	},
		{ "FLOW",		FLOW		},
		{ "VLAN1",		VLAN1		},
		{ "VLAN2",		VLAN2		},
		{ "WUFF",		WUFF		},
		{ "WUCSR",		WUCSR		},
		{ "COE_CR",     COE_CR      }};

	int i, iNumMacRegs = (int)(sizeof(macCsr) / sizeof(SHOW_REG));
	UINT32 dwChipId;
	UINT32 dwIntFlags=0;
	VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

	/* for the devices that do not support COE */
	dwChipId = (pDrvCtrl->dwIdRev & 0xFFFF0000);
	if ((dwChipId != 0x92100000UL) &&
	 	(dwChipId != 0x92110000UL) &&
	 	(dwChipId != 0x92200000UL) &&
	 	(dwChipId != 0x92210000UL))
	{
		iNumMacRegs--;
	}
	
	for (i=0; i<iNumMacRegs; i++)
	{
		logMsg("%8s (0x%02lX) = 0x%08lX\n",
				(UINT32)macCsr[i].szName, 
				macCsr[i].dwOffset, 
				Mac_GetRegDW(pDrvCtrl, macCsr[i].dwOffset,keyCode),
				4, 5, 6);
	}

	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Lan_ShowSysRegs

Description :  Print LAN chip registers 

input: 	LAN9118_END_DEV pointer	

output:  N/A 

**********************************************************************************/
LOCAL void Lan_ShowSysRegs (PLAN9118_END_DEV pDrvCtrl)
{
	static const SHOW_REG sysCsr[] = {
		{ "ID_REV",			0x50UL		},
		{ "INT_CFG",		0x54UL		},
		{ "INT_STS",		0x58UL		},
		{ "INT_EN",			0x5CUL		},
		{ "DMA_CFG",		0x60UL		},
		{ "BYTE_TEST",		0x64UL		},
		{ "FIFO_INT",		0x68UL		},
		{ "RX_CFG",			0x6CUL		},
		{ "TX_CFG",			0x70UL		},
		{ "HW_CFG",			0x74UL		},
		{ "RX_DP_CTRL",		0x78UL		},
		{ "RX_FIFO_INF",	0x7CUL		},
		{ "TX_FIFO_INF",	0x80UL		},
		{ "PMT_CTRL",		0x84UL		},
		{ "GPIO_CFG",		0x88UL		},
		{ "GPT_CFG",		0x8CUL		},
		{ "GPT_CNT",		0x90UL		},
		{ "FPGA_REV",		0x94UL		},
		{ "ENDIAN",			0x98UL		},
		{ "FREE_RUN",		0x9CUL		},
		{ "RX_DROP",		0xA0UL		},
		{ "MAC_CSR_CMD",	0xA4UL		},
		{ "MAC_CSR_DATA",	0xA8UL		},
		{ "AFC_CFG",		0xACUL		},
		{ "E2P_CMD",		0xB0UL		},
		{ "E2P_DATA",		0xB4UL		},
		{ "TEST_REG_A",		0xC0UL		}};

	int i, iNumSysRegs;
	UINT32 dwOldMacCmdReg, dwOldMacDataReg;

	iNumSysRegs = (int)(sizeof(sysCsr) / sizeof(SHOW_REG));
	for (i=0; i<iNumSysRegs; i++)
	{

		logMsg("%12s (0x%02lX) = 0x%08lX\n", 
				(UINT32)sysCsr[i].szName, 
				sysCsr[i].dwOffset, 
				/* *((volatile UINT32 *)(pDrvCtrl->dwLanBase+sysCsr[i].dwOffset)), */
				Lan_GetRegDW(sysCsr[i].dwOffset),
				4, 5, 6);

	}
}

/*********************************************************************************
Function Name:  Lan_ShowPhyRegs

Description :  Print PHY registers 

input: 	LAN9118_END_DEV pointer	

output:  N/A 

**********************************************************************************/
LOCAL void Lan_ShowPhyRegs (PLAN9118_END_DEV pDrvCtrl)
{
	static const SHOW_REG phyCsr[] = {
		{ "BCR",			PHY_BCR				}, 	/* 0 */
		{ "BSR",			PHY_BSR				}, 	/* 1 */
		{ "ID1",			PHY_ID_1			},	/* 2 */
		{ "ID2",			PHY_ID_2			},	/* 3 */
		{ "ANEG_ADV",		PHY_ANEG_ADV		},	/* 4 */
		{ "ANEG_LPA",		PHY_ANEG_LPA		},	/* 5 */
		{ "ANEG_EXP",		6					},	/* 6 */
		{ "MODE_CTRL_STS",	PHY_MODE_CTRL_STS	},	/* 17 */
		{ "CTRL_STS_IND",	27					},	/* 27 */
		{ "INT_SRC",		PHY_INT_SRC			},	/* 29 */
		{ "INT_MASK",		PHY_INT_MASK		},	/* 30 */
		{ "SPECIAL",		PHY_SPECIAL			}};	/* 31 */

	int i, iNumPhyRegs = (int)(sizeof(phyCsr) / sizeof(SHOW_REG));

	UINT32 dwIntFlags=0;
	VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

	for (i=0; i<iNumPhyRegs; i++)
	{

		logMsg("%14s (%2d) = 0x%04X\n", 
				(UINT32)phyCsr[i].szName, 
				phyCsr[i].dwOffset, 
				Phy_GetRegW(pDrvCtrl, phyCsr[i].dwOffset,keyCode),
				4, 5, 6);
	}
	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Led_TestPoint

Description :   utility to trace the software by leds

input: 		device handler, bit set pattern

output: void 

**********************************************************************************/
LOCAL void Led_TestPoint
	(
    	PLAN9118_END_DEV pDrvCtrl,
		UINT8	TestPointNum
	)
{
	UINT32 temp;

	if (!(LAN9118Debug_mode & LAN9118_DEBUG_GPIO))
	{
		printf("Led_TestPoint: GPIO debug not enabled!!!\n");
		return;		
	}

	if (TestPointNum & ~GPIOs)
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_GPIO, "Led_TestPoint: Invalid GPIO/GPO pin selected. TestPointNum = \n", TestPointNum, 2, 3, 4, 5, 6);
#endif
		return;		
	}

	g_GpioSetting &= (~GPIOs);
	g_GpioSetting |= TestPointNum;
	Lan_SetRegDW (GPIO_CFG, g_GpioSetting );
}


/*********************************************************************************
Function Name:  Lan_SetGPIO

Description :   This function will program the GPIO_CFG 0-4 register bits like this:
				- those bits set in the GpiosToSet parameter will be set
				- those bits clear in the GpiosToSet parameter will be left unchanged

input: 		device handler, bit set pattern

output: void 

**********************************************************************************/
LOCAL void Lan_SetGPIO 
	(
    	PLAN9118_END_DEV pDrvCtrl,
		UINT8	GpiosToSet
	)
{
	UINT32 temp;

	if (!(LAN9118Debug_mode & LAN9118_DEBUG_GPIO))
	{
		printf("Lan_SetGPIO: GPIO debug not enabled!!!\n");
		return;		
	}

	if (GpiosToSet & ~GPIOs)
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_GPIO, "Lan_SetGPIO: Invalid GPIO/GPO pin selected. GpiosToSet = \n", GpiosToSet, 2, 3, 4, 5, 6);
#endif
		return;		
	}

	g_GpioSetting |= GpiosToSet;
	Lan_SetRegDW (GPIO_CFG, g_GpioSetting );
}

/*********************************************************************************
Function Name:  Lan_ClrGPIO

Description :   This function will clear the GPIO_CFG 0-4 register bits like this:
				- those bits set in the GpiosToClr parameter will be cleared
				- those bits clear in the GpiosToClr parameter will be left unchanged

input: 		device handler 

output: void 

**********************************************************************************/
LOCAL void Lan_ClrGPIO 
	(
    	PLAN9118_END_DEV pDrvCtrl,
		UINT8	GpiosToClr
	)
{
	UINT32 temp;

	if (!(LAN9118Debug_mode&LAN9118_DEBUG_GPIO))
	{
		printf("Lan_ClrGPIO: GPIO debug not enabled!!!\n");
		return;		
	}

	if (GpiosToClr & ~GPIOs)
	{
#ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_GPIO, "Lan_ClrGPIO: Invalid GPIO/GPO pin selected. GpiosToClr = \n", GpiosToClr, 2, 3, 4, 5, 6);
#endif
		return;		
	}

	g_GpioSetting &= (~GpiosToClr);
	Lan_SetRegDW (GPIO_CFG, g_GpioSetting );
}

/*********************************************************************************
Function Name:  Tx_Initialize

Description :   Initialize the Tramit related parameters

input: 		device handler 

output: void 

**********************************************************************************/
#define TX_FIFO_LOW_THRESHOLD	(1600)

LOCAL void Tx_Initialize(
    	PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 dwRegVal=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->dwLanBase!=0);
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_TX, "\n\nTx_Initialize\n\n", 1, 2, 3, 4, 5, 6);
#endif
	dwRegVal=Lan_GetRegDW(HW_CFG);
	dwRegVal&=(HW_CFG_TX_FIF_SZ_|0x00000FFFUL);
	dwRegVal|=HW_CFG_MBO_;
	Lan_SetRegDW(HW_CFG,dwRegVal);

	/*Set TX COE*/
	if(pDrvCtrl->UseTxCsum)
 	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		UINT32 dwCoeCr=Mac_GetRegDW(pDrvCtrl,COE_CR,keyCode);
		
		dwCoeCr|=(TX_COE_EN);
		Mac_SetRegDW(pDrvCtrl,COE_CR,dwCoeCr,keyCode);
		
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}	

	Lan_SetTDFL(pDrvCtrl,0xFF);
	Lan_EnableInterrupt(pDrvCtrl,INT_EN_TDFA_EN_);


	if(pDrvCtrl->dwTxDmaCh>=TRANSFER_PIO) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_TX, "\nTx will use PIO\n", 1, 2, 3, 4, 5, 6);
#endif
	} else {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_TX, "Tx will use DMA channel %d\n",pDrvCtrl->dwTxDmaCh, 2, 3, 4, 5, 6);
#endif
		SMSC_ASSERT(Platform_IsValidDmaChannel(pDrvCtrl->dwTxDmaCh));
		if(!Platform_DmaInitialize(pDrvCtrl, pDrvCtrl->dwTxDmaCh))
		{
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_TX, "Failed Platform_DmaInitialize, dwTxDmaCh=%u\n", pDrvCtrl->dwTxDmaCh, 2, 3, 4, 5, 6);
#endif
		}
		pDrvCtrl->TxDmaXfer.dwLanReg=pDrvCtrl->dwLanBase+TX_DATA_FIFO;
		pDrvCtrl->TxDmaXfer.pdwBuf=NULL;/*this will be reset per dma request */
		pDrvCtrl->TxDmaXfer.dwDmaCh=pDrvCtrl->dwTxDmaCh;
		pDrvCtrl->TxDmaXfer.dwDwCnt=0;/*this will be reset per dma request */
		pDrvCtrl->TxDmaXfer.fMemWr=FALSE;
	}

	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		UINT32 dwMacCr=Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);

		dwMacCr|=(MAC_CR_TXEN_|MAC_CR_HBDIS_);
		Mac_SetRegDW(pDrvCtrl,MAC_CR,dwMacCr,keyCode);
		Lan_SetRegDW(TX_CFG,TX_CFG_TX_ON_);

		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}

	pDrvCtrl->dwTxQueueDisableMask=0;
	pDrvCtrl->TxInitialized=TRUE;
}

/*********************************************************************************
Function Name:  Tx_HandleInterrupt

Description :   Interrupt Handler for Transmit Related interrupts

input: 		device Handler, Interruptstatus data

output: void 

**********************************************************************************/
LOCAL void Tx_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);

	Lan_SetTDFL(pDrvCtrl,0xFF);
	Tx_WakeQueue(pDrvCtrl,0x02UL);
	{
		UINT32 flag = intLock();
		Lan_SetRegDW(INT_STS,INT_STS_TDFA_);
		Lan_EnableInterrupt(pDrvCtrl,INT_EN_TDFA_EN_);
		intUnlock(flag);
	}
}

#ifdef SCATTER_GATHER_EN
/*******************************************************************************
*
* Tx_MblkWalk - walk the mBlk
*
* This routine walks the mBlk whose address is in <pMblk>, computes
* the number of fragments and packet length it is made of, and returns it in the parameter
* <pFragNum> and <pPktLen>.
*
* RETURNS: void.
*/

LOCAL void Tx_MblkWalk
    (
    M_BLK * pMblk,          /* pointer to the mBlk */
    int *   pFragNum,       /* number of fragments in this mBlk */
    int *   pPktLen         /* Packet Length */
    )
{
    M_BLK * pCurr = pMblk;  /* the current mBlk */
    
    (*pPktLen) = 0; /* zero out the length */

    #ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_TX, "Tx_MblkWalk: start\n", 1, 2, 3, 4, 5, 6);
    #endif

    /* walk this mBlk */
    do
    {
        if (pCurr->mBlkHdr.mLen)
        {
        	/* keep track of the number of fragments in this mBlk */
        	(*pFragNum)++;

        	/* add up length of the fragment */
			(*pPktLen) += pCurr->mBlkHdr.mLen;
			#ifdef TX_LED_TRACE_EN
			if ((*pFragNum) & 1) {
				Led_TestPoint(saved_pDrvCtrl, 3);
			} else {
				Led_TestPoint(saved_pDrvCtrl, 5);
			}
			#endif
		}
		
        #ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_TX, "Tx_MblkWalk: Frg%d (%d)\n", (*pFragNum), pCurr->mBlkHdr.mLen, 3, 4, 5, 6);
        #endif

        pCurr = ((M_BLK *)pCurr->mBlkHdr.mNext);
        #ifdef VXWORKS_DEBUG
        DRV_LOG (LAN9118_DEBUG_TX, "Tx_MblkWalk: next\n", 1, 2, 3, 4, 5, 6);
        #endif
    } while (pCurr != (M_BLK *)NULL);
}


/*******************************************************************************
*
* Tx_SendFragBuf - fragmented packet transmit
*
* This routine transmits the packet described by the given parameters
* over the network
*
* RETURNS: OK, or ERROR if no resources were available or condition
* does not meet.
*/

LOCAL STATUS Tx_SendFragBuf
    (
    PLAN9118_END_DEV pDrvCtrl,    
    M_BLK*		pMblk
    )
{
    int     status = OK; 
    int     txLen = 0;
    UINT32  dwFreeSpace=0;
    int     FragNum = 0;       /* number of fragments in this mBlk */
	UINT32 dwTxCmdA=0;
	UINT32 dwTxCmdB=0;
    
    SMSC_ASSERT(pDrvCtrl!=NULL);
    SMSC_ASSERT(pDrvCtrl->dwLanBase!=0);

    #ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_TX, "Tx_SendFragBuf Tx ...\n", 1, 2, 3, 4, 5, 6);
    #endif

    /* figure out number of fragments and packet length */
    Tx_MblkWalk (pMblk, &FragNum, &txLen);
    #ifdef VXWORKS_DEBUG
    DRV_LOG (LAN9118_DEBUG_TX, "Tx_SendFragBuf: #Frg=%d BufLen=%d\n", FragNum, txLen, 3, 4, 5, 6);
    #endif

	#ifdef TX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 4);
	#endif
	
	dwFreeSpace=Lan_GetRegDW(TX_FIFO_INF);
	dwFreeSpace&=TX_FIFO_INF_TDFREE_;
	if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
	#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_TX, "Tx Data Fifo Low, space available = %d\n",dwFreeSpace, 2, 3, 4, 5, 6);
	#endif
	}

	if(pDrvCtrl->dwTxDmaCh>=TRANSFER_PIO) {
		/*Use PIO only */

		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 5);
		#endif
        /* See if we need to tx fragmented packet */
        if (FragNum == 1) { /* No fragmentation */
			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 6);
			#endif
			dwTxCmdA=
			((((UINT32)(pMblk->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
			TX_CMD_A_INT_FIRST_SEG_ | TX_CMD_A_INT_LAST_SEG_ | 
			((UINT32)(txLen));
			
			dwTxCmdB=
			(((UINT32)(txLen))<<16) |
			((UINT32)(txLen));

			#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
				DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
				DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (PIO) = %ld\n",((((UINT32)(txLen))+3+(((UINT32)(pMblk->mBlkHdr.mData))&0x03UL))>>2)<<2, 2, 3, 4, 5, 6);
			#endif
			
			SET_GPIO(GPIO_TX_COPY);
			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
			Platform_WriteFifo(
				pDrvCtrl->dwLanBase,
				(UINT32 *)(((UINT32)(pMblk->mBlkHdr.mData))&0xFFFFFFFCUL),
				(((UINT32)(txLen))+3+
				(((UINT32)(pMblk->mBlkHdr.mData))&0x03UL))>>2);

			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 7);
			#endif
        } else {            /* Multiple fragments */
       	    M_BLK * pCurr = pMblk;  /* the current mBlk */
       	    
			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 7);
			#endif
       	    do
       	    {
		        /* if first buffer in the packet */
        		if (pCurr == pMblk) {
					dwTxCmdA=
					((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
					TX_CMD_A_INT_FIRST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen));
			
		        /* if last buffer in the packet */
        		} else if (pCurr->mBlkHdr.mNext == (M_BLK *)NULL) {
					dwTxCmdA=
					((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
					TX_CMD_A_INT_LAST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen));
				} else {
					dwTxCmdA=
					((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
					((UINT32)(pCurr->mBlkHdr.mLen));
				}
				
				dwTxCmdB=
				(((UINT32)(txLen))<<16) |
				((UINT32)(txLen));
		        
				#ifdef VXWORKS_DEBUG
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (PIO) = %ld\n",((((UINT32)(pCurr->mBlkHdr.mLen))+3+(((UINT32)(pCurr->mBlkHdr.mData))&0x03UL))>>2)<<2, 2, 3, 4, 5, 6);
				#endif

				/* write the segment into Tx FIFO */
				SET_GPIO(GPIO_TX_COPY);
				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
				Platform_WriteFifo(
					pDrvCtrl->dwLanBase,
					(UINT32 *)(((UINT32)(pCurr->mBlkHdr.mData))&0xFFFFFFFCUL),
					(((UINT32)(pCurr->mBlkHdr.mLen))+3+
					(((UINT32)(pCurr->mBlkHdr.mData))&0x03UL))>>2);

        		pCurr = ((M_BLK *)pCurr->mBlkHdr.mNext);
       	    } while (pCurr != (M_BLK *)NULL);
			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 8);
			#endif
        }
    } else {
		/*Use DMA and PIO */
		UINT32 dwDmaCh=pDrvCtrl->dwTxDmaCh;
		SMSC_ASSERT(TX_FIFO_LOW_THRESHOLD>(txLen+32));

		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 6);
		#endif
 		if(txLen>=pDrvCtrl->dwTxDmaThreshold) {
			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 7);
			#endif
			/*use DMA */
	        if (FragNum == 1) { /* No fragmentation */
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 8);
				#endif
				dwTxCmdA=
				#if (PLATFORM_CACHE_LINE_BYTES == 16)
				(0x01UL<<24)|/*16 byte end alignment */
				#endif
				#if (PLATFORM_CACHE_LINE_BYTES == 32)
				(0x02UL<<24)|/*32 byte end alignment */
				#endif
				((((UINT32)(pMblk->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1))<<16) |/*16 Byte start alignment */
				TX_CMD_A_INT_FIRST_SEG_ |
				TX_CMD_A_INT_LAST_SEG_ |
				((UINT32)(txLen));/*buffer length */

				dwTxCmdB=(((UINT32)(txLen))<<16)|((UINT32)(txLen)&0x7FFUL);

				pDrvCtrl->TxDmaXfer.pdwBuf= (UINT32 *)(((UINT32)(pMblk->mBlkHdr.mData))&
											(~(PLATFORM_CACHE_LINE_BYTES-1)));

				pDrvCtrl->TxDmaXfer.dwDwCnt=((((UINT32)(txLen))+(PLATFORM_CACHE_LINE_BYTES-1)+
											(((UINT32)(pMblk->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1)))&
											(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;

				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 9);
				#endif
				cacheFlush (DATA_CACHE, pDrvCtrl->TxDmaXfer.pdwBuf, (pDrvCtrl->TxDmaXfer.dwDwCnt)<<2);

				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 10);
				#endif
				#ifdef VXWORKS_DEBUG
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (DMA) = %ld\n",((pDrvCtrl->TxDmaXfer.dwDwCnt)<<2), 2, 3, 4, 5, 6);
				#endif

				/* Now really start the transfer */
				SET_GPIO(GPIO_TX_COPY);	

				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
				
				if(!Platform_DmaStartXfer(pDrvCtrl,&(pDrvCtrl->TxDmaXfer)))
				{
				#ifdef VXWORKS_DEBUG
					DRV_LOG (LAN9118_DEBUG_TX, "Failed Platform_DmaStartXfer\n", 1, 2, 3, 4, 5, 6);
				#endif
				}
				
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 11);
				#endif
				/* Since the MBlk will be released when it returns from this function, we */
				/* have to wait here till DMA complete */
				Platform_DmaComplete(pDrvCtrl,dwDmaCh);

				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 12);
				#endif
			} else {            /* Multiple fragments */
	       	    M_BLK * pCurr = pMblk;  /* the current mBlk */
	       	    
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 9);
				#endif
	       	    do
	       	    {
					dwTxCmdA=
					#if (PLATFORM_CACHE_LINE_BYTES == 16)
					(0x01UL<<24);/*16 byte end alignment */
					#endif
					#if (PLATFORM_CACHE_LINE_BYTES == 32)
					(0x02UL<<24);/*32 byte end alignment */
					#endif
					
			        /* if first buffer in the packet */
	        		if (pCurr == pMblk) {
						dwTxCmdA|=
						(((((UINT32)(pCurr->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1))<<16) | /*UINT32 alignment adjustment */
						TX_CMD_A_INT_FIRST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen)));
				
			        /* if last buffer in the packet */
	        		} else if (pCurr->mBlkHdr.mNext == (M_BLK *)NULL) {
						dwTxCmdA|=
						(((((UINT32)(pCurr->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1))<<16) | /*UINT32 alignment adjustment */
						TX_CMD_A_INT_LAST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen)));
					} else {
						dwTxCmdA|=
						(((((UINT32)(pCurr->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1))<<16) | /*UINT32 alignment adjustment */
						((UINT32)(pCurr->mBlkHdr.mLen)));
					}
					
					dwTxCmdB=(((UINT32)(txLen))<<16)|((UINT32)(txLen)&0x7FFUL);

					pDrvCtrl->TxDmaXfer.pdwBuf= (UINT32 *)(((UINT32)(pCurr->mBlkHdr.mData))&
												(~(PLATFORM_CACHE_LINE_BYTES-1)));

					pDrvCtrl->TxDmaXfer.dwDwCnt=((((UINT32)(pCurr->mBlkHdr.mLen))+(PLATFORM_CACHE_LINE_BYTES-1)+
												(((UINT32)(pCurr->mBlkHdr.mData))&(PLATFORM_CACHE_LINE_BYTES-1)))&
												(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;

					cacheFlush (DATA_CACHE, pDrvCtrl->TxDmaXfer.pdwBuf, (pDrvCtrl->TxDmaXfer.dwDwCnt)<<2);
					#ifdef TX_LED_TRACE_EN
					Led_TestPoint(pDrvCtrl, 10);
					#endif

					#ifdef VXWORKS_DEBUG
						DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
						DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
						DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (DMA) = %ld\n",((pDrvCtrl->TxDmaXfer.dwDwCnt)<<2), 2, 3, 4, 5, 6);
					#endif
					
					/* Now really start the transfer */
					SET_GPIO(GPIO_TX_COPY);	

					Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
					Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
				
					if(!Platform_DmaStartXfer(pDrvCtrl,&(pDrvCtrl->TxDmaXfer)))
					{
					#ifdef VXWORKS_DEBUG
						DRV_LOG (LAN9118_DEBUG_TX, "Failed Platform_DmaStartXfer\n", 1, 2, 3, 4, 5, 6);
					#endif
					}
				
					#ifdef TX_LED_TRACE_EN
					Led_TestPoint(pDrvCtrl, 11);
					#endif
					/* Since the MBlk will be released when it returns from this function, we */
					/* have to wait here till DMA complete */
					Platform_DmaComplete(pDrvCtrl,dwDmaCh);

					#ifdef TX_LED_TRACE_EN
					Led_TestPoint(pDrvCtrl, 12);
					#endif
	        		pCurr = ((M_BLK *)pCurr->mBlkHdr.mNext);
	       	    } while (pCurr != (M_BLK *)NULL);
			}
			
		} else {
			/*use PIO */

			#ifdef TX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 8);
			#endif
	        /* See if we need fragmented packet */
	        if (FragNum == 1) { /* No fragmentation */
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 9);
				#endif
				dwTxCmdA=
				((((UINT32)(pMblk->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
				TX_CMD_A_INT_FIRST_SEG_ | TX_CMD_A_INT_LAST_SEG_ | 
				((UINT32)(txLen));
				
				dwTxCmdB=
				(((UINT32)(txLen))<<16) |
				((UINT32)(txLen));

				#ifdef VXWORKS_DEBUG
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (PIO) = %ld\n",((((UINT32)(txLen))+3+(((UINT32)(pMblk->mBlkHdr.mData))&0x03UL))>>2)<<2, 2, 3, 4, 5, 6);
				#endif

				SET_GPIO(GPIO_TX_COPY);
				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
				Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
				Platform_WriteFifo(
					pDrvCtrl->dwLanBase,
					(UINT32 *)(((UINT32)(pMblk->mBlkHdr.mData))&0xFFFFFFFCUL),
					(((UINT32)(txLen))+3+
					(((UINT32)(pMblk->mBlkHdr.mData))&0x03UL))>>2);

				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 10);
				#endif
	        } else {            /* Multiple fragments */
	       	    M_BLK * pCurr = pMblk;  /* the current mBlk */
	       	    
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 10);
				#endif
	       	    do
	       	    {
			        /* if first buffer in a packet */
	        		if (pCurr == pMblk) {
						dwTxCmdA=
						((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
						TX_CMD_A_INT_FIRST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen));
				
			        /* if last buffer in a packet */
	        		} else if (pCurr->mBlkHdr.mNext == (M_BLK *)NULL) {
						dwTxCmdA=
						((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
						TX_CMD_A_INT_LAST_SEG_ | ((UINT32)(pCurr->mBlkHdr.mLen));
					} else {
						dwTxCmdA=
						((((UINT32)(pCurr->mBlkHdr.mData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
						((UINT32)(pCurr->mBlkHdr.mLen));
					}
					
					dwTxCmdB=
					(((UINT32)(txLen))<<16) |
					((UINT32)(txLen));
			        
					#ifdef VXWORKS_DEBUG
						DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdA = 0x%08lX\n",dwTxCmdA, 2, 3, 4, 5, 6);
						DRV_LOG (LAN9118_DEBUG_TX, "dwTxCmdB = 0x%08lX\n",dwTxCmdB, 2, 3, 4, 5, 6);
						DRV_LOG (LAN9118_DEBUG_TX, "PacketLength (PIO) = %ld\n",((((UINT32)(pCurr->mBlkHdr.mLen))+3+(((UINT32)(pCurr->mBlkHdr.mData))&0x03UL))>>2)<<2, 2, 3, 4, 5, 6);
					#endif

					/* write the segment into Tx FIFO */
					SET_GPIO(GPIO_TX_COPY);
					Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
					Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
					Platform_WriteFifo(
						pDrvCtrl->dwLanBase,
						(UINT32 *)(((UINT32)(pCurr->mBlkHdr.mData))&0xFFFFFFFCUL),
						(((UINT32)(pCurr->mBlkHdr.mLen))+3+
						(((UINT32)(pCurr->mBlkHdr.mData))&0x03UL))>>2);

	        		pCurr = ((M_BLK *)pCurr->mBlkHdr.mNext);
	       	    } while (pCurr != (M_BLK *)NULL);
				#ifdef TX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 11);
				#endif
	        }
		}
    }

	dwFreeSpace-=(txLen+32);
	CLR_GPIO(GPIO_TX_COPY);

	if(Tx_GetTxStatusCount(pDrvCtrl)>=30)
	{
		Tx_UpdateTxCounters(pDrvCtrl);
	}
	
	#ifdef TX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 13);
	#endif
	if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		Tx_StopQueue(pDrvCtrl,0x02UL);
		Lan_SetTDFL(pDrvCtrl,0x32);
	}
	
    return (status);
}

#else /* SCATTER_GATHER_EN */
/*******************************************************************************
*
* Tx_SendBuf - send data over the NIC network interface
*
* This routine transfers data to the NIC device via the remote DMA, and
* then signal for a transmission.
*
* RETURNS: OK, or ERROR if the transmitter signalled an error.
*/
LOCAL STATUS Tx_SendBuf
    (
    PLAN9118_END_DEV pDrvCtrl,    
    M_BLK*		pMblk
    )
{
    int			status = OK; 
    int			txLen = 0;
    UINT32 		dwFreeSpace=0;
    
    SMSC_ASSERT(pDrvCtrl!=NULL);
    SMSC_ASSERT(pDrvCtrl->dwLanBase!=0);

	if(pDrvCtrl->dwTxDmaCh>=TRANSFER_PIO) 
	{
		/*Use PIO only */
		UINT32 dwTxCmdA=0;
		UINT32 dwTxCmdB=0;

		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 3);
		#endif

		txLen = netMblkToBufCopy (pMblk, pDrvCtrl->txData, NULL);
    	txLen = max (txLen, ETHERSMALL);

		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 4);
		#endif

		dwFreeSpace=Lan_GetRegDW(TX_FIFO_INF);
		dwFreeSpace&=TX_FIFO_INF_TDFREE_;
		if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_TX, "Tx Data Fifo Low, space available = %d\n",dwFreeSpace, 2, 3, 4, 5, 6);
#endif
		}
		
		dwTxCmdA=
			((((UINT32)(pDrvCtrl->txData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
			TX_CMD_A_INT_FIRST_SEG_ | TX_CMD_A_INT_LAST_SEG_ | 
			((UINT32)(txLen));
			
		dwTxCmdB=
			(((UINT32)(txLen))<<16) |
			((UINT32)(txLen));

		SET_GPIO(GPIO_TX_COPY);
		Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
		Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
		Platform_WriteFifo(
			pDrvCtrl->dwLanBase,
			(UINT32 *)(((UINT32)(pDrvCtrl->txData))&0xFFFFFFFCUL),
			(((UINT32)(txLen))+3+
			(((UINT32)(pDrvCtrl->txData))&0x03UL))>>2);
		dwFreeSpace-=(txLen+32);
		CLR_GPIO(GPIO_TX_COPY);
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 5);
		#endif
	} 
	else 
	{
		/*Use DMA and PIO */
		UINT32 dwDmaCh=pDrvCtrl->dwTxDmaCh;
		SMSC_ASSERT(TX_FIFO_LOW_THRESHOLD>(txLen+32));

		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 4);
		#endif
	    /* Get exclusive access to Fifo port for Tx Dma */

		/*
		 * FIXFIX: Note that we are locking for more time than we really need here
		 * due to the fact that we needed to get the pkt len to do the conditional
		 * branch based on it and the routine that gives the length also does a copy
		 * which would destroy the DMA buffer even if the DMA is not done.
		 *
		 * TBD: explore a non-copy way of getting the pkt len so that I can safely move 
		 * this locking / check for DMa complete down further after the cache flushing 
		 * for the DMA case and before the PIO transfer for the PIO case.
		 */
		Platform_DmaComplete(pDrvCtrl,dwDmaCh);
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 5);
		#endif
		txLen = netMblkToBufCopy (pMblk, pDrvCtrl->txData, NULL);
	    txLen = max (txLen, ETHERSMALL);
		#ifdef TX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 6);
		#endif

		if(txLen>=pDrvCtrl->dwTxDmaThreshold) 
		{
			/*use DMA */
			UINT32 dwTxCmdA;
			UINT32 dwTxCmdB;
			/*prepare for 16 byte alignment */
			
			dwTxCmdA=
#if (PLATFORM_CACHE_LINE_BYTES == 16)
				(0x01UL<<24)|/*16 byte end alignment */
#endif
#if (PLATFORM_CACHE_LINE_BYTES == 32)
				(0x02UL<<24)|/*32 byte end alignment */
#endif
				((((UINT32)(pDrvCtrl->txData))&(PLATFORM_CACHE_LINE_BYTES-1))<<16) |/*16 Byte start alignment */
				TX_CMD_A_INT_FIRST_SEG_ |
				TX_CMD_A_INT_LAST_SEG_ |
				((UINT32)(txLen));/*buffer length */

			dwTxCmdB=(((UINT32)(txLen))<<16)|((UINT32)(txLen)&0x7FFUL);

			pDrvCtrl->TxDmaXfer.pdwBuf= (UINT32 *)(((UINT32)(pDrvCtrl->txData))&
										(~(PLATFORM_CACHE_LINE_BYTES-1)));

			pDrvCtrl->TxDmaXfer.dwDwCnt=((((UINT32)(txLen))+(PLATFORM_CACHE_LINE_BYTES-1)+
										(((UINT32)(pDrvCtrl->txData))&(PLATFORM_CACHE_LINE_BYTES-1)))&
										(~(PLATFORM_CACHE_LINE_BYTES-1)))>>2;

#ifndef UNCASHED_TX_BUF
			cacheFlush (DATA_CACHE, pDrvCtrl->TxDmaXfer.pdwBuf, (pDrvCtrl->TxDmaXfer.dwDwCnt)<<2);
#endif			

			dwFreeSpace=Lan_GetRegDW(TX_FIFO_INF);
			dwFreeSpace&=TX_FIFO_INF_TDFREE_;
			if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_TX, "Tx DATA FIFO LOW, space available = %d\n", dwFreeSpace, 2, 3, 4, 5, 6);
#endif
			}

			/* Now really start the transfer */
			PULSE_GPIO(GPIO_TX_COPY);	/* We don't know when it ends, so don't set, just pulse */

			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
				
			if(!Platform_DmaStartXfer(pDrvCtrl,&(pDrvCtrl->TxDmaXfer)))
			{
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_TX, "Failed Platform_DmaStartXfer\n", 1, 2, 3, 4, 5, 6);
#endif
			}

			dwFreeSpace-=(txLen+32);
		}
		else 
		{
			/*use PIO */
			UINT32 dwTxCmdA=0;
			UINT32 dwTxCmdB=0;

			/* Exclusive access to Fifo port for Pio is now guaranteed (Tx Dma is done) */
			dwFreeSpace=Lan_GetRegDW(TX_FIFO_INF);
			dwFreeSpace&=TX_FIFO_INF_TDFREE_;
			if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
				SMSC_WARNING("Tx DATA FIFO LOW, not enough space available");
			}
			dwTxCmdA=
				((((UINT32)(pDrvCtrl->txData))&0x03UL)<<16) | /*UINT32 alignment adjustment */
				TX_CMD_A_INT_FIRST_SEG_ | TX_CMD_A_INT_LAST_SEG_ | 
				((UINT32)(txLen));
			dwTxCmdB=
				(((UINT32)(txLen))<<16) |
				((UINT32)(txLen));
		
			SET_GPIO(GPIO_TX_COPY);
			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
			Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
			Platform_WriteFifo(
				pDrvCtrl->dwLanBase,
				(UINT32 *)(((UINT32)(pDrvCtrl->txData))&0xFFFFFFFCUL),
				(((UINT32)(txLen))+3+
				(((UINT32)(pDrvCtrl->txData))&0x03UL))>>2);
			CLR_GPIO(GPIO_TX_COPY);

			dwFreeSpace-=(txLen+32);
		}
	}
	
	if(Tx_GetTxStatusCount(pDrvCtrl)>=30)
	{
		Tx_UpdateTxCounters(pDrvCtrl);
	}
	
	if(dwFreeSpace<TX_FIFO_LOW_THRESHOLD) {
		Tx_StopQueue(pDrvCtrl,0x02UL);
		Lan_SetTDFL(pDrvCtrl,0x32);
	}
	
	return (status);
}
#endif /* SCATTER_GATHER_EN */
/*********************************************************************************
Function Name:  Tx_StopQueue

Description :   stops the Transmitter 

input: 		device handler , source 

output: void 

**********************************************************************************/
LOCAL void Tx_StopQueue (PLAN9118_END_DEV pDrvCtrl, UINT32 dwSource)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->TxInitialized);
	if(pDrvCtrl->dwTxQueueDisableMask==0) {
		/* there is not any mux call to stop tx queue */
	}
	pDrvCtrl->dwTxQueueDisableMask|=dwSource;
}

/*********************************************************************************
Function Name:  Tx_WakeQueue

Description :   this function enables the transmit Queue 

input: 		device Handler , source	

output: void

**********************************************************************************/
LOCAL void Tx_WakeQueue (PLAN9118_END_DEV pDrvCtrl, UINT32 dwSource)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->TxInitialized);

	pDrvCtrl->dwTxQueueDisableMask&=(~dwSource);
	if(pDrvCtrl->dwTxQueueDisableMask==0)
	{
		muxTxRestart (&pDrvCtrl->endObj); /* ask mux to restart tx */
	}
}

/*********************************************************************************
Function Name:  Tx_GetTxStatusCount

Description :   get the transmit status count 

input: 		device handler

output: void

**********************************************************************************/
LOCAL UINT32 Tx_GetTxStatusCount (PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->dwLanBase!=0);
	result=Lan_GetRegDW(TX_FIFO_INF);
	if(OLD_REGISTERS(pDrvCtrl)) {
		result&=TX_FIFO_INF_TSFREE_;
		result>>=16;
		if(result>0x80) {
			SMSC_WARNING("TX_FIFO_INF_TSFREE_>0x80");
			result=0x80;
		}
		result=0x80-result;
	} else {
		result&=TX_FIFO_INF_TSUSED_;
		result>>=16;
	}
	return result;
}


/*********************************************************************************
Function Name:  Tx_CompleteTx

input: 		device Handler 

output: void 

**********************************************************************************/
LOCAL UINT32 Tx_CompleteTx (PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->dwLanBase!=0);
	SMSC_ASSERT(pDrvCtrl->TxInitialized==TRUE);

	result=Lan_GetRegDW(TX_FIFO_INF);
	if(OLD_REGISTERS(pDrvCtrl)) {
		result&=TX_FIFO_INF_TSFREE_;
		if(result!=0x00800000UL) {
			result=Lan_GetRegDW(TX_STATUS_FIFO);
		} else {
			result=0;
		}
	} else {
		result&=TX_FIFO_INF_TSUSED_;
		if(result!=0x00000000UL) {
			result=Lan_GetRegDW(TX_STATUS_FIFO);
		} else {
			result=0;
		}
	}
	return result;
}

/*********************************************************************************
Function Name:  Tx_UpdateTxCounters

input: 		Device Handler

output: void 

**********************************************************************************/
LOCAL void Tx_UpdateTxCounters (PLAN9118_END_DEV pDrvCtrl)
{

	UINT32 dwTxStatus=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	while((dwTxStatus=Tx_CompleteTx(pDrvCtrl))!=0)
	{
		if(dwTxStatus&0x80000000UL) {
			SMSC_WARNING("Packet tag reserved bit is high");
			/* pDrvCtrl->stats.tx_errors++; */ 
			END_ADD_IF_OUTERR(pDrvCtrl, 1);
		} else if(dwTxStatus&0x00007080UL) {
			SMSC_WARNING("Tx Status reserved bits are high");
			/* pDrvCtrl->stats.tx_errors++; */
			END_ADD_IF_OUTERR(pDrvCtrl, 1);
		} else {
			if(dwTxStatus&0x00008000UL) {
				/* pDrvCtrl->stats.tx_errors++; */
				END_ADD_IF_OUTERR(pDrvCtrl, 1);
			} else {
				/* it is done in MUX */
				/* pDrvCtrl->stats.tx_packets++; */
				/* pDrvCtrl->stats.tx_bytes+=(dwTxStatus>>16); */
			}
			if(dwTxStatus&0x00000100UL) {
				/* pDrvCtrl->stats.collisions+=16; */
				/* pDrvCtrl->stats.tx_aborted_errors+=1; */
				END_ADD_IF_COL(pDrvCtrl, 16);
			} else {
				/* pDrvCtrl->stats.collisions+= */
				/*	((dwTxStatus>>3)&0xFUL); */ 
				if (((dwTxStatus>>3)&0xFUL))
				{
					END_ADD_IF_COL(pDrvCtrl, ((dwTxStatus>>3)&0xFUL));
				}
			}
			if(dwTxStatus&0x00000800UL) {
				/* pDrvCtrl->stats.tx_carrier_errors+=1; */
			}
			if(dwTxStatus&0x00000200UL) {
				/* pDrvCtrl->stats.collisions++; */
				/* pDrvCtrl->stats.tx_aborted_errors++; */
				END_ADD_IF_COL(pDrvCtrl, 1);
			}
		}
	}
}

/*********************************************************************************
Function Name:  Rx_Initialize

Description :   Initialize the revice related parameter

input: 		Device Handler

output: void 

**********************************************************************************/
LOCAL void Rx_Initialize(
        PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);


	if(pDrvCtrl->dwRxDmaCh>=TRANSFER_PIO) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_RX, "Rx will use PIO\n", 1, 2, 3, 4, 5, 6);
#endif
		Platform_GetFlowControlParameters(pDrvCtrl, &(pDrvCtrl->RxFlowParameters), FALSE);
	} else {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_RX, "Rx will use DMA Channel %d\n", pDrvCtrl->dwRxDmaCh, 2, 3, 4, 5, 6);
#endif
		SMSC_ASSERT(Platform_IsValidDmaChannel(pDrvCtrl->dwRxDmaCh));
		if(!Platform_DmaInitialize(pDrvCtrl, pDrvCtrl->dwRxDmaCh))
		{
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_RX, "Failed Platform_DmaInitialize, dwRxDmaCh=%u\n",pDrvCtrl->dwRxDmaCh, 2, 3, 4, 5, 6);
#endif
		}
		Platform_GetFlowControlParameters(pDrvCtrl, &(pDrvCtrl->RxFlowParameters), TRUE);
	}

	if(pDrvCtrl->max_throughput!=0xFFFFFFFFUL) {
		pDrvCtrl->RxFlowParameters.MaxThroughput=pDrvCtrl->max_throughput;
	}

	if(pDrvCtrl->max_packet_count!=0xFFFFFFFFUL) {
		pDrvCtrl->RxFlowParameters.MaxPacketCount=pDrvCtrl->max_packet_count;
	}

	if(pDrvCtrl->packet_cost!=0xFFFFFFFFUL) {
		pDrvCtrl->RxFlowParameters.PacketCost=pDrvCtrl->packet_cost;
	}

	if(pDrvCtrl->burst_period!=0xFFFFFFFFUL) {
		pDrvCtrl->RxFlowParameters.BurstPeriod=pDrvCtrl->burst_period;
	}

	if(pDrvCtrl->RxFlowParameters.BurstPeriod==0) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_RX, "burst_period of 0 is not allowed\n", 1, 2, 3, 4, 5, 6);
		DRV_LOG (LAN9118_DEBUG_RX, " resetting burst_period to 100\n", 1, 2, 3, 4, 5, 6);
#endif
		pDrvCtrl->RxFlowParameters.BurstPeriod=100;
	}
	
	if(pDrvCtrl->max_work_load!=0xFFFFFFFFUL) {
		pDrvCtrl->RxFlowMaxWorkLoad=pDrvCtrl->max_work_load;
	} else {
		pDrvCtrl->RxFlowMaxWorkLoad=
			pDrvCtrl->RxFlowParameters.MaxThroughput+
			(pDrvCtrl->RxFlowParameters.MaxPacketCount*
			pDrvCtrl->RxFlowParameters.PacketCost);
	}
	
	pDrvCtrl->RxFlowBurstMaxWorkLoad=
			(pDrvCtrl->RxFlowMaxWorkLoad*
			pDrvCtrl->RxFlowParameters.BurstPeriod)/1000;
	
	if(pDrvCtrl->int_deas!=0xFFFFFFFFUL) {
		Lan_SetIntDeas(pDrvCtrl,pDrvCtrl->int_deas);
	} else {
		Lan_SetIntDeas(pDrvCtrl,pDrvCtrl->RxFlowParameters.IntDeas);
	}

	/* Set RX COE */
 	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		Mac_SetRegDW(pDrvCtrl,VLAN1,0x8100,keyCode);
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}

	if (pDrvCtrl->UseRxCsum) {

		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		UINT32 dwCoeCr=Mac_GetRegDW(pDrvCtrl,COE_CR,keyCode);		

		dwCoeCr|=(RX_COE_EN |  RX_COE_MODE);
		Mac_SetRegDW(pDrvCtrl,COE_CR,dwCoeCr,keyCode);

		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}

	/*initially the receiver is off */
	/*  a following link up detection will turn the receiver on */
	pDrvCtrl->dwRxOffCount=1;
	Lan_SetRegDW(RX_CFG,0x00000200UL);
	Rx_ReceiverOn(pDrvCtrl, 0);

	Lan_SetRSFL(pDrvCtrl,0x00); 
	pDrvCtrl->RxInterrupts=INT_EN_RSFL_EN_;
	pDrvCtrl->RxInterrupts|=INT_EN_RXE_EN_;
	pDrvCtrl->RxInterrupts|=INT_EN_RDFO_EN_;
	pDrvCtrl->RxInterrupts|=INT_EN_RXDFH_INT_EN_;
	Lan_EnableInterrupt(pDrvCtrl,pDrvCtrl->RxInterrupts);
	
	pDrvCtrl->FCRxInterruptsEn = TRUE;
}

/**************************************************************************************/
/*function :  Rx_HandleOverrun 
*
*Description :  Set the flag for overrun situation and update counters.
*
* input: device handler
*
* output: void
*
/***************************************************************************************/
LOCAL void Rx_HandleOverrun(PLAN9118_END_DEV pDrvCtrl)
{
	pDrvCtrl->RxOverrunCount++;
}

/*********************************************************************************
Function Name:  Rx_CompleteMulticastUpdate

Description : updates the multicast table. set the HAsh high and low address in the 
				table. and make the revicer on for the multicast  

input: 	LAN9118_END_DEVICE	

output: void 

**********************************************************************************/
LOCAL void Rx_CompleteMulticastUpdate (PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 local_MACCR;
	VL_KEY keyCode=0;
	UINT32 dwIntFlags=0;

	keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
	if (pDrvCtrl->MulticastUpdatePending) {
		Mac_SetRegDW(pDrvCtrl,HASHH,pDrvCtrl->HashHi,keyCode);
		Mac_SetRegDW(pDrvCtrl,HASHL,pDrvCtrl->HashLo,keyCode);
		local_MACCR = Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);
		local_MACCR |= pDrvCtrl->set_bits_mask;
		local_MACCR &= ~(pDrvCtrl->clear_bits_mask);
		Mac_SetRegDW(pDrvCtrl,MAC_CR,local_MACCR,keyCode);
		Rx_ReceiverOn(pDrvCtrl,keyCode);
		pDrvCtrl->MulticastUpdatePending = FALSE;
	}

	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Rx_BeginMulticastUpdate

Description :   

input: 	LAN9118_END_DEVICE	

output: void 

**********************************************************************************/
LOCAL void Rx_BeginMulticastUpdate (PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 startTime, currentTime;
	UINT32 timeout;
	UINT32 dwTemp, flags;

	/* NOTE: we can't rely on privateData->dwLinkSpeed because */
	/* it updates only once per second and may be out dated. */
	Rx_ReceiverOff (pDrvCtrl);
	/* For concord or later there is no
	 overrun processing that might turn off the receiver.
	 there for we can rely on RxStop Int. */

	/* poll for MAC RX to stop */
	timeout = 200UL;
	do
	{
		/* Atomic interrupt status read */
		flags = intLock ();
		dwTemp = Lan_GetRegDW(INT_STS);
		intUnlock (flags);
		/* wait 1 uSec */
		Lan_udelay(pDrvCtrl, 1);
		timeout--;
	}
	while ((timeout)&&(!(dwTemp&INT_STS_RXSTOP_INT_)));

	if(timeout==0) {
		/* this is probably a 10Mb link, therefore prepare
		   interrupt for update later.*/
		Lan_EnableInterrupt(pDrvCtrl,INT_EN_RXSTOP_INT_EN_);

		/* if this is a 10Mbps half duplex connection
		   then Rx stop is only 99.6%  reliable
		   Therefor we must schedule Gpt callback as 
		   back up */

		/* using 18*(100uS) because we already waited 200uS */
		Gpt_ScheduleCallBack(pDrvCtrl,GptCB_RxCompleteMulticast, 18UL);

	} else {
		/* Rx is stopped */
		flags = intLock ();
		Lan_SetRegDW(INT_STS,INT_STS_RXSTOP_INT_);/* clear interrupt signal */
		intUnlock (flags);
		Rx_CompleteMulticastUpdate(pDrvCtrl);
	}
}

/*********************************************************************************
Function Name:  Rx_PopRxStatus

Description :   

input: 	LAN9118_END_DEV

output: UINT32 

**********************************************************************************/
LOCAL UINT32 Rx_PopRxStatus(
			PLAN9118_END_DEV pDrvCtrl, 
			UINT32 dwIntSts)
{
	UINT32 result=Lan_GetRegDW(RX_FIFO_INF); /*retrive the status of the Rx FIFO used for the Data and the status 0x7C*/
	
	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 6);
	#endif
	if((pDrvCtrl->RxCongested==FALSE)|| /* fixme - no pDrvCtrl->RxCongested assignment */
		((pDrvCtrl->RxCongested==TRUE)&&((result&0x00FF0000UL)==0UL)))
	{
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 7);
		#endif
		if(result&0x00FF0000UL) {

			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 8);
			#endif
			if(dwIntSts&INT_STS_RDFO_) {
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 31);
				#endif
				Rx_HandleOverrun(pDrvCtrl);
			}

			#ifdef RX_LED_TRACE_EN
			Led_TestPoint(pDrvCtrl, 9);
			#endif
			if((pDrvCtrl->RxFlowControlActive==FALSE)||
				((pDrvCtrl->RxFlowControlActive==TRUE)&&
				 (pDrvCtrl->RxFlowBurstActive==TRUE)))
			{
				/*Rx status is available, read it */
				result=Lan_GetRegDW(RX_STATUS_FIFO);
				pDrvCtrl->RxStatusDWReadCount++;
				pDrvCtrl->LastRxStatus3=
					pDrvCtrl->LastRxStatus2;
				pDrvCtrl->LastRxStatus2=
					pDrvCtrl->LastRxStatus1;
				pDrvCtrl->LastRxStatus1=result;
	            
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 10);
				#endif
				if(pDrvCtrl->RxOverrun) {
					UINT32 dwPacketLength=((result&0x3FFF0000UL)>>16);
					UINT32 dwByteCount=((dwPacketLength+2+3)&0xFFFFFFFCUL);
					#ifdef RX_LED_TRACE_EN
					Led_TestPoint(pDrvCtrl, 11);
					#endif
					if((pDrvCtrl->RxUnloadProgress+dwByteCount)>=
						((pDrvCtrl->RxMaxDataFifoSize)-16))
					{
						/*This is the packet that crosses the corruption point */
						/*  so just ignore it and complete the overrun processing. */
						result=0;
						#ifdef RX_LED_TRACE_EN
						Led_TestPoint(pDrvCtrl, 31);
						#endif
						goto FINISH_OVERRUN_PROCESSING;
					}
					pDrvCtrl->RxUnloadProgress+=dwByteCount;
					pDrvCtrl->RxUnloadPacketProgress++;
				}
				
				pDrvCtrl->RxFlowCurrentThroughput+=
						((((result&0x3FFF0000UL)>>16)-4UL));
				pDrvCtrl->RxFlowCurrentPacketCount++;
				pDrvCtrl->RxFlowCurrentWorkLoad+=
						((((result&0x3FFF0000UL)>>16)-4UL)+pDrvCtrl->RxFlowParameters.PacketCost);
				if(pDrvCtrl->RxFlowControlActive) {
					#ifdef RX_LED_TRACE_EN
					Led_TestPoint(pDrvCtrl, 12);
					#endif
					pDrvCtrl->RxFlowBurstWorkLoad+=
						((((result&0x3FFF0000UL)>>16)-4UL)+pDrvCtrl->RxFlowParameters.PacketCost);
					if(pDrvCtrl->RxFlowBurstWorkLoad>=
						pDrvCtrl->RxFlowBurstMaxWorkLoad) 
					{
						#ifdef RX_LED_TRACE_EN
						Led_TestPoint(pDrvCtrl, 13);
						#endif
						pDrvCtrl->RxFlowBurstActive=FALSE;
						pDrvCtrl->FCRxInterruptsEn = FALSE;
						Lan_DisableInterrupt(pDrvCtrl,pDrvCtrl->RxInterrupts);
					}
				}
			} else {
				result=0;
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 31);
				#endif
			}
		} 
		else 
		{
				#ifdef RX_LED_TRACE_EN
				Led_TestPoint(pDrvCtrl, 14);
				#endif
FINISH_OVERRUN_PROCESSING: 
			if(pDrvCtrl->RxOverrun) {
				UINT32 timeOut;
				UINT32 temp, dwTemp=0;
				temp=dwIntSts;
				{
					timeOut=2000;
					temp|=Lan_GetRegDW(INT_STS);
					while((timeOut>0)&&(!(temp&INT_STS_RXSTOP_INT_))) {
						temp=Lan_GetRegDW(INT_STS);
						Lan_udelay(pDrvCtrl, 1);
						timeOut--;
					}
					if(timeOut==0) {
						SMSC_WARNING("Timed out waiting for Rx to Stop\n");
					}
					Lan_SetRegDW(INT_STS,INT_STS_RXSTOP_INT_);
				}

				if(pDrvCtrl->dwRxDmaCh<TRANSFER_REQUEST_DMA) {

					/* make sure DMA has stopped before doing RX Dump */
					/* fixme - we don't need to check in scatter-gather mode
					if(privateData->RxSkb) {
						Platform_DmaComplete(
							&(privateData->PlatformData),
							privateData->dwRxDmaCh);
						Rx_HandOffSkb(privateData,privateData->RxSkb);
						privateData->RxSkb=NULL;
					}
					*/
				}

			 	temp=Lan_GetRegDW(RX_CFG);
				Lan_SetRegDW(RX_CFG,(temp&0x3FFFFFFFUL));
				timeOut=10000000;
				Lan_SetBitsDW(RX_CFG,RX_CFG_RX_DUMP_);
				dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
				dwTemp = dwTemp;
				while((timeOut>0)&&(Lan_GetRegDW(RX_CFG)&(RX_CFG_RX_DUMP_))) {
					Lan_udelay(pDrvCtrl, 1);
					timeOut--;
				}
				if(timeOut==0) {
					SMSC_WARNING("Timed out waiting for Rx Dump to complete\n");
				}
				Lan_SetRegDW(RX_CFG,temp);

				pDrvCtrl->RxDumpCount++;
				Rx_ReceiverOn(pDrvCtrl, 0);
				pDrvCtrl->RxOverrun=FALSE;
			}
			result=0;
			pDrvCtrl->LastReasonForReleasingCPU=1;/*Status FIFO Empty */
		}
	} else {
		/*disable and reenable the INT_EN
		  This will allow the deassertion interval to begin */
		UINT32 temp=Lan_GetRegDW(INT_EN);
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
		Lan_SetRegDW(INT_EN,0);
		Lan_SetRegDW(INT_EN,temp);
		result=0;
		pDrvCtrl->LastReasonForReleasingCPU=2;/*High Congestion */
	}
	return result;
}

/*********************************************************************************
Function Name:  Rx_CountErrors

Description :   this functions is counter for the RX error

input: 	LAN9118_END_DEVICE and 	UINT32

output: void 

**********************************************************************************/
LOCAL void Rx_CountErrors(PLAN9118_END_DEV pDrvCtrl,UINT32 dwRxStatus) 
{
	BOOL crcError=FALSE;
	if(dwRxStatus&0x00008000UL) {
		/* pDrvCtrl->stats.rx_errors++; */
		if(dwRxStatus&0x00000002UL) {
			/* pDrvCtrl->stats.rx_crc_errors++; */
			/* END_ADD_IF_INERR(pDrvCtrl, 1); */
			crcError=TRUE;
		}
	}
	if(!crcError) {
		if((dwRxStatus&0x00001020UL)==0x00001020UL) {
			/* Frame type indicates length, and length error is set */
			/* pDrvCtrl->stats.rx_length_errors++;*/
			/* END_ADD_IF_INERR(pDrvCtrl, 1); */
		}
		if(dwRxStatus&RX_STS_MCAST_) {
			/* it is counted in MUX */ 
			/* pDrvCtrl->stats.multicast++; */
		}
	}
}

/*********************************************************************************
Function Name:  Rx_FastForward

Description :   This function reads the data from the receive FIFO if there is a error in the 
				recived data the data in the FIFO has to read and discard from the FIFO 

input: 	LAN9118_END_DEVICE	,UINT32 no of wors ro discard

output: void 

**********************************************************************************/
LOCAL void Rx_FastForward(PLAN9118_END_DEV pDrvCtrl,UINT32 dwDwordCount)
{
	pDrvCtrl->RxFastForwardCount++;
	if(dwDwordCount>=4)
	{
		UINT32 dwTimeOut=500;
		UINT32 dwTemp=0UL; 
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 24);
		#endif
		Lan_SetRegDW(RX_DP_CTRL,(dwDwordCount|RX_DP_CTRL_FFWD_BUSY_));
		dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
		dwTemp = dwTemp;
		while((dwTimeOut)&&(Lan_GetRegDW(RX_DP_CTRL)&
				RX_DP_CTRL_FFWD_BUSY_))
		{
			Lan_udelay(pDrvCtrl, 1);
			dwTimeOut--;
		}
		if(dwTimeOut==0) {

			SMSC_WARNING("timed out waiting for RX FFWD to finish");
		}
	} else {
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 25);
		#endif
		while(dwDwordCount) {
			UINT32 dwTemp=Lan_GetRegDW(RX_DATA_FIFO);
			dwTemp=dwTemp;
			dwDwordCount--;
		}
	}
}

/*********************************************************************************
Function Name:  Rx_ReceiverOff

Description :   Rx_ReceiverOffuse a reference counter because they are used in both the Rx code 
			and the link management count

input: 		LAN9118_END_DEVICE, uSec

output: void 

**********************************************************************************/
LOCAL void Rx_ReceiverOff(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 dwIntFlags=0;
	VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
	
	if(pDrvCtrl->dwRxOffCount==0) {
		UINT32 dwMacCr=Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);
		if(!(dwMacCr&MAC_CR_RXEN_)) {
#ifdef VXWORKS_DEBUG
			DRV_LOG (LAN9118_DEBUG_RX, "Rx_ReceiverOff: Receiver is already Off\n", 1, 2, 3, 4, 5, 6);
#endif
		}
		dwMacCr&=(~MAC_CR_RXEN_);
		Mac_SetRegDW(pDrvCtrl,MAC_CR,dwMacCr,keyCode);
	}
	pDrvCtrl->dwRxOffCount++;
	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Rx_ReceiverOn

Description :   Rx_ReceiverOn use a reference counter because they are used in both the Rx code 
			and the link management count

input: 		LAN9118_END_DEVICE, uSec

output: void 

**********************************************************************************/
LOCAL void Rx_ReceiverOn(PLAN9118_END_DEV pDrvCtrl, VL_KEY callerKeyCode)
{
	UINT32 dwIntFlags=0;
	VL_KEY keyCode=0;

	if (callerKeyCode == 0) {
		keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
	}
	else {
		SMSC_ASSERT(Vl_CheckLock(&(pDrvCtrl->MacPhyLock),callerKeyCode));
		keyCode = callerKeyCode;
	}

	if(pDrvCtrl->dwRxOffCount>0) {
		pDrvCtrl->dwRxOffCount--;
		if(pDrvCtrl->dwRxOffCount==0) {
			UINT32 dwMacCr=Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);
			if(dwMacCr&MAC_CR_RXEN_) {
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_RX, "Rx_ReceiverOn: Receiver is already on\n", 1, 2, 3, 4, 5, 6);
#endif
			}
			dwMacCr|=MAC_CR_RXEN_;
			Mac_SetRegDW(pDrvCtrl,MAC_CR,dwMacCr,keyCode);
		}
	} else {
		SMSC_ASSERT(FALSE);
	}

	if (callerKeyCode == 0) {
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}
}

/*********************************************************************************
Function Name:  Rx_HandleInterrupt

Description :   Rx handler 

input: 		Device Handler , Interrupt status 

output: void

**********************************************************************************/
LOCAL void Rx_HandleInterrupt(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwIntSts)
{
	UINT32	dwRxStatus;
	SMSC_ASSERT(pDrvCtrl!=NULL);


	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 0);
	#endif
	pDrvCtrl->LastReasonForReleasingCPU=0;

	if(dwIntSts&INT_STS_RXE_) {
		SMSC_TRACE("Rx_HandleInterrupt: RXE signalled");
		/* pDrvCtrl->stats.rx_errors++; */ 
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
	}

	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 1);
	#endif
	if(dwIntSts&INT_STS_RXDFH_INT_) {
		/* pDrvCtrl->stats.rx_dropped+=Lan_GetRegDW(RX_DROP); */ 
		END_ADD_IF_DROP(pDrvCtrl, Lan_GetRegDW(RX_DROP));
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR, "Rx_HandleInterrupt:INT_STS_RXDFH_INT_\n", 1, 2, 3, 4, 5, 6);	
#endif
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
	}
	
	#ifdef RX_LED_TRACE_EN
	Led_TestPoint(pDrvCtrl, 2);
	#endif
	if(dwIntSts&(INT_STS_RDFO_)) {
		Rx_HandleOverrun(pDrvCtrl);
		DRV_LOG (LAN9118_DEBUG_RX | LAN9118_DEBUG_RX_ERR,"Rx_HandleOverrun	INT_STS_RDFO_\n",1, 2, 3, 4, 5, 6);	
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 31);
		#endif
	}

	if((!(dwIntSts&INT_STS_RSFL_))&&(pDrvCtrl->RxOverrun==FALSE)) {
    	
		#ifdef RX_LED_TRACE_EN
		Led_TestPoint(pDrvCtrl, 3);
		#endif
		Lan_EnableInterrupt(pDrvCtrl,pDrvCtrl->RxInterrupts);
		return;
	}
	
	if(pDrvCtrl->MeasuringRxThroughput==FALSE) {
		pDrvCtrl->MeasuringRxThroughput=TRUE;
		Gpt_ScheduleCallBack(pDrvCtrl,GptCB_MeasureRxThroughput,1000);
		pDrvCtrl->RxFlowCurrentThroughput=0;
		pDrvCtrl->RxFlowCurrentPacketCount=0;
		pDrvCtrl->RxFlowCurrentWorkLoad=0;
	}

    Lan_HandleRxInt(pDrvCtrl, dwIntSts);
}

/*********************************************************************************
Function Name:  Rx_HandleInterrupt

Description :   Rx handler 

input: 		Device Handler , Interrupt status 

output: void

**********************************************************************************/
LOCAL void RxStop_HandleInterrupt(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwIntSts)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);

	Gpt_CancelCallBack (pDrvCtrl, GptCB_RxCompleteMulticast);
	Rx_CompleteMulticastUpdate (pDrvCtrl);
}

/*********************************************************************************
Function Name:  Mac_Initialize

Description :  This function is kept only as a place holder

input: 		LAN9118_END_DEVICE	device handler 	

output: void 

**********************************************************************************/
LOCAL STATUS Mac_Initialize (PLAN9118_END_DEV pDrvCtrl)
{
    UINT32 dwHigh16;
    UINT32 dwLow32;
    UINT32 local_MACCR;
    UINT32 dwTimeOut;
    UINT32 e2p_cmd;

	SMSC_ASSERT (pDrvCtrl != NULL);

	if (E2Prm_LoadFlag) {
		dwTimeOut=1000;
		e2p_cmd = Lan_GetRegDW (E2P_CMD);
		while((dwTimeOut>0)&&(e2p_cmd&E2P_CMD_EPC_BUSY_)) {
			Lan_udelay(pDrvCtrl, 5);
			dwTimeOut--;
			e2p_cmd = Lan_GetRegDW (E2P_CMD);
		}
					
		Eeprom_EnableAccess(pDrvCtrl);
		if(Eeprom_Reload(pDrvCtrl)){
			Eeprom_DisableAccess(pDrvCtrl);
			if(Eeprom_IsMacAddressLoaded(pDrvCtrl)) {
				UINT32 dwIntFlags=0;
				VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

				dwHigh16 =Mac_GetRegDW(pDrvCtrl,ADDRH,keyCode);
				dwLow32 =Mac_GetRegDW(pDrvCtrl,ADDRL,keyCode);
				
				Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);

				pDrvCtrl->enetAddr [0]=LOBYTE(LOWORD(dwLow32));
				pDrvCtrl->enetAddr [1]=HIBYTE(LOWORD(dwLow32));
				pDrvCtrl->enetAddr [2]=LOBYTE(HIWORD(dwLow32));
				pDrvCtrl->enetAddr [3]=HIBYTE(HIWORD(dwLow32));
				pDrvCtrl->enetAddr [4]=LOBYTE(LOWORD(dwHigh16));
				pDrvCtrl->enetAddr [5]=HIBYTE(LOWORD(dwHigh16));
			} else {
				goto MAC_DEFAULT;
			}
	
		} else {
			Eeprom_DisableAccess(pDrvCtrl);
			goto MAC_DEFAULT;
		}
	} else {
 MAC_DEFAULT:
		Lan_EnetAddrGet (pDrvCtrl, (char*) &(pDrvCtrl->enetAddr));
		dwHigh16 = (pDrvCtrl->enetAddr [4]) |
       	           (pDrvCtrl->enetAddr [5] << 8);
	    dwLow32 = (pDrvCtrl->enetAddr [3] << 24) | 
			      (pDrvCtrl->enetAddr [2] << 16) |
			      (pDrvCtrl->enetAddr [1] << 8) |
			      (pDrvCtrl->enetAddr [0]);
		{	
			UINT32 dwIntFlags=0;
			VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

        	Mac_SetRegDW (pDrvCtrl, ADDRH, dwHigh16,keyCode);
        	Mac_SetRegDW (pDrvCtrl, ADDRL, dwLow32,keyCode);

			Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
		}
	}
	
	DRV_LOG (LAN9118_DEBUG_MAC, "Mac Address is set to 0x%04X%08X\n", dwHigh16, dwLow32, 3, 4, 5, 6);
    
    if(END_MIB_INIT (&pDrvCtrl->endObj, M2_ifType_ethernet_csmacd,
                      &pDrvCtrl->enetAddr[0], LAN9118_EADR_LEN, ETHERMTU,
                      LAN9118_SPEED) == ERROR) {
        return FALSE;
    }
		
    return TRUE;
}

/*********************************************************************************
Function Name:  MacNotBusy

Description :   checks the MAC status 

input: 		LAN9118_END_DEVICE	device handler 		

output: STATUS 

**********************************************************************************/
LOCAL STATUS MacNotBusy (PLAN9118_END_DEV pDrvCtrl)
{
	int i = 0;

	/* wait for MAC not busy, w/ timeout */
	for (i = 0; i < 40; i++)
	{
		if ((Lan_GetRegDW(MAC_CSR_CMD) & MAC_CSR_CMD_CSR_BUSY_) == 0) {
			return TRUE;
		}
	}
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_MAC, "timeout waiting for MAC not BUSY. MAC_CSR_CMD = 0x%08X\n", Lan_GetRegDW (MAC_CSR_CMD), 2, 3, 4, 5, 6);
#endif
	return FALSE;
}

/*********************************************************************************
Function Name:  Mac_GetRegDW

Description :   

input: LAN9118_END_DEVICE device handler and register offset	

output: UINT32 

**********************************************************************************/
LOCAL UINT32 Mac_GetRegDW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegOffset,VL_KEY keyCode)
{
	UINT32 result = 0xFFFFFFFF;
	UINT32 dwTemp = 0;

	SMSC_ASSERT (pDrvCtrl != NULL);
	SMSC_ASSERT (pDrvCtrl->LanInitialized == TRUE);
	SMSC_ASSERT (Vl_CheckLock(&(pDrvCtrl->MacPhyLock),keyCode));
	SMSC_ASSERT (pDrvCtrl->dwLanBase != 0);

	/* wait until not busy */
	if (Lan_GetRegDW (MAC_CSR_CMD) & MAC_CSR_CMD_CSR_BUSY_)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_MAC, "Mac_GetRegDW() failed, MAC already busy at entry\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	/* send the MAC Cmd w/ offset */
	Lan_SetRegDW (MAC_CSR_CMD,
		((dwRegOffset & 0x000000FF) | 
		MAC_CSR_CMD_CSR_BUSY_ | MAC_CSR_CMD_R_NOT_W_));
	dwTemp = Lan_GetRegDW (BYTE_TEST);	/* to flush previous write */
	dwTemp = dwTemp;

	/* wait for the read to happen, w/ timeout */
	if (MacNotBusy (pDrvCtrl) == FALSE)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_MAC, "Mac_GetRegDW() failed, waiting for MAC not busy after read\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	} 
	else 
	{
		/* finally, return the read data */
		result = Lan_GetRegDW (MAC_CSR_DATA);
	}
DONE:
	return result;
}

/*********************************************************************************
Function Name:  Mac_SetRegDW

Description :   

input: LAN9118_END_DEVICE device handler and register offset	 		

output: void 

**********************************************************************************/
LOCAL void Mac_SetRegDW (PLAN9118_END_DEV pDrvCtrl, UINT32 dwRegOffset, UINT32 dwVal,VL_KEY keyCode)
{
	UINT32 dwTemp = 0;

	SMSC_ASSERT (pDrvCtrl != NULL);
	SMSC_ASSERT (pDrvCtrl->LanInitialized == TRUE);
	SMSC_ASSERT (Vl_CheckLock(&(pDrvCtrl->MacPhyLock),keyCode));
	SMSC_ASSERT (pDrvCtrl->dwLanBase != 0);

	if (Lan_GetRegDW (MAC_CSR_CMD) & MAC_CSR_CMD_CSR_BUSY_)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_MAC, "Mac_SetRegDW() failed, MAC already busy at entry\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	/* send the data to write */
	Lan_SetRegDW (MAC_CSR_DATA, dwVal);

	/* do the actual write */
	Lan_SetRegDW (MAC_CSR_CMD, ((dwRegOffset & 0x000000FF) | MAC_CSR_CMD_CSR_BUSY_));
	dwTemp = Lan_GetRegDW (BYTE_TEST);	/* force flush of previous write */
	dwTemp = dwTemp;

	/* wait for the write to complete, w/ timeout */
	if (MacNotBusy (pDrvCtrl) == FALSE)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_MAC, "Mac_SetRegDW() failed, waiting for MAC not busy after write\n", 1, 2, 3, 4, 5, 6);
#endif
	}
DONE:
	return;
}

/*********************************************************************************
Function Name:  Phy_Reset

Description :   function to reset the PHY

input: 	LAN9118_END_DEVICE	

output: UINT32 

**********************************************************************************/
#ifdef USE_PHY_WORK_AROUND

LOCAL UINT32 Phy_Reset(PLAN9118_END_DEV pDrvCtrl,VL_KEY keyCode)
{
	UINT32 result=FALSE;
	UINT16 wTemp=0;
	UINT32 dwLoopCount=100000;
	SMSC_TRACE("Performing PHY BCR Reset");
	Phy_SetRegW(pDrvCtrl,PHY_BCR,PHY_BCR_RESET_,keyCode);
	do {
		Lan_udelay(pDrvCtrl, 10);
		wTemp=Phy_GetRegW(pDrvCtrl,PHY_BCR,keyCode);
		dwLoopCount--;
	} while((dwLoopCount>0)&&(wTemp&PHY_BCR_RESET_));
	if(wTemp&PHY_BCR_RESET_) {
		SMSC_WARNING("Phy Reset failed to complete.");
		goto DONE;
	}
	/* extra delay required because the phy may not be completed with its reset
	  when PHY_BCR_RESET_ is cleared.
	  They say 256 uS is enough delay but I'm using 500 here to be safe
	*/
	Lan_udelay(pDrvCtrl, 500);
	result=TRUE;
DONE:
	return result;
}

/*********************************************************************************
Function Name:  Phy_LBT_GetTxStatus

Description :   gets the stats about the TX FIF0

input: 	LAN9118_END_DEVICE	

output: UINT32 

**********************************************************************************/
LOCAL UINT32 Phy_LBT_GetTxStatus(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=Lan_GetRegDW(TX_FIFO_INF);
	if(OLD_REGISTERS(pDrvCtrl)) {
		result&=TX_FIFO_INF_TSFREE_;
		if(result!=0x00800000UL) {
			result=Lan_GetRegDW(TX_STATUS_FIFO);
		} else {
			result=0;
		}
	} else {
		result&=TX_FIFO_INF_TSUSED_;
		if(result!=0x00000000UL) {
			result=Lan_GetRegDW(TX_STATUS_FIFO);
		} else {
			result=0;
		}
	}
	return result;
}



/*********************************************************************************
Function Name:  Phy_LBT_GetRxStatus

Description :   Gets the status of RX FIFO

input: 	LAN9118_END_DEVICE	device handler

output: UINT32 

**********************************************************************************/
LOCAL UINT32 Phy_LBT_GetRxStatus(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=Lan_GetRegDW(RX_FIFO_INF);
	if(result&0x00FF0000UL) {
		/*Rx status is available, read it */
		result=Lan_GetRegDW(RX_STATUS_FIFO);
	} else {
		result=0;
	}
	return result;
}

/*********************************************************************************
Function Name:  Phy_CheckLoopBackPacket

Description :   function to check the PHY Loop back Packet

input: 	LAN9118_END_DEVICE	device handler

output: UINT32

**********************************************************************************/
LOCAL UINT32 Phy_CheckLoopBackPacket(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=FALSE;
	UINT32 tryCount=0;
	UINT32 dwLoopCount=0;
	for(tryCount=0;tryCount<10;tryCount++)
	{
		UINT32 dwTxCmdA=0;
		UINT32 dwTxCmdB=0;
		UINT32 dwStatus=0;
		UINT32 dwPacketLength=0;
		
		/*zero-out Rx Packet memory */
		memset(pDrvCtrl->LoopBackRxPacket,0,MIN_PACKET_SIZE);
		
		/*write Tx Packet to 118 */
		dwTxCmdA=
			((((UINT32)(pDrvCtrl->LoopBackTxPacket))&0x03UL)<<16) | /*UINT32 alignment adjustment */
			TX_CMD_A_INT_FIRST_SEG_ | TX_CMD_A_INT_LAST_SEG_ |
			((UINT32)(MIN_PACKET_SIZE));
		dwTxCmdB=
			(((UINT32)(MIN_PACKET_SIZE))<<16) |
			((UINT32)(MIN_PACKET_SIZE));
		Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdA);
		Lan_SetRegDW(TX_DATA_FIFO,dwTxCmdB);
		Platform_WriteFifo(
			pDrvCtrl->dwLanBase,
			(UINT32 *)(((UINT32)(pDrvCtrl->LoopBackTxPacket))&0xFFFFFFFCUL),
			(((UINT32)(MIN_PACKET_SIZE))+3+
			(((UINT32)(pDrvCtrl->LoopBackTxPacket))&0x03UL))>>2);

		/*wait till transmit is done */
		dwLoopCount=60;
		while((dwLoopCount>0)&&((dwStatus=Phy_LBT_GetTxStatus(pDrvCtrl))==0)) {
			Lan_udelay(pDrvCtrl, 5);
			dwLoopCount--;
		}
		if(dwStatus==0) {
			SMSC_WARNING("Failed to Transmit during Loop Back Test");
			continue;
		}
		if(dwStatus&0x00008000UL) {
			SMSC_WARNING("Transmit encountered errors during Loop Back Test");
			continue;
		}

		/*wait till receive is done */
		dwLoopCount=60;
		while((dwLoopCount>0)&&((dwStatus=Phy_LBT_GetRxStatus(pDrvCtrl))==0))
		{
	         Lan_udelay(pDrvCtrl, 5);
	         dwLoopCount--;
		}
		if(dwStatus==0) {
			SMSC_WARNING("Failed to Receive during Loop Back Test");
			continue;
		}
		if(dwStatus&RX_STS_ES_)
		{
			SMSC_WARNING("Receive encountered errors during Loop Back Test");
			continue;
		}

		dwPacketLength=((dwStatus&0x3FFF0000UL)>>16);

		Platform_ReadFifo(
			pDrvCtrl->dwLanBase,
			((UINT32 *)(pDrvCtrl->LoopBackRxPacket)),
			(dwPacketLength+3+(((UINT32)(pDrvCtrl->LoopBackRxPacket))&0x03UL))>>2);
		
		if(dwPacketLength!=(MIN_PACKET_SIZE+4)) {
			SMSC_WARNING("Unexpected packet size during loop back test. Will retry");
		} else {
			UINT32 byteIndex=0;
			UINT32 foundMissMatch=FALSE;
			for(byteIndex=0;byteIndex<MIN_PACKET_SIZE;byteIndex++) {
				if(pDrvCtrl->LoopBackTxPacket[byteIndex]!=pDrvCtrl->LoopBackRxPacket[byteIndex])
				{
					foundMissMatch=TRUE;
					break;
				}         
			}
			if(!foundMissMatch) {
				SMSC_TRACE("Successfully Verified Loop Back Packet");
				result=TRUE;
				goto DONE;
			} else {
				SMSC_WARNING("Data miss match during loop back test, will retry.");
			}
		}
	}
DONE:
	return result;
}

/*********************************************************************************
Function Name:  Phy_LoopBackTest

Description :   Test case for the PHY Loop back packet

input: 	LAN9118_END_DEVICE	device handler		

output: UINT32 

**********************************************************************************/
LOCAL UINT32 Phy_LoopBackTest(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 result=FALSE;
	UINT32 byteIndex=0;
	UINT32 tryCount=0;

	/*Initialize Tx Packet */
	for(byteIndex=0;byteIndex<6;byteIndex++) {
		/*use broadcast destination address */
		pDrvCtrl->LoopBackTxPacket[byteIndex]=(UINT8)0xFF;
	}
	for(byteIndex=6;byteIndex<12;byteIndex++) {
		/*use incrementing source address */
		pDrvCtrl->LoopBackTxPacket[byteIndex]=(UINT8)byteIndex;
	}
	/*Set length type field */
	pDrvCtrl->LoopBackTxPacket[12]=0x00;
	pDrvCtrl->LoopBackTxPacket[13]=0x00;
	for(byteIndex=14;byteIndex<MIN_PACKET_SIZE;byteIndex++)
	{
		pDrvCtrl->LoopBackTxPacket[byteIndex]=(UINT8)byteIndex;
	}
	{
		UINT32 dwRegVal=Lan_GetRegDW(HW_CFG);
		dwRegVal&=(HW_CFG_TX_FIF_SZ_|0x00000FFFUL);
		dwRegVal|=HW_CFG_MBO_;
		Lan_SetRegDW(HW_CFG,dwRegVal);
	}
	Lan_SetRegDW(TX_CFG,TX_CFG_TX_ON_);

	Lan_SetRegDW(RX_CFG,(((UINT32)(pDrvCtrl->LoopBackRxPacket))&0x03)<<8);
	
	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		/*Set Phy to 10/FD, no ANEG, */
		Phy_SetRegW(pDrvCtrl,PHY_BCR,0x0100,keyCode);

		/*enable MAC Tx/Rx, FD */
		Mac_SetRegDW(pDrvCtrl,MAC_CR,MAC_CR_FDPX_|MAC_CR_TXEN_|MAC_CR_RXEN_,keyCode);

		/*set Phy to loopback mode */
		Phy_SetRegW(pDrvCtrl,PHY_BCR,0x4100,keyCode);

		for(tryCount=0;tryCount<10;tryCount++) {
			if(Phy_CheckLoopBackPacket(pDrvCtrl))
			{
				result=TRUE;
				goto DONE;
			}
			pDrvCtrl->dwResetCount++;
			/*disable MAC rx */
			Mac_SetRegDW(pDrvCtrl,MAC_CR,0UL,keyCode);
			Phy_Reset(pDrvCtrl,keyCode);

			/*Set Phy to 10/FD, no ANEG, and Loopbackmode */
			Phy_SetRegW(pDrvCtrl,PHY_BCR,0x4100,keyCode);

			/*enable MAC Tx/Rx, FD */
			Mac_SetRegDW(pDrvCtrl,MAC_CR,MAC_CR_FDPX_|MAC_CR_TXEN_|MAC_CR_RXEN_,keyCode);
		}
DONE:
		/*disable MAC */
		Mac_SetRegDW(pDrvCtrl,MAC_CR,0UL,keyCode);
		/*Cancel Phy loopback mode */
		Phy_SetRegW(pDrvCtrl,PHY_BCR,0U,keyCode);
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}

	Lan_SetRegDW(TX_CFG,0UL);
	Lan_SetRegDW(RX_CFG,0UL);
	
	return result;
}
#endif /* USE_PHY_WORK_AROUND */

/*********************************************************************************
Function Name:  Phy_SetLink

Description :   

input: 			pointer to LAN9118_END_DEVICE

output: void 

**********************************************************************************/
LOCAL void Phy_SetLink(PLAN9118_END_DEV pDrvCtrl) 
{
	UINT32 dwLinkRequest = pDrvCtrl->link_mode;
	UINT32 dwIntFlags=0;
	VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

	if(dwLinkRequest&LINK_AUTO_NEGOTIATE) {
		UINT16 wTemp;
		wTemp=Phy_GetRegW(pDrvCtrl, PHY_ANEG_ADV,keyCode);
		wTemp&=~PHY_ANEG_ADV_PAUSE_;
		if(dwLinkRequest&LINK_ASYMMETRIC_PAUSE) {
			wTemp|=PHY_ANEG_ADV_ASYMP_;
		} 
		if(dwLinkRequest&LINK_SYMMETRIC_PAUSE) {
			wTemp|=PHY_ANEG_ADV_SYMP_;
		}
		wTemp&=~PHY_ANEG_ADV_SPEED_;
		if(dwLinkRequest&LINK_SPEED_10HD) {
			wTemp|=PHY_ANEG_ADV_10H_;
		}
		if(dwLinkRequest&LINK_SPEED_10FD) {
			wTemp|=PHY_ANEG_ADV_10F_;
		}
		if(dwLinkRequest&LINK_SPEED_100HD) {
			wTemp|=PHY_ANEG_ADV_100H_;
		}
		if(dwLinkRequest&LINK_SPEED_100FD) {
			wTemp|=PHY_ANEG_ADV_100F_;
		}
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY,
				 "Phy_SetLink: link_mode=0x%04X\n",
				 dwLinkRequest, 2, 3, 4, 5, 6);
		DRV_LOG (LAN9118_DEBUG_PHY,
				 "                     PHY_ANEG_ADV before 0x%04X\n",
				 Phy_GetRegW(pDrvCtrl,PHY_ANEG_ADV,keyCode), 2, 3, 4, 5, 6);
		DRV_LOG (LAN9118_DEBUG_PHY,
				 "                     PHY_ANEG_ADV set to 0x%04X\n",
				 wTemp, 2, 3, 4, 5, 6);
#endif
		Phy_SetRegW(pDrvCtrl,PHY_ANEG_ADV,wTemp,keyCode);
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY,
				 "                     PHY_ANEG_ADV verify 0x%04X\n",
				 Phy_GetRegW(pDrvCtrl,PHY_ANEG_ADV,keyCode), 2, 3, 4, 5, 6);
#endif


		/* begin to establish link */
		pDrvCtrl->dwRemoteFaultCount=0;
		Phy_SetRegW(pDrvCtrl,
			PHY_BCR,
			PHY_BCR_AUTO_NEG_ENABLE_|
			PHY_BCR_RESTART_AUTO_NEG_,keyCode);
	} else {
		UINT16 wTemp=0;
		if(dwLinkRequest&(LINK_SPEED_100FD)) {
			dwLinkRequest=LINK_SPEED_100FD;
		} else if(dwLinkRequest&(LINK_SPEED_100HD)) {
			dwLinkRequest=LINK_SPEED_100HD;
		} else if(dwLinkRequest&(LINK_SPEED_10FD)) {
			dwLinkRequest=LINK_SPEED_10FD;
		} else if(dwLinkRequest&(LINK_SPEED_10HD)) {
			dwLinkRequest=LINK_SPEED_10HD;
		}
		if(dwLinkRequest&(LINK_SPEED_10FD|LINK_SPEED_100FD)) {
			wTemp|=PHY_BCR_DUPLEX_MODE_;
		}
		if(dwLinkRequest&(LINK_SPEED_100HD|LINK_SPEED_100FD)) {
			wTemp|=PHY_BCR_SPEED_SELECT_;
		}
		Phy_SetRegW(pDrvCtrl,PHY_BCR,wTemp,keyCode);
	}

	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Phy_SetAutoMdixSts

Description :   

input: 			pointer to LAN9118_END_DEVICE

output: void 

**********************************************************************************/
LOCAL void Phy_SetAutoMdixSts(PLAN9118_END_DEV pDrvCtrl, 
                        UINT16 wAutoMdixSts)
{
	UINT16 SpecialCtrlSts=0U;

	if (((pDrvCtrl->dwGeneration)>2) && (!(pDrvCtrl->ExtPhy))) 

	{
		if (wAutoMdixSts > 2)
		{
			UINT32 dwIntFlags=0;
			VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

			SpecialCtrlSts=Phy_GetRegW(pDrvCtrl, PHY_SPLMODE_STSIND,keyCode);
			SpecialCtrlSts = (SpecialCtrlSts&0x1FFF);
			Phy_SetRegW(pDrvCtrl, PHY_SPLMODE_STSIND,SpecialCtrlSts,keyCode);

			Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);	

			if (Lan_GetRegDW(HW_CFG) & HW_CFG_AMDIX_EN_) {
				SMSC_TRACE("Auto-MDIX Enable by default!!!");
			}
			else {
				SMSC_TRACE("Auto-MDIX Disable by default!!!");
			}
        }
		else 
		{
            UINT32 dwIntFlags=0;
			VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

			SpecialCtrlSts=Phy_GetRegW(pDrvCtrl, PHY_SPLMODE_STSIND,keyCode);
			SpecialCtrlSts = (((wAutoMdixSts+4) << 13) | (SpecialCtrlSts&0x1FFF));
			Phy_SetRegW(pDrvCtrl, PHY_SPLMODE_STSIND,SpecialCtrlSts,keyCode);

			Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);	

			if (wAutoMdixSts & AMDIX_ENABLE) {
				SMSC_TRACE("Override Strap, Enable Auto-MDIX ");
			} else if (wAutoMdixSts & AMDIX_DISABLE_CROSSOVER) {
				SMSC_TRACE("Override Strap, Disable Auto-MDIX, CrossOver Cable");		
			} else {
				SMSC_TRACE("Override Strap, Disable Auto-MDIX, Straight Cable");
			}
			
		}
	}
	else {

		SMSC_TRACE("This chip or PHY doesn't support HP AMDIX!!!");
	}
}

/*********************************************************************************
Function Name:  Phy_GetAutoMdixSts

Description :   

input: 			pointer to LAN9118_END_DEVICE

output: void 

**********************************************************************************/
LOCAL UINT16 Phy_GetAutoMdixSts(PLAN9118_END_DEV pDrvCtrl)
{
	UINT16 SpecialCtrlSts=0U;
	UINT32 dwIntFlags=0;

	if (((pDrvCtrl->dwGeneration)>2) && (!(pDrvCtrl->ExtPhy))) 

	{
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
        SpecialCtrlSts=Phy_GetRegW(pDrvCtrl, PHY_SPLMODE_STSIND,keyCode);
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);	

        if (SpecialCtrlSts & SPLMODE_STSIND_OVRIDE_AMDIX_) {
			if (SpecialCtrlSts & SPLMODE_STSIND_AMIDX_EN_) {
				SMSC_TRACE("AutoMdix Status: Override Strap, Enable Auto Mdix");
			}
			else if (SpecialCtrlSts & SPLMODE_STSIND_AMDIX_STS_) {
				SMSC_TRACE("AutoMdix Status: Override Strap, Disable Auto Mdix, CrossOver Cable");
			} else {
				SMSC_TRACE("AutoMdix Status: Override Strap, Disable Auto Mdix, Straight Cable");
			}
		}
		else {
			if (Lan_GetRegDW(HW_CFG) & HW_CFG_AMDIX_EN_) {
				SMSC_TRACE("AutoMdix Status: Enable by default!!!");
			}
			else {
				SMSC_TRACE("AutoMdix Status: Disable by default!!!");
			}
		}
	}
	else {
		SMSC_TRACE("This chip or PHY doesn't support HP AMDIX!!!");
	}
	
	if (!(SpecialCtrlSts & SPLMODE_STSIND_OVRIDE_AMDIX_)) {
	    return (UINT16)3;
	} else {
	    return (UINT16)((SpecialCtrlSts & (UINT16)0x6000)>>(UINT16)13);
	}
}

/*********************************************************************************
Function Name:  Phy_Initialize

Description :   

input: LAN9118_END_DEVICE device handler		

output: STATUS 

**********************************************************************************/
LOCAL STATUS Phy_Initialize (PLAN9118_END_DEV pDrvCtrl)
{
	STATUS result = ERROR;
	BOOL bConfigureAutoMdix = FALSE;
	UINT32 dwTemp = 0;
	UINT16 wTemp = 0;
	UINT32 dwLoopCount = 0;

	SMSC_ASSERT(pDrvCtrl != NULL);
	SMSC_ASSERT(pDrvCtrl->dwLanBase != 0);
	SMSC_ASSERT(pDrvCtrl->link_mode <= 0x7F);
	pDrvCtrl->ExtPhy=FALSE;

	if(pDrvCtrl->phy_addr!=0xFFFFFFFF) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "-->Auto Phy detection  %x \n", (pDrvCtrl->dwIdRev&0xFFFF0000), 2, 3, 4, 5, 6);
#endif
		switch(pDrvCtrl->dwIdRev&0xFFFF0000) {
			case 0x217A0000UL:
			case 0x215A0000UL:
			case 0x117A0000UL:
			case 0x115A0000UL:
				pDrvCtrl->dwIdRev|=1;
			case 0x01170000UL:
			case 0x01150000UL:
				if(pDrvCtrl->dwIdRev&0x0000FFFF) {
					UINT32 dwHwCfg=0;

					pDrvCtrl->dwIdRev = Lan_GetRegDW(ID_REV);
					dwHwCfg=Lan_GetRegDW(HW_CFG);

					if(dwHwCfg&HW_CFG_EXT_PHY_DET_) {
              			/*External phy is requested, supported, and detected Attempt to switch */
						/*	NOTE: Assuming Rx and Tx are stopped because Phy_Initialize is called before 
				  			Rx_Initialize and Tx_Initialize */
						UINT16 wPhyId1=0;
						UINT16 wPhyId2=0;

						/*Disable phy clocks to the mac */
						dwHwCfg&= (~HW_CFG_PHY_CLK_SEL_);
						dwHwCfg|= HW_CFG_PHY_CLK_SEL_CLK_DIS_;
						Lan_SetRegDW(HW_CFG,dwHwCfg);
						Lan_udelay(pDrvCtrl, 10);/*wait for clocks to acutally stop */

						dwHwCfg|=HW_CFG_EXT_PHY_EN_;
						Lan_SetRegDW(HW_CFG,dwHwCfg);

						dwHwCfg&= (~HW_CFG_PHY_CLK_SEL_);
						dwHwCfg|= HW_CFG_PHY_CLK_SEL_EXT_PHY_;
						Lan_SetRegDW(HW_CFG,dwHwCfg);
						Lan_udelay(pDrvCtrl, 10);/*wait for clocks to actually start */

						dwHwCfg|=HW_CFG_SMI_SEL_;
						Lan_SetRegDW(HW_CFG,dwHwCfg);

						{
							UINT32 dwIntFlags=0;
							VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
			
							if(pDrvCtrl->phy_addr<=31) {
								/* only check the phy address specified */
								wPhyId1=Phy_GetRegW(pDrvCtrl,PHY_ID_1,keyCode);
								wPhyId2=Phy_GetRegW(pDrvCtrl,PHY_ID_2,keyCode);

							} else {
								/*auto detect phy */
								UINT32 address=0;
								for(address=0;address<=31;address++) {
									pDrvCtrl->phy_addr=address;
									wPhyId1=Phy_GetRegW(pDrvCtrl,PHY_ID_1,keyCode);
									wPhyId2=Phy_GetRegW(pDrvCtrl,PHY_ID_2,keyCode);
									if((wPhyId1!=0xFFFFU)||(wPhyId2!=0xFFFFU)) {
#ifdef VXWORKS_DEBUG
										DRV_LOG (LAN9118_DEBUG_PHY, "Detected Phy at address = 0x%02X = %d\n", address, address, 3, 4, 5, 6);
#endif
										break;
									}
								}
								
								if(address>=32) {
#ifdef VXWORKS_DEBUG
									DRV_LOG (LAN9118_DEBUG_PHY, "Failed to auto detect external phy\n", 1, 2, 3, 4, 5, 6);
#endif
								}
							}
							
							Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
						}
					
						if((wPhyId1==0xFFFFU)&&(wPhyId2==0xFFFFU)) {
#ifdef VXWORKS_DEBUG
							DRV_LOG (LAN9118_DEBUG_PHY, "External Phy is not accessable\n", 1, 2, 3, 4, 5, 6);
							DRV_LOG (LAN9118_DEBUG_PHY, "  using internal phy instead\n", 1, 2, 3, 4, 5, 6);
#endif
							/*revert back to interal phy settings. */
							/*Disable phy clocks to the mac */
							dwHwCfg&= (~HW_CFG_PHY_CLK_SEL_);
							dwHwCfg|= HW_CFG_PHY_CLK_SEL_CLK_DIS_;
							Lan_SetRegDW(HW_CFG,dwHwCfg);
							Lan_udelay(pDrvCtrl, 10);/*wait for clocks to actually stop */

							dwHwCfg&=(~HW_CFG_EXT_PHY_EN_);
							Lan_SetRegDW(HW_CFG,dwHwCfg);
		
							dwHwCfg&= (~HW_CFG_PHY_CLK_SEL_);
							dwHwCfg|= HW_CFG_PHY_CLK_SEL_INT_PHY_;
							Lan_SetRegDW(HW_CFG,dwHwCfg);
							Lan_udelay(pDrvCtrl, 10);/*wait for clocks to actually start */

							dwHwCfg&=(~HW_CFG_SMI_SEL_);
							Lan_SetRegDW(HW_CFG,dwHwCfg);
							goto USE_INTERNAL_PHY;
						} else {
#ifdef VXWORKS_DEBUG
							DRV_LOG (LAN9118_DEBUG_PHY, "Successfully switched to external phy\n", 1, 2, 3, 4, 5, 6);
#endif

#ifdef USE_LED1_WORK_AROUND
							pDrvCtrl->NotUsingExtPhy=0;
#endif
						}
					} else {
#ifdef VXWORKS_DEBUG
						DRV_LOG (LAN9118_DEBUG_PHY, "No External Phy Detected\n", 1, 2, 3, 4, 5, 6);
						DRV_LOG (LAN9118_DEBUG_PHY, "  using internal phy instead\n", 1, 2, 3, 4, 5, 6);
#endif
						goto USE_INTERNAL_PHY;
					}
				} else {
#ifdef VXWORKS_DEBUG
					DRV_LOG (LAN9118_DEBUG_PHY, "External Phy is not supported\n", 1, 2, 3, 4, 5, 6);
					DRV_LOG (LAN9118_DEBUG_PHY, "  using internal phy instead\n", 1, 2, 3, 4, 5, 6);
#endif
					goto USE_INTERNAL_PHY;
				};
				break;
	
			default:
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_PHY, "External Phy is not supported\n", 1, 2, 3, 4, 5, 6);
				DRV_LOG (LAN9118_DEBUG_PHY, "  using internal phy instead\n", 1, 2, 3, 4, 5, 6);
#endif
				goto USE_INTERNAL_PHY;
		}
	} else {
USE_INTERNAL_PHY:
		pDrvCtrl->phy_addr=1;
		pDrvCtrl->ExtPhy=FALSE;
#ifdef USE_LED1_WORK_AROUND
	if(pDrvCtrl->dwGeneration<=2) {
			pDrvCtrl->NotUsingExtPhy=1;
		} else {
			/*Generation 3 or higher has the LED problem fixed
			  to disable the workaround pretend the phy is external*/
			pDrvCtrl->NotUsingExtPhy=0;
		}
#endif
		bConfigureAutoMdix = TRUE;
	}
	
	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

		dwTemp=Phy_GetRegW(pDrvCtrl,PHY_ID_2,keyCode);
		pDrvCtrl->bPhyRev=((UINT8)(dwTemp&(0x0F)));
		pDrvCtrl->bPhyModel=((UINT8)((dwTemp>>4)&(0x3F)));
		pDrvCtrl->dwPhyId=((dwTemp&(0xFC00))<<8);
		dwTemp=Phy_GetRegW(pDrvCtrl,PHY_ID_1,keyCode);
		pDrvCtrl->dwPhyId|=((dwTemp&(0x0000FFFF))<<2);
	
		pDrvCtrl->dwLinkSpeed=LINK_OFF;
		pDrvCtrl->dwLinkSettings=LINK_OFF;
		/*reset the PHY */
		Phy_SetRegW(pDrvCtrl,PHY_BCR,PHY_BCR_RESET_,keyCode);
	    dwLoopCount=100000;
		do {

			Lan_udelay(pDrvCtrl, 10);
			wTemp=Phy_GetRegW(pDrvCtrl,PHY_BCR,keyCode);
			dwLoopCount--;
		} while((dwLoopCount>0) && (wTemp&PHY_BCR_RESET_));

		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
#ifdef VXWORKS_DEBUG	
		DRV_LOG (LAN9118_DEBUG_PHY, "dwPhyId==0x%08X,bPhyModel==0x%02X,bPhyRev==0x%02X\n", pDrvCtrl->dwPhyId, pDrvCtrl->bPhyModel, pDrvCtrl->bPhyRev, 4, 5, 6);
#endif
	}

	if(wTemp&PHY_BCR_RESET_) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "PHY reset failed to complete.\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}
	
	if (bConfigureAutoMdix == TRUE)
	{
		Phy_SetAutoMdixSts(pDrvCtrl,pDrvCtrl->AutoMdix);	
	}
	
#ifdef USE_PHY_WORK_AROUND
if(pDrvCtrl->dwGeneration<=2){
	if(Phy_LoopBackTest(pDrvCtrl) == FALSE) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "Failed Loop back test\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	} else {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "Passed Loop Back Test\n", 1, 2, 3, 4, 5, 6);
#endif
	}
}
#endif /*USE_PHY_WORK_AROUND*/

	/*
	switch((pDrvCtrl->dwIdRev) & 0xFFFF0000UL)
		{
		case 0x92100000UL:
		case 0x92110000UL:
		case 0x92200000UL:
		case 0x92210000UL:
		case 0x118A0000UL: 
		case 0x117A0000UL:
		case 0x115A0000UL:
		Phy_SetAutoMdixSts(pDrvCtrl,pDrvCtrl->AutoMdix);
		break;
		default:
		break;
		}
	*/
	Phy_SetLink(pDrvCtrl);


	/* prepare GPT callback 1 second, resolution is 100 uSec */
	/* argument is always pDrvCtrl */
	Gpt_ScheduleCallBack (pDrvCtrl, (void (*) (LAN9118_END_DEV *)) Phy_CheckLink, 10000UL);/*10000*/
	result=TRUE;
DONE:
#ifdef VXWORKS_DEBUG
	if (result) {
	DRV_LOG (LAN9118_DEBUG_PHY, "<--Phy_Initialize, result=TRUE\n", 1, 2, 3, 4, 5, 6);

	}
	else {
	DRV_LOG (LAN9118_DEBUG_PHY, "<--Phy_Initialize, result=FALSE\n", 1, 2, 3, 4, 5, 6);
	}
#endif
	return result;
}



/*********************************************************************************
Function Name:  Phy_GetRegW

Description :   

input: 		LAN9118_END_DEVICE	device handler 	and register index

output: UINT16 

**********************************************************************************/
LOCAL UINT16 Phy_GetRegW(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwRegIndex,VL_KEY keyCode)
{
	UINT32 dwAddr=0;
	int i=0;
	UINT16 result=0xFFFFU;

	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized==TRUE);
	SMSC_ASSERT(Vl_CheckLock(&(pDrvCtrl->MacPhyLock),keyCode));

	/* confirm MII not busy */
	if ((Mac_GetRegDW(pDrvCtrl, MII_ACC,keyCode) & MII_ACC_MII_BUSY_) != 0)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "MII is busy in Phy_GetRegW???\n", 1, 2, 3, 4, 5, 6);
#endif
		result=0;
		goto DONE;
	}

	/* set the address, index & direction (read from PHY) */
	dwAddr = ((pDrvCtrl->phy_addr&0x1F)<<11) | ((dwRegIndex & 0x1F)<<6);
	Mac_SetRegDW(pDrvCtrl, MII_ACC, dwAddr,keyCode);

	/* wait for read to complete w/ timeout */
	for(i=0;i<100;i++) {
		/* see if MII is finished yet */
		if ((Mac_GetRegDW(pDrvCtrl, MII_ACC,keyCode) & MII_ACC_MII_BUSY_) == 0)
		{
			/* get the read data from the MAC & return i */
			result=((UINT16)Mac_GetRegDW(pDrvCtrl, MII_DATA,keyCode));
			goto DONE;
		}
	}
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_PHY, "timeout waiting for MII write to finish\n", 1, 2, 3, 4, 5, 6);
#endif

DONE:
	return result;
}

/*********************************************************************************
Function Name:  Phy_SetRegW

Description :   

input: 		LAN9118_END_DEVICE	device handler ,register index and value to set

output: void 

**********************************************************************************/
LOCAL void Phy_SetRegW(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwRegIndex,UINT16 wVal,VL_KEY keyCode)
{
	UINT32 dwAddr=0;
	int i=0;

	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(pDrvCtrl->LanInitialized==TRUE);
	SMSC_ASSERT(Vl_CheckLock(&(pDrvCtrl->MacPhyLock),keyCode));

	if(dwRegIndex==0) {
		if((wVal&0x1200)==0x1200) {
			pDrvCtrl->wLastADVatRestart=pDrvCtrl->wLastADV;
		}
	}
	if(dwRegIndex==4) {
		pDrvCtrl->wLastADV=wVal;
	}

	/* confirm MII not busy */
	if ((Mac_GetRegDW(pDrvCtrl, MII_ACC,keyCode) & MII_ACC_MII_BUSY_) != 0)
	{
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "MII is busy in Phy_SetRegW???\n", 1, 2, 3, 4, 5, 6);
#endif
		goto DONE;
	}

	/* put the data to write in the MAC */
	Mac_SetRegDW(pDrvCtrl, MII_DATA, (UINT32)wVal,keyCode);

	/* set the address, index & direction (write to PHY) */
	dwAddr = ((pDrvCtrl->phy_addr&0x1F)<<11) | ((dwRegIndex & 0x1F)<<6) | MII_ACC_MII_WRITE_;
	Mac_SetRegDW(pDrvCtrl, MII_ACC, dwAddr,keyCode);

	/* wait for write to complete w/ timeout */
	for(i=0;i<100;i++) {
		/* see if MII is finished yet */
		if ((Mac_GetRegDW(pDrvCtrl, MII_ACC,keyCode) & MII_ACC_MII_BUSY_) == 0)
		{
			goto DONE;
		}
	}
#ifdef VXWORKS_DEBUG
	DRV_LOG (LAN9118_DEBUG_PHY, "timeout waiting for MII write to finish\n", 1, 2, 3, 4, 5, 6);
#endif
DONE:
	return;
}

/*********************************************************************************
Function Name:  Phy_UpdateLinkMode

Description :   this function Updates the Link mode 

input: 	Device handler 	

output: void 

**********************************************************************************/
LOCAL void Phy_UpdateLinkMode(PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 dwOldLinkSpeed=pDrvCtrl->dwLinkSpeed;
	UINT32 dwIntFlags=0;
	VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

	Phy_GetLinkMode(pDrvCtrl, keyCode);

	if(dwOldLinkSpeed!=(pDrvCtrl->dwLinkSpeed)) {
		if(pDrvCtrl->dwLinkSpeed!=LINK_OFF) {
			UINT32 dwRegVal=0;
			
		    /* Initialize pointer to interface structure */
		    if ( pDrvCtrl->pIf == NULL )
		    {
		      char ifUnitStr[10];
		      char ifNameStr[40];
		      
		      sprintf(ifUnitStr, "%d", pDrvCtrl->unit);
		      strcpy (ifNameStr, LAN9118_DEV_NAME);
		      strcat (ifNameStr, ifUnitStr);
		      pDrvCtrl->pIf = ifunit (ifNameStr);
		      if (pDrvCtrl->pIf == NULL)
		      {
		        SMSC_WARNING ("ifunit returns NULL!!!");	
		      }
		    }
		    
			switch(pDrvCtrl->dwLinkSpeed) {
			case LINK_SPEED_10HD:
				logMsg ("Link is now UP at 10Mbps HD\n", 1, 2, 3, 4, 5, 6);
				break;
			case LINK_SPEED_10FD:
				logMsg ("Link is now UP at 10Mbps FD\n", 1, 2, 3, 4, 5, 6);
				break;
			case LINK_SPEED_100HD:
				logMsg ("Link is now UP at 100Mbps HD\n", 1, 2, 3, 4, 5, 6);
				break;
			case LINK_SPEED_100FD:
				logMsg ("Link is now UP at 100Mbps FD\n", 1, 2, 3, 4, 5, 6);
				break;
			default:
				logMsg ("Link is now UP at Unknown Link Speed, dwLinkSpeed=0x%08X\n", pDrvCtrl->dwLinkSpeed, 2, 3, 4, 5, 6);
				break;
			}
		
			dwRegVal=Mac_GetRegDW(pDrvCtrl,MAC_CR,keyCode);
			dwRegVal&=~(MAC_CR_FDPX_|MAC_CR_RCVOWN_);
			switch(pDrvCtrl->dwLinkSpeed) {
			case LINK_SPEED_10HD:
			case LINK_SPEED_100HD:
				dwRegVal|=MAC_CR_RCVOWN_;
				break;
			case LINK_SPEED_10FD:
			case LINK_SPEED_100FD:
				dwRegVal|=MAC_CR_FDPX_;
				break;
			default:break;
			}

			Mac_SetRegDW(pDrvCtrl,
				MAC_CR,dwRegVal,keyCode);

			if(pDrvCtrl->dwLinkSettings&LINK_AUTO_NEGOTIATE) {
				UINT16 linkPartner=0;
				UINT16 localLink=0;
				localLink=Phy_GetRegW(pDrvCtrl,4,keyCode);
				linkPartner=Phy_GetRegW(pDrvCtrl,5,keyCode);
				switch(pDrvCtrl->dwLinkSpeed) {
				case LINK_SPEED_10FD:
				case LINK_SPEED_100FD:
					if(((localLink&linkPartner)&((UINT16)0x0400U)) != ((UINT16)0U)) {
						/*Enable PAUSE receive and transmit */
						Mac_SetRegDW(pDrvCtrl,FLOW,0xFFFF0002,keyCode);
						Lan_SetBitsDW(AFC_CFG,(pDrvCtrl->afc_cfg&0x0000000F));
					} else if(((localLink&((UINT16)0x0C00U))==((UINT16)0x0C00U)) &&
							((linkPartner&((UINT16)0x0C00U))==((UINT16)0x0800U)))
					{
						/*Enable PAUSE receive, disable PAUSE transmit */
						Mac_SetRegDW(pDrvCtrl,FLOW,0xFFFF0002,keyCode);
						Lan_ClrBitsDW(AFC_CFG,0x0000000F);
					} else {
						/*Disable PAUSE receive and transmit */
						Mac_SetRegDW(pDrvCtrl,FLOW,0,keyCode);
						Lan_ClrBitsDW(AFC_CFG,0x0000000F);
					};break;
				case LINK_SPEED_10HD:
				case LINK_SPEED_100HD:
					Mac_SetRegDW(pDrvCtrl,FLOW,0,keyCode);
					Lan_SetBitsDW(AFC_CFG,0x0000000F);
					break;
				default:break;
				}
#ifdef VXWORKS_DEBUG
				DRV_LOG (LAN9118_DEBUG_PHY, "LAN9118: %s,%s,%s,%s,%s,%s\n",
					(UINT32)((localLink&PHY_ANEG_ADV_ASYMP_)?"ASYMP":"     "),
					(UINT32)((localLink&PHY_ANEG_ADV_SYMP_)?"SYMP ":"     "),
					(UINT32)((localLink&PHY_ANEG_ADV_100F_)?"100FD":"     "),
					(UINT32)((localLink&PHY_ANEG_ADV_100H_)?"100HD":"     "),
					(UINT32)((localLink&PHY_ANEG_ADV_10F_)?"10FD ":"     "),
					(UINT32)((localLink&PHY_ANEG_ADV_10H_)?"10HD ":"     "));
	
				DRV_LOG (LAN9118_DEBUG_PHY, "Partner: %s,%s,%s,%s,%s,%s\n",
					(UINT32)((linkPartner&PHY_ANEG_LPA_ASYMP_)?"ASYMP":"     "),
					(UINT32)((linkPartner&PHY_ANEG_LPA_SYMP_)?"SYMP ":"     "),
					(UINT32)((linkPartner&PHY_ANEG_LPA_100FDX_)?"100FD":"     "),
					(UINT32)((linkPartner&PHY_ANEG_LPA_100HDX_)?"100HD":"     "),
					(UINT32)((linkPartner&PHY_ANEG_LPA_10FDX_)?"10FD ":"     "),
					(UINT32)((linkPartner&PHY_ANEG_LPA_10HDX_)?"10HD ":"     "));
#endif
			} else {
				switch(pDrvCtrl->dwLinkSpeed) {
				case LINK_SPEED_10HD:
				case LINK_SPEED_100HD:
					Mac_SetRegDW(pDrvCtrl,FLOW,0x0,keyCode);
					Lan_SetBitsDW(AFC_CFG,0x0000000F);
					break;
				default:
					Mac_SetRegDW(pDrvCtrl,FLOW,0x0,keyCode);
					Lan_ClrBitsDW(AFC_CFG,0x0000000F);
					break;
				}
			}
			
			/* Flush Rx FIFO and turn on Rx */
			{
			 	UINT32 temp, timeOut;
			 	
			 	/* Dump Rx FIFO before enabling Rx */
			 	temp=Lan_GetRegDW(RX_CFG);
				Lan_SetRegDW(RX_CFG,(temp&0x3FFFFFFFUL));
				timeOut=10000000;
				Lan_SetBitsDW(RX_CFG,RX_CFG_RX_DUMP_);
				while((timeOut>0)&&(Lan_GetRegDW(RX_CFG)&(RX_CFG_RX_DUMP_))) {
					Lan_udelay(pDrvCtrl, 1);
					timeOut--;
				}
				if(timeOut==0) {
					SMSC_WARNING("Timed out waiting for Rx Dump to complete\n");
				}
				Lan_SetRegDW(RX_CFG,temp);
				
				pDrvCtrl->dwRxOffCount = 1;
				Rx_ReceiverOn(pDrvCtrl, 0);
			}

			Tx_WakeQueue(pDrvCtrl,0x01);
#ifdef USE_LED1_WORK_AROUND
			if ((g_GpioSettingOriginal & GPIO_CFG_LED1_EN_) &&
				pDrvCtrl->NotUsingExtPhy)
			{
				/* Restore orginal GPIO configuration */
				g_GpioSetting = g_GpioSettingOriginal;
				Lan_SetRegDW(GPIO_CFG,g_GpioSetting);
			}
#endif /* USE_LED1_WORK_AROUND */
		} else {
			logMsg ("Link is now DOWN\n", 1, 2, 3, 4, 5, 6);

			Tx_StopQueue(pDrvCtrl,0x01);
			
			/* Turn off the receiver */
			Rx_ReceiverOff(pDrvCtrl); 
			
			Mac_SetRegDW(pDrvCtrl,FLOW,0,keyCode);
			Lan_ClrBitsDW(AFC_CFG,0x0000000F);
#ifdef USE_LED1_WORK_AROUND
			/* Check global setting that LED1 usage is 10/100 indicator */
/*			g_GpioSetting = Lan_GetRegDW(GPIO_CFG); */
			if ((g_GpioSetting & GPIO_CFG_LED1_EN_) &&
				pDrvCtrl->NotUsingExtPhy)
			{
				/*Force 10/100 LED off, after saving orginal GPIO configuration */
				g_GpioSettingOriginal = g_GpioSetting;

				g_GpioSetting &= ~GPIO_CFG_LED1_EN_;
				g_GpioSetting |=
					(GPIO_CFG_GPIOBUF0_|GPIO_CFG_GPIODIR0_|GPIO_CFG_GPIOD0_);
				Lan_SetRegDW(GPIO_CFG,g_GpioSetting);
			}
#endif /* USE_LED1_WORK_AROUND */
		}
	}

	Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
}

/*********************************************************************************
Function Name:  Phy_CheckLink

Description :   

input: 	device handler	

output: void 

**********************************************************************************/
LOCAL void Phy_CheckLink(unsigned long ptr)
{
	LAN9118_END_DEV *pDrvCtrl=(LAN9118_END_DEV *) ptr;
	if(pDrvCtrl==NULL) {
#ifdef VXWORKS_DEBUG
		DRV_LOG (LAN9118_DEBUG_PHY, "Phy_CheckLink(ptr==0)\n", 1, 2, 3, 4, 5, 6);
#endif
		return;
	}

	END_ADD_IF_DROP(pDrvCtrl, Lan_GetRegDW(RX_DROP)); 

	/*must call this twice */ 
	Phy_UpdateLinkMode(pDrvCtrl);
	Phy_UpdateLinkMode(pDrvCtrl);

	if(!(pDrvCtrl->StopLinkPolling)) {
		Gpt_ScheduleCallBack (pDrvCtrl, (void (*) (LAN9118_END_DEV *)) Phy_CheckLink, 10000UL);/*10000*/
	}
}

/*********************************************************************************
Function Name:  Phy_GetLinkMode

Description :   This function get the Current link Mode 

input: 		Device Handler

output: void 

**********************************************************************************/
LOCAL void Phy_GetLinkMode(
	PLAN9118_END_DEV pDrvCtrl,
	VL_KEY keyCode)
{
	UINT32 result=LINK_OFF;
	UINT16 wRegVal=0;
	UINT16 wRegBSR=Phy_GetRegW(
		pDrvCtrl,
		PHY_BSR,keyCode);
	pDrvCtrl->dwLinkSettings=LINK_OFF;
	if(wRegBSR&PHY_BSR_LINK_STATUS_) {
		wRegVal=Phy_GetRegW(
			pDrvCtrl,
			PHY_BCR,keyCode);
		if(wRegVal&PHY_BCR_AUTO_NEG_ENABLE_) {
			UINT32 linkSettings=LINK_AUTO_NEGOTIATE;
			UINT16 wRegADV=pDrvCtrl->wLastADVatRestart;
			UINT16 wRegLPA=Phy_GetRegW(
				pDrvCtrl,
				PHY_ANEG_LPA,keyCode);
			if(wRegADV&PHY_ANEG_ADV_ASYMP_) {
				linkSettings|=LINK_ASYMMETRIC_PAUSE;
			}
			if(wRegADV&PHY_ANEG_ADV_SYMP_) {
				linkSettings|=LINK_SYMMETRIC_PAUSE;
			}
			if(wRegADV&PHY_ANEG_LPA_100FDX_) {
				linkSettings|=LINK_SPEED_100FD;
			}
			if(wRegADV&PHY_ANEG_LPA_100HDX_) {
				linkSettings|=LINK_SPEED_100HD;
			}
			if(wRegADV&PHY_ANEG_LPA_10FDX_) {
				linkSettings|=LINK_SPEED_10FD;
			}
			if(wRegADV&PHY_ANEG_LPA_10HDX_) {
				linkSettings|=LINK_SPEED_10HD;
			}
			pDrvCtrl->dwLinkSettings=linkSettings;
			wRegLPA&=wRegADV;
			if(wRegLPA&PHY_ANEG_LPA_100FDX_) {
				result=LINK_SPEED_100FD;
			} else if(wRegLPA&PHY_ANEG_LPA_100HDX_) {
				result=LINK_SPEED_100HD;
			} else if(wRegLPA&PHY_ANEG_LPA_10FDX_) {
				result=LINK_SPEED_10FD;
			} else if(wRegLPA&PHY_ANEG_LPA_10HDX_) {
				result=LINK_SPEED_10HD;
			}
		} else {
			if(wRegVal&PHY_BCR_SPEED_SELECT_) {
				if(wRegVal&PHY_BCR_DUPLEX_MODE_) {
					pDrvCtrl->dwLinkSettings=result=LINK_SPEED_100FD;
				} else {
					pDrvCtrl->dwLinkSettings=result=LINK_SPEED_100HD;
				}
			} else {
				if(wRegVal&PHY_BCR_DUPLEX_MODE_) {
					pDrvCtrl->dwLinkSettings=result=LINK_SPEED_10FD;
				} else {
					pDrvCtrl->dwLinkSettings=result=LINK_SPEED_10HD;
				}
			}
		}
	}
	pDrvCtrl->dwLinkSpeed=result;
}

/*********************************************************************************
Function Name:  Gpt_FreeRunCompare

Description :   Compares the time 

input: 	time to compare 	

output: void 

**********************************************************************************/
LOCAL TIME_SPAN Gpt_FreeRunCompare(UINT32 time1,UINT32 time2) 
{
	return ((TIME_SPAN)(time1-time2));
}

/*********************************************************************************
Function Name:  Gpt_ScheduleInterrupt

Description :   ISR for the General purpose timer 

input: 		device handler and time span

output: void 

**********************************************************************************/
LOCAL void Gpt_ScheduleInterrupt(PLAN9118_END_DEV pDrvCtrl,TIME_SPAN timeSpan) 
{

	UINT32 timerValue=0;
	if(timeSpan<0) timeSpan=0;
	timerValue=(UINT32)timeSpan;
	if((timerValue%2500)>=1250) {
		timerValue=(timerValue/2500)+1;
	} else {
		timerValue=(timerValue/2500);
	}
	if(timerValue>0x0000FFFFUL) {
		timerValue=0x0000FFFF;
	}
	Lan_SetRegDW(GPT_CFG,(timerValue|GPT_CFG_TIMER_EN_));
	{
		UINT32 flag = intLock();
		Lan_SetRegDW(INT_STS,INT_STS_GPT_INT_);
		intUnlock(flag);
	}
}

/*********************************************************************************
Function Name:  

Description :   

input: 		

output: void 

**********************************************************************************/
LOCAL void Gpt_CancelInterrupt(PLAN9118_END_DEV pDrvCtrl) 
{
	Lan_SetRegDW(GPT_CFG,0UL);
	{
		UINT32 flag = intLock();
		Lan_SetRegDW(INT_STS,INT_STS_GPT_INT_);
		intUnlock(flag);
	}
}

/*********************************************************************************
Function Name:  Gpt_ScheduleCallBack

input: 	device handler and Timer call back function 	

output: void 

**********************************************************************************/
LOCAL void Gpt_ScheduleCallBack(
	PLAN9118_END_DEV pDrvCtrl,
	void (*callBackFunction)(PLAN9118_END_DEV pDrvCtrl),
	UINT32 callBackTime)
{
	UINT32 slot_index=GPT_SCHEDULE_DEPTH;
	BOOL result=FALSE;

	if((callBackFunction!=NULL)&&(callBackTime!=0)) {
		UINT32 dwIntFlags=0;
		SMSC_ASSERT(pDrvCtrl!=NULL);
		spin_lock_irqsave(&(pDrvCtrl->GpTimerLock),dwIntFlags);
		{
			UINT32 index=0;
			UINT32 currentTime=Lan_GetRegDW(FREE_RUN);
			TIME_SPAN nextCallTime=MAX_TIME_SPAN;
			TIME_SPAN timeSpan=MAX_TIME_SPAN;
			BOOL rescheduleRequired=FALSE;
			for(index=0;index<GPT_SCHEDULE_DEPTH;index++) {
				if(pDrvCtrl->GptFunction[index]==NULL) {
					if(!result) {
						result=TRUE;
						pDrvCtrl->GptFunction[index]=(void *)callBackFunction;
						pDrvCtrl->GptCallTime[index]=currentTime+(2500*callBackTime);
						timeSpan=Gpt_FreeRunCompare(pDrvCtrl->GptCallTime[index],currentTime);
						if(nextCallTime>timeSpan) {
							nextCallTime=timeSpan;
							rescheduleRequired=TRUE;
							slot_index = index;
						}
					}
				} else {
					timeSpan=Gpt_FreeRunCompare(pDrvCtrl->GptCallTime[index],currentTime);
					if(nextCallTime>=timeSpan) {
						nextCallTime=timeSpan;
						rescheduleRequired=FALSE;
					}
				}
			}
			if(rescheduleRequired) {
				pDrvCtrl->Gpt_scheduled_slot_index = slot_index;
				Gpt_ScheduleInterrupt(pDrvCtrl,nextCallTime);
			}
		}
		spin_unlock_irqrestore(&(pDrvCtrl->GpTimerLock),dwIntFlags);	
	}
	if(!result) {
		SMSC_WARNING("Gpt_ScheduleCallBack: Failed");
	}
}

/*********************************************************************************
Function Name:  Gpt_CancelCallBack

input: 		Device Handler, call back function 

output: void 

**********************************************************************************/
void Gpt_CancelCallBack(
	PLAN9118_END_DEV pDrvCtrl,
	void (*callBackFunction)(PLAN9118_END_DEV pDrvCtrl))
{
	BOOL result=FALSE;
	if(callBackFunction!=NULL) {
		UINT32 dwIntFlags=0;
		SMSC_ASSERT(pDrvCtrl!=NULL);
		spin_lock_irqsave(&(pDrvCtrl->GpTimerLock),dwIntFlags);
		{
			UINT32 index=0;
			UINT32 currentTime=Lan_GetRegDW(FREE_RUN);
			TIME_SPAN nextCallTime=MAX_TIME_SPAN;
			TIME_SPAN timeSpan=MAX_TIME_SPAN;
			BOOL rescheduleRequired=FALSE;
			for(index=0;index<GPT_SCHEDULE_DEPTH;index++) {
				if(pDrvCtrl->GptFunction[index]==callBackFunction) {
					result=TRUE;
					pDrvCtrl->GptFunction[index]=(void *)NULL;
				}
				else if(pDrvCtrl->GptFunction[index]!=NULL) {
					timeSpan=Gpt_FreeRunCompare(pDrvCtrl->GptCallTime[index],currentTime);
					if(nextCallTime>=timeSpan && pDrvCtrl->Gpt_scheduled_slot_index != index) {
						nextCallTime=timeSpan;
						rescheduleRequired=TRUE;
						pDrvCtrl->Gpt_scheduled_slot_index = index;
					}
				}
			}
			if(rescheduleRequired) {
				Gpt_ScheduleInterrupt(pDrvCtrl,nextCallTime);
			}
			else if (pDrvCtrl->Gpt_scheduled_slot_index==GPT_SCHEDULE_DEPTH) {
				Gpt_CancelInterrupt(pDrvCtrl); 
			}
		}
		spin_unlock_irqrestore(&(pDrvCtrl->GpTimerLock),dwIntFlags);	
	}
	if(!result) {
	SMSC_WARNING("Gpt_CancelCallBack: Failed");
	}
}

/*********************************************************************************
Function Name:  Gpt_HandleInterrupt

input: 		Device handler , Interrupt status 

output: void

**********************************************************************************/
LOCAL void Gpt_HandleInterrupt (PLAN9118_END_DEV pDrvCtrl, UINT32 dwIntSts)
{
	UINT32 dwIntFlags=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	
	spin_lock_irqsave(&(pDrvCtrl->GpTimerLock),dwIntFlags);
	{
		UINT32 index=0;
		UINT32 currentTime=Lan_GetRegDW(FREE_RUN);
		TIME_SPAN timeSpan=MAX_TIME_SPAN;
		TIME_SPAN nextCallTime=MAX_TIME_SPAN;
		UINT32 rescheduleRequired=FALSE;

		for(index=0;index<GPT_SCHEDULE_DEPTH;index++) {
			if(pDrvCtrl->GptFunction[index]!=NULL) {
				timeSpan=Gpt_FreeRunCompare(pDrvCtrl->GptCallTime[index],currentTime);
				if(timeSpan<1250) {
					void (*callBackFunction)(LAN9118_END_DEV *pDrvCtrl);
					callBackFunction=pDrvCtrl->GptFunction[index];
					pDrvCtrl->GptFunction[index]=NULL;
					spin_unlock_irqrestore(&(pDrvCtrl->GpTimerLock),dwIntFlags);
					pDrvCtrl->Gpt_scheduled_slot_index = GPT_SCHEDULE_DEPTH;
					callBackFunction(pDrvCtrl);
					spin_lock_irqsave(&(pDrvCtrl->GpTimerLock),dwIntFlags);
				}
			}
		}

		for(index=0;index<GPT_SCHEDULE_DEPTH;index++) {
			if(pDrvCtrl->GptFunction[index]!=NULL) {
				rescheduleRequired=TRUE;
				timeSpan=Gpt_FreeRunCompare(pDrvCtrl->GptCallTime[index],currentTime);
				if(nextCallTime>timeSpan) {
					nextCallTime=timeSpan;
					pDrvCtrl->Gpt_scheduled_slot_index = index;
				}
			}
		}

		if(rescheduleRequired) {
			Gpt_ScheduleInterrupt(pDrvCtrl,nextCallTime);
		}
	}
	spin_unlock_irqrestore(&(pDrvCtrl->GpTimerLock),dwIntFlags);

   	Lan_EnableInterrupt(pDrvCtrl,INT_EN_GPT_INT_EN_);
}

/*********************************************************************************
Function Name:  GptCB_RxCompleteMulticast

Description :    
			

input: 		LAN9118_END_DEVICE

output: void 

**********************************************************************************/
LOCAL void GptCB_RxCompleteMulticast(PLAN9118_END_DEV pDrvCtrl)
{
	Rx_CompleteMulticastUpdate (pDrvCtrl);
}

/*********************************************************************************
Function Name:  GptCB_RestartBurst

Description :   

input: 	LAN9118_END_DEVICE	

output: void 

**********************************************************************************/
LOCAL void GptCB_RestartBurst(PLAN9118_END_DEV pDrvCtrl)
{
	if(pDrvCtrl->RxFlowControlActive) {
		pDrvCtrl->RxFlowBurstActive=TRUE;
		if(pDrvCtrl->RxFlowBurstWorkLoad>pDrvCtrl->RxFlowBurstMaxWorkLoad) {
			pDrvCtrl->RxFlowBurstWorkLoad-=pDrvCtrl->RxFlowBurstMaxWorkLoad;
		} else {
			pDrvCtrl->RxFlowBurstWorkLoad=0;
		}
		Gpt_ScheduleCallBack(pDrvCtrl,GptCB_RestartBurst,
				pDrvCtrl->RxFlowParameters.BurstPeriod);
	}

	Lan_EnableInterrupt(pDrvCtrl,pDrvCtrl->RxInterrupts);
	pDrvCtrl->FCRxInterruptsEn = TRUE;
}

/*********************************************************************************
Function Name:  GptCB_MeasureRxThroughput

Description :   

input: 	pointer to LAN9118_END_DEVICE

output: void 

**********************************************************************************/
LOCAL void GptCB_MeasureRxThroughput(PLAN9118_END_DEV pDrvCtrl)
{
	if(pDrvCtrl->RxFlowMeasuredMaxThroughput<pDrvCtrl->RxFlowCurrentThroughput) {
		pDrvCtrl->RxFlowMeasuredMaxThroughput=pDrvCtrl->RxFlowCurrentThroughput;
	}
	if(pDrvCtrl->RxFlowMeasuredMaxPacketCount<pDrvCtrl->RxFlowCurrentPacketCount) {
		pDrvCtrl->RxFlowMeasuredMaxPacketCount=pDrvCtrl->RxFlowCurrentPacketCount;
	}
	if(pDrvCtrl->RxFlowCurrentThroughput!=0) {
		if(pDrvCtrl->RxFlowMaxWorkLoad!=0) {
			if(!(pDrvCtrl->RxFlowControlActive)) {

				UINT32 activationLevel=
					(pDrvCtrl->RxFlowMaxWorkLoad*(100+RX_FLOW_ACTIVATION))/100;
				if(pDrvCtrl->RxFlowCurrentWorkLoad>=activationLevel) {
					pDrvCtrl->RxFlowControlActive=TRUE;
					pDrvCtrl->RxFlowBurstActive=TRUE;
					pDrvCtrl->RxFlowBurstWorkLoad=0;
					Gpt_ScheduleCallBack(pDrvCtrl,GptCB_RestartBurst,
							pDrvCtrl->RxFlowParameters.BurstPeriod);
				}
			} else {
				UINT32 deactivationLevel=
					(pDrvCtrl->RxFlowMaxWorkLoad*(100-RX_FLOW_DEACTIVATION))/100;
				if(pDrvCtrl->RxFlowCurrentWorkLoad<=deactivationLevel) {
					pDrvCtrl->RxFlowControlActive=FALSE;
				}
			}
		}
		pDrvCtrl->RxFlowCurrentThroughput=0;
		pDrvCtrl->RxFlowCurrentPacketCount=0;
		pDrvCtrl->RxFlowCurrentWorkLoad=0;
		Gpt_ScheduleCallBack(pDrvCtrl,GptCB_MeasureRxThroughput,1000);/*1000*/
	} else {
		if(pDrvCtrl->RxFlowMaxWorkLoad!=0) {
			if(pDrvCtrl->RxFlowControlActive) {
				pDrvCtrl->RxFlowControlActive=FALSE;
			}
		}
		pDrvCtrl->MeasuringRxThroughput=FALSE;
	}
}

/************************************************************************
Function Name:Eeprom_EnableAccess

Inputs:	LAN9118_END_DEVICE 

Description:This function Enable the EEPROM access

Return: Returns one if sucess
***********************************************************************/
void Eeprom_EnableAccess(PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	Lan_SetRegDW(GPIO_CFG,(g_GpioSetting&0xFF0FFFFFUL)); /* sets EEPROM  EEDIO and EECLK*/
	Lan_udelay (pDrvCtrl,100);
}

/************************************************************************
Function Name:Eeprom_DisableAccess

Inputs:	LAN9118_END_DEVICE 

Description:This function disable the EEPROM access

Return: Returns one if sucess
***********************************************************************/
void Eeprom_DisableAccess(PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	Lan_SetRegDW(GPIO_CFG,g_GpioSetting);
}

/************************************************************************
Function Name:Eeprom_IsMacAddressLoaded

Inputs:	LAN9118_END_DEVICE 

Description:This function checks the MAC address is loaded from the EEPROM 

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_IsMacAddressLoaded(PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	return (Lan_GetRegDW(E2P_CMD)&
		E2P_CMD_MAC_ADDR_LOADED_)?TRUE:FALSE;
}

/************************************************************************
Function Name:Eeprom_IsBusy

Inputs:	LAN9118_END_DEVICE 

Description:This function checks the EEPROM is BUSY in loading the ADDRH

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_IsBusy(PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	return (Lan_GetRegDW(E2P_CMD)&
		E2P_CMD_EPC_BUSY_)?TRUE:FALSE;
}

/************************************************************************
Function Name:Eeprom_Timeout

Inputs:	LAN9118_END_DEVICE 

Description:this is time out function fo the  EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_Timeout(PLAN9118_END_DEV pDrvCtrl)
{
	SMSC_ASSERT(pDrvCtrl!=NULL);
	return (Lan_GetRegDW(E2P_CMD)&
		E2P_CMD_EPC_TIMEOUT_)?TRUE:FALSE;
}

/************************************************************************
Function Name:Eeprom_WriteLocation

Inputs:	LAN9118_END_DEVICE , address and Data

Description:This function reads data from given address of the  EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_ReadLocation(
	PLAN9118_END_DEV pDrvCtrl,
	BYTE address, BYTE * data)
{
	UINT32 timeout=100000;
	UINT32 temp=0;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	SMSC_ASSERT(data!=NULL);
	if((temp=Lan_GetRegDW(E2P_CMD))&E2P_CMD_EPC_BUSY_) {
		printf("Eeprom_ReadLocation: Busy at start, E2P_CMD=0x%08lX");
		return FALSE;
	}
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_READ_|((UINT32)address)));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
	Lan_udelay(pDrvCtrl,10);
	timeout--;
	}
	if(timeout==0) {
		return FALSE;
	}
	(*data)=(BYTE)(Lan_GetRegDW(E2P_DATA));
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_WriteLocation

Inputs:	LAN9118_END_DEVICE 

Description:This function enable the access to  EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_EnableEraseAndWrite(
	PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 timeout=100000;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_) {
		SMSC_WARNING("Eeprom_EnableEraseAndWrite: Busy at start");
		return FALSE;
	}
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_EWEN_));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
		Lan_udelay(pDrvCtrl,10);
		timeout--;
	}
	if(timeout==0) {
		return FALSE;
	}
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_WriteLocation

Inputs:	LAN9118_END_DEVICE 

Description:This function disable the access to  EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_DisableEraseAndWrite(
	PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 timeout=100000;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_) {
		SMSC_WARNING("Eeprom_DisableEraseAndWrite: Busy at start");
		return FALSE;
	}
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_EWDS_));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
		Lan_udelay(pDrvCtrl,10);
		timeout--;
	}
	if(timeout==0) {
		return FALSE;
	}
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_WriteLocation

Inputs:	LAN9118_END_DEVICE, Address and Data

Description:This function is used to write the data in given address in the EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_WriteLocation(
	PLAN9118_END_DEV pDrvCtrl,BYTE address,BYTE data)
{
	UINT32 timeout=100000;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_) {
		SMSC_WARNING("Eeprom_WriteLocation: Busy at start");
		return FALSE;
	}
	Lan_SetRegDW(E2P_DATA,((UINT32)data));
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_WRITE_|((UINT32)address)));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
		Lan_udelay(pDrvCtrl,10);
		timeout--;
	}
	if(timeout==0) {
		return FALSE;
	}
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_EraseAll

Inputs:	LAN9118_END_DEVICE

Description:This function erase the MAC ADDRESS stored in the EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_EraseAll(
	PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 timeout=100000;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_) {
		SMSC_WARNING("Eeprom_EraseAll: Busy at start");
		return FALSE;
	}
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_ERAL_));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
		Lan_udelay(pDrvCtrl,10);
		timeout--;
	}
	if(timeout==0) {
		return FALSE;
	}
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_Reload

Inputs:	LAN9118_END_DEVICE

Description:This function saves the MAC ADDRESS into the EEPROM

Return: Returns one if sucess
***********************************************************************/
BOOL Eeprom_Reload(
	PLAN9118_END_DEV pDrvCtrl)
{
	UINT32 timeout=100000;
	UINT32 dwTemp=0;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	if(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_) {
		SMSC_WARNING("Eeprom_Reload: Busy at start"); /*SMSC_WARNING*/
		return FALSE;
	}
	Lan_SetRegDW(E2P_CMD,
		(E2P_CMD_EPC_BUSY_|E2P_CMD_EPC_CMD_RELOAD_));
	dwTemp = Lan_GetRegDW(BYTE_TEST); /* force flush of previous write */
	dwTemp = dwTemp;
	while((timeout>0)&&
		(Lan_GetRegDW(E2P_CMD)&E2P_CMD_EPC_BUSY_))
	{
		Lan_udelay(pDrvCtrl,10);
		timeout--;
	}
	if(timeout==0) {
		SMSC_WARNING(" timeout");
		return FALSE;
	}
	return TRUE;
}

/************************************************************************
Function Name:Eeprom_SaveMacAddress

Inputs:	LAN9118_END_DEVICE  ,UINT32, UINT32

Description:This function savs the MAC ADDRESS into the EEPROM

Output: returns one if sucess
***********************************************************************/
BOOL Eeprom_SaveMacAddress(
	PLAN9118_END_DEV pDrvCtrl,
	UINT32 dwHi16,UINT32 dwLo32)
{
	BOOL result=FALSE;
	SMSC_ASSERT(pDrvCtrl!=NULL);
	Eeprom_EnableAccess(pDrvCtrl);
	if(!Eeprom_EnableEraseAndWrite(pDrvCtrl)) goto DONE;
	if(!Eeprom_EraseAll(pDrvCtrl)) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,0,0xA5)) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,1,LOBYTE(LOWORD(dwLo32)))) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,2,HIBYTE(LOWORD(dwLo32)))) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,3,LOBYTE(HIWORD(dwLo32)))) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,4,HIBYTE(HIWORD(dwLo32)))) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,5,LOBYTE(LOWORD(dwHi16)))) goto DONE;
	if(!Eeprom_WriteLocation(pDrvCtrl,6,HIBYTE(LOWORD(dwHi16)))) goto DONE;
	if(!Eeprom_DisableEraseAndWrite(pDrvCtrl)) goto DONE;

	if(!Eeprom_Reload(pDrvCtrl)) goto DONE;
	if(!Eeprom_IsMacAddressLoaded(pDrvCtrl)) 
	{
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);

		if(dwHi16!=Mac_GetRegDW(pDrvCtrl,ADDRH,keyCode)) { 
			Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			goto DONE;
		}
		
		if(dwLo32!=Mac_GetRegDW(pDrvCtrl,ADDRL,keyCode)) {
			Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
			goto DONE;
		}
		
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);
	}	
	result=TRUE;
DONE:
	Eeprom_DisableAccess(pDrvCtrl);
	return result;
}

/*********************************************************************************
Function Name: cmd 

Description : in-line command in VxWorks shell  

input: 	Command String, arg1,...,arg3

output: void 

**********************************************************************************/
void cmd (char *str, UINT32 arg1, UINT32 arg2, UINT32 arg3)
{
	PLAN9118_END_DEV pDrvCtrl = saved_pDrvCtrl;

	if (!strcmp(str, "GET_REG")) {
		logMsg ("GET_REG::REG(offset=0x%04lX) = 0x%08lX\n", arg1, Lan_GetRegDW(arg1), 0, 0, 0, 0);	

	} else if (!strcmp(str, "SET_REG")) {
		Lan_SetRegDW(arg1, arg2);
		logMsg ("SET_REG::REG(offset=0x%04lX) = 0x%08lX\n", arg1, Lan_GetRegDW(arg1), 0, 0, 0, 0);	

	} else if (!strcmp(str, "GET_MAC_REG")) {
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		logMsg ("GET_MAC_REG::REG(offset=0x%04lX) = 0x%08lX\n", arg1, Mac_GetRegDW(pDrvCtrl,arg1,keyCode), 0, 0, 0, 0);	
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);

	} else if (!strcmp(str, "SET_MAC_REG")) {
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		Mac_SetRegDW(pDrvCtrl,arg1, arg2, keyCode);
		logMsg ("SET_MAC_REG::REG(offset=0x%04lX) = 0x%08lX\n", arg1, Mac_GetRegDW(pDrvCtrl,arg1,keyCode), 0, 0, 0, 0);	
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);

	} else if (!strcmp(str, "GET_PHY_REG")) {
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		logMsg ("GET_PHY_REG::REG(offset=0x%04lX) = 0x%04X\n", arg1, Phy_GetRegW(pDrvCtrl,arg1,keyCode), 0, 0, 0, 0);	
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);

	} else if (!strcmp(str, "SET_PHY_REG")) {
		UINT32 dwIntFlags=0;
		VL_KEY keyCode=Vl_WaitForLock(&(pDrvCtrl->MacPhyLock),&dwIntFlags);
		Phy_SetRegW(pDrvCtrl,arg1, (UINT16)arg2, keyCode);
		logMsg ("SET_PHY_REG::REG(offset=0x%04lX) = 0x%04X\n", arg1, Phy_GetRegW(pDrvCtrl,arg1,keyCode), 0, 0, 0, 0);	
		Vl_ReleaseLock(&(pDrvCtrl->MacPhyLock),keyCode,&dwIntFlags);

	} else if (!strcmp(str, "DUMP_REGS")) {
		Lan_ShowSysRegs (saved_pDrvCtrl);

	} else if (!strcmp(str, "DUMP_MAC_REGS")) {
		Lan_ShowMacRegs (saved_pDrvCtrl);

	} else if (!strcmp(str, "DUMP_PHY_REGS")) {
		Lan_ShowPhyRegs (saved_pDrvCtrl);

	} else if (!strcmp(str, "RX_XFER_MODE")) {
		if (arg1) {
		    pDrvCtrl->rx_dma = TRANSFER_REQUEST_DMA;
		} else {
		    pDrvCtrl->rx_dma = TRANSFER_PIO;
		}
		logMsg ("RX_XFER_MODE::Rx Xfer Mode = %s\n", (UINT32)((pDrvCtrl->rx_dma == TRANSFER_PIO)?"PIO":"DMA"), 0, 0, 0, 0, 0);	

        if(pDrvCtrl->rx_dma==TRANSFER_REQUEST_DMA) {
			pDrvCtrl->dwRxDmaCh=Platform_RequestDmaChannel(pDrvCtrl,RX_DMA);
            SMSC_ASSERT(pDrvCtrl->dwRxDmaCh!=TRANSFER_PIO);
            if(pDrvCtrl->dwRxDmaCh!=TRANSFER_PIO) {
                pDrvCtrl->RxDmaChReserved=TRUE;
            }
        } else {
       		pDrvCtrl->dwRxDmaCh=TRANSFER_PIO;
	        pDrvCtrl->RxDmaChReserved=FALSE;        	
       	}

	} else if (!strcmp(str, "TX_XFER_MODE")) {
		if (arg1) {
		    pDrvCtrl->tx_dma = TRANSFER_REQUEST_DMA;
		} else {
		    pDrvCtrl->tx_dma = TRANSFER_PIO;
		}
		logMsg ("TX_XFER_MODE::Tx Xfer Mode = %s\n", (UINT32)((pDrvCtrl->tx_dma == TRANSFER_PIO)?"PIO":"DMA"), 0, 0, 0, 0, 0);	

       if(pDrvCtrl->tx_dma==TRANSFER_REQUEST_DMA) {
            pDrvCtrl->dwTxDmaCh=Platform_RequestDmaChannel(pDrvCtrl,TX_DMA);
            SMSC_ASSERT(pDrvCtrl->dwTxDmaCh!=TRANSFER_PIO);
            if(pDrvCtrl->dwTxDmaCh!=TRANSFER_PIO) {
                pDrvCtrl->TxDmaChReserved=TRUE;
            }
        } else {
	    	pDrvCtrl->dwTxDmaCh=TRANSFER_PIO;      
            pDrvCtrl->TxDmaChReserved=FALSE;
        }

	} else if (!strcmp(str, "RX_DMA_TRSHLD")) {
	    pDrvCtrl->dwRxDmaThreshold = arg1;
		logMsg ("RX_DMA_TRSHLD::pDrvCtrl->dwRxDmaThreshold = %ld\n", pDrvCtrl->dwRxDmaThreshold, 0, 0, 0, 0, 0);	

	} else if (!strcmp(str, "TX_DMA_TRSHLD")) {
		pDrvCtrl->dwTxDmaThreshold = arg1;	
		logMsg ("TX_DMA_TRSHLD::pDrvCtrl->dwTxDmaThreshold = %ld\n", pDrvCtrl->dwTxDmaThreshold, 0, 0, 0, 0, 0);	

	} else if (!strcmp(str, "?")) {
		logMsg ("cmd \"Command\", arg1, arg2 \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("Command & args: \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("GET_REG, offset \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("SET_REG, offset, val \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("GET_MAC_REG, offset \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("SET_MAC_REG, offset, val \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("GET_PHY_REG, offset \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("SET_PHY_REG, offset, val \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("RX_XFER_MODE, 1=DMA or 0=PIO \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("TX_XFER_MODE, 1=DMA or 0=PIO \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("RX_DMA_TRSHLD, Val \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("TX_DMA_TRSHLD, Val \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("DUMP_REGS \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("DUMP_MAC_REGS \n", 1, 2, 3, 4, 5, 6);	
		logMsg ("DUMP_PHY_REGS \n", 1, 2, 3, 4, 5, 6);	
	}
}

/*********************************************************************************
Function Name:  DumpBuffer

Description :   print buffer

input: 	buffer pointer and length	

output: void 

**********************************************************************************/
#ifdef	LAN9118_DEBUG
LOCAL void DumpBuffer(UINT8 *data, UINT32 len)
{
	int i;
	for (i=0;i<len;i++)
	{
		if ((i%16) ==  0)
		{
			printf("\n%04x:  ", i);
		}
		printf("%02x ", data[i]);
	}
	printf("\n\n");
}
#endif

/*********************************************************************************
Function Name:  Vl_InitLock

Description :   Lock initialization

input: 		

output: void 

**********************************************************************************/
void Vl_InitLock(PVERIFIABLE_LOCK pVl)
{
	SMSC_ASSERT(pVl!=NULL);

	spin_lock_init(&(pVl->Lock));
	pVl->KeyCode=0;
}

/*********************************************************************************
Function Name:  Vl_CheckLock

Description :  Check to see if lock has already been taken 

input: 		

output: void 

**********************************************************************************/
BOOL Vl_CheckLock(PVERIFIABLE_LOCK pVl,VL_KEY keyCode)
{
	BOOL result=FALSE;

	SMSC_ASSERT(pVl!=NULL);
	if(keyCode==pVl->KeyCode) 
		result=TRUE;

	return result;
}

/*********************************************************************************
Function Name:  Vl_WaitForLock

Description :   Wait for lock till becomes available

input: 		

output: void 

**********************************************************************************/
VL_KEY Vl_WaitForLock(PVERIFIABLE_LOCK pVl,UINT32 *pdwIntFlags)
{
	VL_KEY result=0;

	SMSC_ASSERT(pVl!=NULL);
	SMSC_ASSERT(pVl->Lock!=NULL);
	
	spin_lock_irqsave(&(pVl->Lock),	(*pdwIntFlags));
	pVl->KeyCode++;
	if(pVl->KeyCode>0x80000000UL) {
		pVl->KeyCode=1;
	}
	result=pVl->KeyCode;

	return result;
}

/*********************************************************************************
Function Name: Vl_ReleaseLock 

Description :   Release lock

input: 		

output: void 

**********************************************************************************/
void Vl_ReleaseLock(PVERIFIABLE_LOCK pVl,VL_KEY keyCode,UINT32 *pdwIntFlags)
{
	SMSC_ASSERT(pVl!=NULL);
	SMSC_ASSERT(pVl->Lock!=NULL);
	SMSC_ASSERT(pVl->KeyCode==keyCode);

	spin_unlock_irqrestore(&(pVl->Lock), (*pdwIntFlags));
}

